/** ScaFES
 *  Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#include "ScaFES.hpp"
template<typename CT, std::size_t DIM>
class HeatEqnFDM : public ScaFES::Problem<HeatEqnFDM<CT,DIM>, CT, DIM> {
    public:
        HeatEqnFDM(ScaFES::Parameters const& cl,
                   ScaFES::GridGlobal<DIM> const& gg,
            bool useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile
                = std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>() )
            : ScaFES::Problem<HeatEqnFDM<CT, DIM>, CT, DIM>(
                     cl, gg, useLeapfrog, nameDatafield, stencilWidth,
                     isKnownDf, nLayers, defaultValue, writeToFile,
                     computeError, geomparamsInit) { }

        void evalInner(std::vector< ScaFES::DataField<CT, DIM> >& vNew,
                       ScaFES::Ntuple<int,DIM> const& idxNode,
                       int const& /*timestep*/) {
            ScaFES::Ntuple<double,DIM> x = this->coordinates(idxNode);
            vNew[0](idxNode) = 0.0;
            vNew[1](idxNode) = 0.0;
            vNew[2](idxNode) = 1.0;
            for (std::size_t pp = 0; pp < DIM; ++pp) {
                vNew[2](idxNode) *= ( x[pp] * (x[pp] - 0.5)
                                      * (x[pp] - 0.5) * (1.0 - x[pp]) );
            }
        }
        void evalBorder(std::vector< ScaFES::DataField<CT, DIM> >& vNew,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                        int const& timestep) {
            this->evalInner(vNew, idxNode, timestep);
        }

        template<typename TT>
        void initInner(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                       std::vector<TT> const& /*vOld*/,
                       ScaFES::Ntuple<int,DIM> const& idxNode,
                       int const& /*timestep*/) {
            ScaFES::Ntuple<double,DIM> x = this->coordinates(idxNode);
            vNew[0](idxNode) = 1.0;
            for (std::size_t pp = 0; pp < DIM; ++pp) {
                vNew[0](idxNode) *= ( x[pp] * (x[pp] - 0.5)
                                      * (x[pp] - 0.5) * (1.0 - x[pp]) );
            }
        }
        template<typename TT>
        void initBorder(std::vector< ScaFES::DataField<TT, DIM> >& vNew,
                        std::vector<TT> const& vOld,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                        int const& timestep) {
            this->template initInner<TT>(vNew, vOld, idxNode, timestep);
        }

        template<typename TT>
        void updateInner(std::vector<ScaFES::DataField<TT,DIM>>& vNew,
                        std::vector<ScaFES::DataField<TT,DIM>> const& vOld,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                       int const& /*timestep*/) {
         vNew[0](idxNode) = vOld[0](idxNode)
                            - this->tau() * this->knownDf(0, idxNode);
         for (std::size_t pp = 0; pp < DIM; ++pp) {
            vNew[0](idxNode) += this->tau() * (
                 2.0 * vOld[0](idxNode)
                     - vOld[0](this->connect(idxNode, 2*pp))
                     - vOld[0](this->connect(idxNode, 2*pp+1)) )
                     / (this->gridsize(pp) * this->gridsize(pp))
                     -  this->knownDf(1, idxNode);
         }
       }
       template<typename TT>
       void updateBorder(std::vector<ScaFES::DataField<TT,DIM>>& vNew,
                         std::vector<ScaFES::DataField<TT,DIM>>const& /*vOld*/,
                         ScaFES::Ntuple<int,DIM> const& idxNode,
                         int const& /*timestep*/) {
           vNew[0](idxNode) = this->knownDf(1, idxNode);
       }

       template<typename TT>
       void updateInner2(std::vector<ScaFES::DataField<TT,DIM>>&,
                         std::vector<ScaFES::DataField<TT,DIM>> const&,
                         ScaFES::Ntuple<int,DIM> const&,
                         int const&) {}

       template<typename TT>
       void updateBorder2(std::vector<ScaFES::DataField<TT,DIM>>&,
                          std::vector<ScaFES::DataField<TT,DIM>>const&,
                          ScaFES::Ntuple<int,DIM> const&,
                          int const&) {}
};

const int DIM = 3;

int main(int argc, char *argv[]) {
    ScaFES::Parameters paramsCl(argc, argv);
    ScaFES::GridGlobal<DIM> gg(paramsCl);
    std::vector<std::string> nameDatafield(3);
    nameDatafield[0] = "F";
    nameDatafield[1] = "G";
    nameDatafield[2] = "U";
    std::vector<int> stencilWidth(3);
    stencilWidth[0] = 0;
    stencilWidth[1] = 0;
    stencilWidth[2] = 1;
    std::vector<bool> isKnownDf(3);
    isKnownDf[0] = true;
    isKnownDf[1] = true;
    isKnownDf[2] = false;
    std::vector<int> nLayers(3);
    nLayers[0] = 0;
    nLayers[1] = 0;
    nLayers[2] = 0;
    std::vector<double> defaultValue(3);
    defaultValue[0] = 0.0;
    defaultValue[1] = 0.0;
    defaultValue[2] = 0.0;
    std::vector<ScaFES::WriteHowOften> writeToFile(3, ScaFES::WriteHowOften::NEVER);
    std::vector<bool> computeError(3);
    computeError[0] = false;
    computeError[1] = false;
    computeError[2] = true;
    std::vector<double> geomparamsInit;

    HeatEqnFDM<double, DIM> ppp(paramsCl, gg, false,
                     nameDatafield, stencilWidth,
                     isKnownDf, nLayers, defaultValue, writeToFile,
                     computeError, geomparamsInit);
    ppp.iterateOverTime();
    return 0;
}

