/** SCADOPT
 *  Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCADOPT_PROBLEM_HPP_
#define SCADOPT_PROBLEM_HPP_

#include "SCADOPT_Config.hpp"

#include <iostream>
#include <ios>
#include <fstream>
#include <string>
#include <vector>
#include <cassert>
#include <unordered_map>

extern "C" {
#ifdef _OPENMP
#include <omp.h>
#endif
}

#ifdef SCADOPT_HAVE_ADOLC
#include <adolc/adolc.h>
#include <adolc/adouble.h>
#ifdef _OPENMP
#include <adolc/adolc_openmp.h>
#endif
#endif

#ifdef SCADOPT_HAVE_BOOST
#include <boost/type_traits.hpp>
#include <boost/assign/list_of.hpp>
#endif

#include "mpi.h"

#ifdef SCADOPT_HAVE_BOOST_MPI
#include <boost/mpi.hpp>
#endif

#ifdef VTRACE
#include "vt_user.h"
#endif

#ifdef SCADOPT_HAVE_LBFGSB
extern "C" {
#include <lbfgsb.h>
}
#endif

#include "ScaFES.hpp"

namespace SCADOPT
{

/*******************************************************************************
 ******************************************************************************/
/** \n The class \c Problem is designed to describe a PDE constrained
 * optimization problem on a given MPI partition. The optimization problem
 * usually consists of an initial boundary value problem, a target functional,
 * and an initial guess of the (geometrical) parameters.
 *
 * \n The user has to implement an own class that is inherited by
 * the class \c ScaFES::Problem containing different methods.
 */
template<class OWNPRBLM, typename CT = double, std::size_t DIM = 3>
class Problem
   : public ScaFES::Problem<Problem<OWNPRBLM, CT, DIM>, CT, DIM>
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
#ifdef SCADOPT_HAVE_ADOLC
        /** Type definition for the member computeTf methods of this class. */
        typedef void (Problem<OWNPRBLM,CT,DIM>::*TypeFuncComputeTfToTrace)(
                adouble&,
                std::vector< ScaFES::DataField<adouble,DIM> > const&,
                int const&);
#endif

        /** Type definition for the member computeTf methods of this class. */
        typedef void (Problem<OWNPRBLM,CT,DIM>::*TypeFuncComputeTfToEval)(
                CT&,
                std::vector<CT>&,
                std::vector< ScaFES::DataField<CT,DIM> > const&,
                std::vector< ScaFES::DataField<CT,DIM> > const&,
                int const&);

        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** Creates own default constructor. */
        Problem();

        /** Creates own constructor. */
        Problem(
            ScaFES::Parameters const& params,
            ScaFES::GridGlobal<DIM> const& gg,
            bool const& useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<bool> const& optimizeDf = std::vector<bool>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile 
                = std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>(),
            std::vector<CT> const& geomparamsSol = std::vector<CT>()
        );

        /** Creates default copy constructor. */
        Problem(Problem<OWNPRBLM,CT,DIM> const& /*ppp*/) = default;

        /** Creates compiler provided assignment operator. */
        Problem& operator=(Problem<OWNPRBLM, CT, DIM> const& /*ppp*/) = default;

        /** Creates own destructor. */
        virtual ~Problem() = default;

        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns current value of target functional. */
        CT const& tf() const;

        /** Returns current value of gradient of target functional. */
        std::vector<CT> const& gradTf() const;

        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Compute new target functional and its gradient. */
        void compNewTfAndGradTf();

        /** Iterates over time (including computation of tf). */
        void iterateOverTimeInternal();

        /** Solves the current inverse problem by optimizing a target function
         * using a given optimizer. Currently: "lbfgsb".
         * This is the interface for optimizers like L-BFGS-B or IP-OPT. */
        void minimizeTargetfct(
            std::string const& optimizer
        );

        /** Evaluates ALL data fields at one given global inner
         *  grid node. 
         * The concrete initialisation function must be implemented by the
         * user in a sub class.
         */

        template<typename TT>
        void evalInner(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Evaluates ALL data fields at one given global border
         *  grid node. 
         * The concrete initialisation function must be implemented by the
         * user in a sub class.
         */
        template<typename TT>
        void evalBorder(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Initializes all unknown data fields at one given global inner
         *  grid node. 
         * The concrete initialisation function must be implemented by the
         * user in a sub class.
         */
        template<typename TT>
        void initInner(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            std::vector<TT> const& dfIndep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Initializes all unknown data fields at one given global border
         *  grid node. 
         * The concrete initialisation function must be implemented by the
         * user in a sub class.
         */
        template<typename TT>
        void initBorder(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            std::vector<TT> const& dfIndep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Initializes all unknown data fields at one given global inner
         *  grid node. 
         * The concrete initialisation function must be implemented by the
         * user in a sub class.
         */
        template<typename TT>
        void initInner(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Initializes all unknown data fields at one given global border
         *  grid node. 
         * The concrete initialisation function must be implemented by the
         * user in a sub class.
         */
        template<typename TT>
        void initBorder(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Updates all unknown data fields at one given global inner
         *  grid node. 
         * The concrete update function must be implemented by the
         * user in a sub class. */
        template<typename TT>
        void updateInner(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Updates all unknown data fields at one given global border
         *  grid node. 
         * The concrete update function must be implemented by the
         * user in a sub class. */
        template<typename TT>
        void updateBorder(
            std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Updates (2nd cycle) all unknown data fields at one given global 
         *  inner grid node using the leap frog scheme.
         * The concrete update function must be implemented by the
         * user in a sub class. */
        template<typename TT>
        void updateInner2(
            std::vector< ScaFES::DataField<TT, DIM> >& vectNew,
            std::vector< ScaFES::DataField<TT, DIM> > const& vectOld,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

        /** Updates (2nd cycle) all unknown data fields at one given global 
         *  border grid node using the leap frog scheme.
         * The concrete update function must be implemented by the
         * user in a sub class. */
        template<typename TT>
        void updateBorder2(
            std::vector< ScaFES::DataField<TT, DIM> >& vectNew,
            std::vector< ScaFES::DataField<TT, DIM> > const& vectOld,
            ScaFES::Ntuple<int,DIM> const& idxNode,
            int const& timestep
        );

    private:
        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns if the target functional phase was already traced. */
        bool const& isTracedTf() const;

        /** Returns the tape identifier corresponding to the tape containing
         *  the target functional at all inner partition grid nodes. */
        short int const& tapeIdComputeTfPartitionInner() const;

        /** Returns the tape identifier corresponding to the tape containing 
         *  the target functional at all border partition grid nodes. */
        short int const& tapeIdComputeTfPartitionBorder() const;

        /** Returns the tape identifier corresponding to the tape containing 
         *  the target functional at all partition grid nodes. */
        short int const& tapeIdComputeTfPartitionAll() const;

        /** Returns the number of variables during optimization (memNormal). */
        int const& nParamsToOptimize() const;

        /** Returns the number of directions. */
        std::vector<bool> const& optimizeDf() const;

        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Solves the current inverse problem by optimizing a target function
         * using L-BFGS-B and based on the grid ansatz or geometrical parameters
         * ansatz. */
        void minimizeTargetfctUsingLbfgsb(int const& optimVersion);

        /*--------------------------------------------------------------------*/
        /** Computes the value of the target functional.
         *  The computation is included in a loop over all time steps 
         *  as the target functional can be time-dependent. */
        void compTargetfunc(int const& timestep);

        /*--------------------------------------------------------------------*/
        /** Computes target functional using asynchronous MPI communication. */
        void computeTfAsynchronously(
            TypeFuncComputeTfToEval funcLocalInner,
            TypeFuncComputeTfToEval funcLocalBorder,
            CT& dfDep,
            std::vector<CT>& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            int const& timestep
        );

        /** Computes target functional using synchronous MPI communication. */
        void computeTfSynchronously(
            TypeFuncComputeTfToEval funcGlobal,
            CT& dfDep,
            std::vector<CT>& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepDomain,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndepDomain,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepBdry,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndepBdry,
            const int& timestep
        );

        /*--------------------------------------------------------------------*/
        /** Computes the target functional at all inner grid partition nodes
         *  at a given time step. */
        template<typename TT>
        void computeTfLocalInner(
            TT& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Computes the target functional at all border grid partition nodes
         *  at a given time step. */
        template<typename TT>
        void computeTfLocalBorder(
            TT& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Computes the target functional at all inner grid global nodes
         *  at a given time step. */
        template<typename TT>
        void computeTfGlobalInner(
            TT& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Computes the target functional at all border grid global nodes
         *  at a given time step. */
        template<typename TT>
        void computeTfGlobalBorder(
            TT& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Traces the target function at a given global node of the
         * GLOBAL grid. */
        void traceComputeTf(
            CT& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            int const& timestep
        );

#ifdef SCADOPT_HAVE_ADOLC
        /** Creates 1 tape for tf phase at all grid partition nodes. */
        void createTapesForComputeTf(
            short int const& tapeId,
            TypeFuncComputeTfToTrace funcComputeTf,
            CT& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            adouble& aDfDep,
            std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
            int const& timestep
        );

        /** Creates 2 tapes for tf phase at all grid partition nodes:
         * 1 tape for the tf phase at all inner grid partition nodes.
         * 1 tape for the tf phase at all border grid partition nodes. */
        void createTapeForComputeTfPartitionAll(
            short int const& tapeId,
            TypeFuncComputeTfToTrace funcInner,
            TypeFuncComputeTfToTrace funcBorder,
            CT& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            adouble& aDfDep,
            std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
            int const& timestep
        );
#endif

        /** Evaluates the target function
         *  at all grid partition nodes using the AD forward mode. */
        void computeTf_zosForward(
            CT& dfDep,
            short int const& tapeId,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Evaluates the target function and the its gradient 
         *  at all grid partition nodes using the AD forward mode. */
        void computeTf_fovForward(
            CT& dfDep,
            std::vector<CT>& dfGradDep,
            short int const& tapeId,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            int const& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperComputeTfPartitionAll(
            CT& dfDep,
            std::vector<CT>& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperComputeTfPartitionBorder(
            CT& dfDep,
            std::vector<CT>& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
         void wrapperComputeTfPartitionInner(
            CT& dfDep,
            std::vector<CT>& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /*--------------------------------------------------------------------*/
        /** Collects a scalar from each grid partition and sums it up
         *  at the given MPI process. */
        void collectScalar(
            CT& scalarGlobal,
            CT const& scalarLocal,
            int const& rankOptim
        );

        /** Collects an one-dimensional MPI local array from all grid 
         * partitions in an one-dimensional MPI global array. */
        void collectArray1D(
            std::vector<CT>& arrayGlobal,
            std::vector<CT>& arrayLocal,
            int** const& rcounts,
            int** const& displs,
            int* const& nVectLocal,
            int const& nTotalLocal,
            int const& nDfsToOptimize,
            int const& rankOptim
        );

        /** Distributes an one-dimensional MPI-global arrary 
         *  to one-dimensional MPI local arrays 
         *  defined on the grid partitions. */
        void distributeArray1D(
            std::vector<CT>& arrayLocal,
            std::vector<CT>& arrayGlobal,
            int** const& rcounts,
            int** const& displs,
            int* const& nVectLocal,
            int const& nTotalLocal,
            int const& nDfsToOptimize,
            int const& rankOptim
        );

        /** Gets the number of data fields after which should be optimized. */
        int getNdataFieldsForOptimization() const;

        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Vector of Wall clock times: Contains accumulated times 
         * of all forward iterations.
         * tIterate, tTrace, tInit, tUpdate, tCompTf,
         * tSync, tWrite, tSolution. */
        std::unordered_map<std::string, double> mWallClockTimesOptim;
    
        /** Tape identifier of target functional phase. */
        short int mTapeIdComputeTfPartitionInner;

        /** Tape identifier of target functional phase. */
        short int mTapeIdComputeTfPartitionBorder;

        /** Tape identifier of target functional phase. */
        short int mTapeIdComputeTfPartitionAll;

        /** Is target functional already traced? */
        bool mIsTracedCompTf;

        /** Optimize data field? */
        std::vector<bool> mOptimizeDf;

        /** Geomparams based ansatz: Reference if known. */
        std::vector<CT> mVectParamsRef;

        /** Value of target functional. */
        CT mTf;

        /** Grid based ansatz: Gradient of target functional (at memAll). */
        std::vector< ScaFES::DataField<CT, DIM> > mVectGradTf;

        /** Memory for gradient of target functional.
         * This will be used for the grid based AND the parameter based ansatz.
         * \remark Has to be a continuous memory lump for ADOL-C and L-BFGS-B. 
         */
        std::vector<CT> mMemoryVectGradTf;

        /** Number of variables which should be optimized (memNormal). */
        int mNparamsToOptimize;

        /** Iterate of optimization loop (at memNormal without halo nodes). */
        std::vector< ScaFES::DataField<CT, DIM> > mVectIterate;

        /** Memory for storing values of iterate of opt. loop. */
        std::vector<CT> mMemoryVectIterate;

        /** Gradient of target functional
         * within optimization loop (at memNormal without halo nodes). */
        std::vector< ScaFES::DataField<CT, DIM> > mVectIterateGradTf;

        /** Memory for storing values of gradient of tf of opt. loop. */
        std::vector<CT> mMemoryVectIterateGradTf;

}; // End of class. //

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::Problem()
    : ScaFES::Problem<Problem<OWNPRBLM, CT, DIM>, CT, DIM>()
    , mWallClockTimesOptim()
    , mTapeIdComputeTfPartitionInner(21844)
    , mTapeIdComputeTfPartitionBorder(22844)
    , mTapeIdComputeTfPartitionAll(23844)
    , mIsTracedCompTf(false)
    , mOptimizeDf()
    , mVectParamsRef()
    , mTf(static_cast<CT>(0))
    , mVectGradTf()
    , mMemoryVectGradTf()
    , mNparamsToOptimize(0)
    , mVectIterate()
    , mMemoryVectIterate()
    , mVectIterateGradTf()
    , mMemoryVectIterateGradTf()
{
     this->mWallClockTimesOptim["communic"] = 0.0;
     this->mWallClockTimesOptim["lbfgsb"] = 0.0;  
     this->mWallClockTimesOptim["compTf"] = 0.0;  
 }
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::Problem(
    ScaFES::Parameters const& params,
    ScaFES::GridGlobal<DIM> const& gg,
    bool const& useLeapfrog,
    std::vector<std::string> const& nameDatafield,
    std::vector<int> const& stencilWidth,
    std::vector<bool> const& isKnownDf,
    std::vector<int> const& nLayers,
    std::vector<bool> const& optimizeDf,
    std::vector<CT> const& defaultValue,
    std::vector<ScaFES::WriteHowOften> const& writeToFile,
    std::vector<bool> const& computeError,
    std::vector<CT> const& geomparamsInit,
    std::vector<CT> const& geomparamsSol
)
    : ScaFES::Problem< Problem<OWNPRBLM, CT, DIM >, CT, DIM>(
        params,
        gg,
        useLeapfrog,
        nameDatafield,
        stencilWidth,
        isKnownDf,
        nLayers,
        defaultValue,
        writeToFile,
        computeError,
        geomparamsInit
    )
    , mWallClockTimesOptim()
    , mTapeIdComputeTfPartitionInner(21844+this->myRank())
    , mTapeIdComputeTfPartitionBorder(22844+this->myRank())
    , mTapeIdComputeTfPartitionAll(23844+this->myRank())
    , mIsTracedCompTf(false)
    , mOptimizeDf(optimizeDf)
    , mVectParamsRef(geomparamsSol)
    , mTf(static_cast<CT>(0))
    , mVectGradTf()
    , mMemoryVectGradTf()
    , mNparamsToOptimize(0)
    , mVectIterate()
    , mMemoryVectIterate()
    , mVectIterateGradTf()
    , mMemoryVectIterateGradTf()
{
    this->mWallClockTimesOptim["communic"] = 0.0;
    this->mWallClockTimesOptim["lbfgsb"] = 0.0; 
    this->mWallClockTimesOptim["compTf"] = 0.0; 

    /*------------------------------------------------------------------------*/
    int idDf = 0;
    int idOptimizeAndUnknownDf = 0;
    int nDataFields = nameDatafield.size();

    /*------------------------------------------------------------------------*/
    int nNodesTotalAllUnknownDf = 0;
    for (std::size_t ii = 0; ii < this->vectUnknownDfsDomNew().size(); ++ii) {
        nNodesTotalAllUnknownDf 
            += this->vectUnknownDfsDomNew()[ii].nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    if (!(this->isParamsBased())) {
        int idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii) {
            if (!this->isKnownDf().at(ii)) {
                if (this->optimizeDf().at(ii)) {
                    this->mNparams
                     += this->vectUnknownDfsDomNew().at(idDf).nElemsAllocated();
                }
                ++idDf;
            }
        }
    }
    /*------------------------------------------------------------------------*/
    // Create memory for local arrays.
    this->mNparamsToOptimize = 0;
    if (this->isParamsBased()) {
        this->mNparamsToOptimize = this->vectParamsCurr().size();
    }
    else {
        this->mNparamsToOptimize = 0;
        idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii) {
            if (!this->isKnownDf().at(ii)) {
                if (this->optimizeDf().at(ii)) {
                    this->mNparamsToOptimize
                      += this->vectUnknownDfsDomNew().at(idDf).memNormal().nNodesSub().size();
                }
                ++idDf;
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Determine number of unknown data fields.
    int nUnknownDfs = 0;
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (!this->isKnownDf().at(ii)) {
            ++nUnknownDfs;
        }
    }

    /*------------------------------------------------------------------------*/
    // Determine number of elements (memAll) of known and unknown data fields.
    std::vector <ScaFES::Ntuple<int,DIM> > nMemAllOfDf(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii) {
        ScaFES::Ntuple<int,DIM> idxNodeFirst
            = gg.partition(params.myWorld().rank()).idxNodeFirstSub()
              - ScaFES::Ntuple<int, DIM>(stencilWidth.at(ii));
        ScaFES::Ntuple<int,DIM> idxNodeLast
            = gg.partition(params.myWorld().rank()).idxNodeLastSub()
              + ScaFES::Ntuple<int, DIM>(stencilWidth.at(ii));
        nMemAllOfDf[ii] 
            = idxNodeLast - idxNodeFirst + ScaFES::Ntuple<int, DIM>(1);
    }

    /*------------------------------------------------------------------------*/
    // Determine number of elements (memNormal) of known + unknown data fields.
    int nMemNormalUnknownDfs = 0;
    int nMemNormalUnknownAndOptimizeDfs = 0;
    std::vector< ScaFES::Ntuple<int,DIM> > nMemNormalOfDf(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii) {
        ScaFES::Ntuple<int,DIM> idxNodeFirstNormal
            = gg.partition(params.myWorld().rank()).idxNodeFirstSub();
        ScaFES::Ntuple<int,DIM> idxNodeLastNormal
            = gg.partition(params.myWorld().rank()).idxNodeLastSub();
        nMemNormalOfDf[ii] 
            = idxNodeLastNormal 
                - idxNodeFirstNormal + ScaFES::Ntuple<int, DIM>(1);
        if (!this->isKnownDf().at(ii)) {
            nMemNormalUnknownDfs += nMemNormalOfDf[ii].size();
            if (this->optimizeDf().at(ii)) {
                nMemNormalUnknownAndOptimizeDfs += nMemNormalOfDf[ii].size();
            }
        }
    }

    /*------------------------------------------------------------------------*/
    CT* elemDataCurr = static_cast<CT*>(0);

    // Create memory for gradient of target functional (memAll)
    // Create data field for gradient of tf.
    const int NCOMPS_TF = 1;
    if (0 < this->nColumns()) {
        this->mMemoryVectGradTf.resize(
            NCOMPS_TF * this->nColumns(),
            static_cast<CT>(0)
        );
    }
    // Create corresponding data field only in case of the grid based ansatz.
    if (!(this->isParamsBased())) {
        int nDataFieldsForOptimization = this->getNdataFieldsForOptimization();
        elemDataCurr = this->mMemoryVectGradTf.data();
        this->mVectGradTf.reserve(nDataFieldsForOptimization);
        idDf = 0;
        idOptimizeAndUnknownDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii) {
            if (!this->isKnownDf().at(ii)) {
                if (this->optimizeDf().at(ii)) {
                    std::string nameDatafieldCurr 
                        = nameDatafield.at(ii) + "GradTf";
                    this->mVectGradTf.push_back(
                        ScaFES::DataField<CT, DIM>(
                            nameDatafieldCurr,
                            this->params(),
                            this->globalGrid(),
                            this->vectUnknownDfsDomOld()[idDf].memAll(),
                            stencilWidth.at(ii),
                            elemDataCurr,
                            NCOMPS_TF,
                            this->vectUnknownDfsDomOld()[idDf].memHole(),
                            nLayers.at(ii),
                            writeToFile.at(ii)
                        )
                    );
                    elemDataCurr += (nMemAllOfDf.at(ii).size() * NCOMPS_TF);
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }
        /*--------------------------------------------------------------------*/
        // Set default value for each iterate data field.
        idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii) {
            if (!(this->isKnownDf().at(ii))) {
                if (this->optimizeDf().at(ii)) {
                    this->mVectGradTf.at(idDf) = defaultValue.at(ii);
                    ++idDf;
                }
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Create (local) iterate of optimization algorithm.
    // Do not create a corresponding data field for the (local) iterate
    // in case of the parameter based ansatz.
    if (this->isParamsBased()) {
        this->mMemoryVectIterate.resize(
            this->nParamsToOptimize(),
            static_cast<CT>(0)
        );
    }
    else {
        if (0 < nMemNormalUnknownAndOptimizeDfs) {
            this->mMemoryVectIterate.resize(
                nMemNormalUnknownAndOptimizeDfs * 1,
                static_cast<CT>(0)
            );
        }
        int nDataFieldsForOptimization = this->getNdataFieldsForOptimization();
        elemDataCurr = this->mMemoryVectIterate.data();
        this->mVectIterate.reserve(nDataFieldsForOptimization);
        idDf = 0;
        idOptimizeAndUnknownDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii) {
            if (!this->isKnownDf().at(ii)) {
                if (this->optimizeDf().at(ii)) {
                    std::string nameDatafieldOld 
                        = nameDatafield.at(ii) + "Iterate";
                    this->mVectIterate.push_back(
                        ScaFES::DataField<CT, DIM>(
                            nameDatafieldOld,
                            this->params(),
                            this->globalGrid(),
                            this->vectUnknownDfsDomOld()[idDf].memAll(),
                            0,
                            elemDataCurr,
                            1,
                            this->vectUnknownDfsDomOld()[idDf].memHole(),
                            nLayers.at(ii),
                            writeToFile.at(ii)
                        )
                    );
                    elemDataCurr += nMemNormalOfDf.at(ii).size() * 1;
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }

        /*--------------------------------------------------------------------*/
        // Set default value for each iterate data field.
        idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii) {
            if (!(this->isKnownDf().at(ii))) {
                if (this->optimizeDf().at(ii)) {
                    this->mVectIterate.at(idDf) = defaultValue.at(ii);
                    ++idDf;
                }
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Create gradient of target func. for optimization algorithm (memNormal).
    if (this->isParamsBased()) {
        if (0 < this->nParamsToOptimize()) {
            this->mMemoryVectIterateGradTf.resize(
                this->nParamsToOptimize(),
                static_cast<CT>(0)
            );
        }
    }
    else {
        if (0 < nMemNormalUnknownAndOptimizeDfs) {
            this->mMemoryVectIterateGradTf.resize(
                nMemNormalUnknownAndOptimizeDfs * 1,
                static_cast<CT>(0)
            );
        }
        int nDataFieldsForOptimization = this->getNdataFieldsForOptimization();
        elemDataCurr = this->mMemoryVectIterateGradTf.data();
        this->mVectIterateGradTf.reserve(nDataFieldsForOptimization);
        idDf = 0;
        idOptimizeAndUnknownDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii) {
            if (!(this->isKnownDf().at(ii))) {
                if (this->optimizeDf().at(ii)) {
                    std::string nameDatafieldCurr
                        = nameDatafield.at(ii) + "GradTfIter";
                    this->mVectIterateGradTf.push_back(
                        ScaFES::DataField<CT, DIM>(
                            nameDatafieldCurr,
                            this->params(),
                            this->globalGrid(),
                            this->vectUnknownDfsDomOld()[idDf].memAll(),
                            0,
                            elemDataCurr,
                            1,
                            this->vectUnknownDfsDomOld()[idDf].memHole(),
                            nLayers.at(ii)
                        )
                    );
                    elemDataCurr += nMemNormalOfDf.at(ii).size() * 1;
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }
    }
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline CT const& Problem<OWNPRBLM, CT, DIM>::tf() const
{
    return this->mTf;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<CT> const& Problem<OWNPRBLM, CT, DIM>::gradTf() const
{
    return this->mMemoryVectGradTf;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::compNewTfAndGradTf()
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute new tf and grad(tf)..." << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    // Copy initial guess to first iterate of optimization loop.
    if (this->isParamsBased()) {
        for (int ll = 0; ll < this->nParamsToOptimize(); ++ll) {
            this->mMemoryVectIterate[ll] = this->vectParamsCurr().at(ll);
        }
    }
    // REMARK: Initial guess is already given as MPI local data 
    // field mVectUnknownDfsDomOld.
    // * mVectIterate works on grid of memNormal.
    // * mVectUnknownDfsDomOld works on grid of memAll.
    else {
        int idOptimizeAndUnknownDf = 0;
        int idDf = 0;
        for (std::size_t ii = 0; ii < this->optimizeDf().size(); ++ii) {
            if (!this->isKnownDf().at(ii)) {
                if (this->optimizeDf().at(ii)) {
                    this->mVectIterate[idOptimizeAndUnknownDf].setValuesAtMemNormal(
                            this->vectUnknownDfsDomOld()[idDf]);
                     ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }
        // REMARK: Synchronization of mVectIterate is not necessary
        // because this data field only lives on memNormal!
    }

    this->iterateOverTimeInternal();

    /*------------------------------------------------------------------------*/
    // Copy resp. convert gradient of target functional.
    if (this->isParamsBased()) {
        for (int ll = 0; ll < this->nParamsToOptimize(); ++ll) {
            this->mMemoryVectIterateGradTf[ll] = this->mMemoryVectGradTf[ll];
        }
    }
    else {
        for (std::size_t ii = 0; ii < this->mVectGradTf.size(); ++ii) {
            this->mVectGradTf[ii].sync(this->nTimesteps());
        }

        // Both objects have the same vector lengths.
        for (std::size_t ii = 0; ii < this->mVectGradTf.size(); ++ii) {
            this->mVectIterateGradTf[ii].setValuesAtMemNormal(mVectGradTf[ii]);
            // mVectIterateGradTf need not to be synchronized as it lives
            // on memNormal.
        }
    }

    /*------------------------------------------------------------------------*/
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::iterateOverTimeInternal()
{
    int currTimestep = 0;

    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Iterate over time..." << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
#ifdef VTRACE
    unsigned int mid;
    mid = VT_MARKER_DEF("Prblm_iterate", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Prblm_iterate_start");
#endif

    this->mTf = static_cast<CT>(0);
    for (std::size_t ii = 0; ii < this->mMemoryVectGradTf.size(); ++ii) {
        this->mMemoryVectGradTf[ii] = static_cast<CT>(0);
    }
    for (currTimestep = 0; currTimestep <= this->nTimesteps(); ++currTimestep) {

        /*--------------------------------------------------------------------*/
        if ((this->params().rankOutput() == this->myRank()) 
            && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank())
                      << " * Process TIME STEP=" << currTimestep
                      << "..." << std::endl;
        }
        this->mParams.decreaseLevel();

        /*--------------------------------------------------------------------*/
        this->evalKnownDfs(currTimestep);
        if (0 < currTimestep) {
            this->updateValsAtNodesOfUnknownDfs(currTimestep);
            if (this->useLeapfrogIntegration()) {
                // Swap old / new data fields, BUT DO NOT SWAP times.
                // ==> Swap times 2x.
                this->swapPointersOfUnknownDfs(); // 1st swap of times.
                this->swapTimesOfUnknownDfs();    // 2nd swap of times.
                this->update2ValsAtNodesOfUnknownDfs(currTimestep);
            }
            // Compute target functional at each time step.
            if (this->params().isTfTimeDependent()) {
                this->compTargetfunc(currTimestep);
            }
            else {
                if (currTimestep == this->params().nTimesteps()) {
                    this->compTargetfunc(currTimestep);
                }
            }
        }
        else {
            this->initUnknownDfs();
        }
        this->writeAllDfs(currTimestep);
        this->compErrOfUnknownDfs();
        this->swapPointersOfUnknownDfs();
        // Update times of old / new data fields AFTER swapping.
        this->updateTimesOfUnknownDfs();

        /*--------------------------------------------------------------------*/
        this->mParams.increaseLevel();
        if ((this->params().rankOutput() == this->myRank()) 
            && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank())
                      << "   Processed." << std::endl;
        }
    }
#ifdef VTRACE
    mid = VT_MARKER_DEF("EndLoop", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Prblm_iterate_end");
#endif

    /*------------------------------------------------------------------------*/
    // Print some statistics.
    double tSum = 0.0;
    for (auto it = this->mWallClockTimes.begin(); 
         it != this->mWallClockTimes.end(); ++it) {
        tSum += it->second;
    }
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << std::scientific << std::fixed
                  << " * Summary of Wall clock times(compNewTfAndGradTf):"
                  << std::endl;
        for (auto it = this->mWallClockTimes.begin(); 
             it != this->mWallClockTimes.end(); ++it) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      <<  "       t(" << std::setw(11) << it->first << ") = "
                      << std::setw(8) << std::setprecision(5)
                  << it->second
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << it->second / tSum * 100.0 << " %%)"
                  << std::endl;
        }
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "       --------------------------------------"
                  << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "       t(compTfGrdTf) = "
                  << std::setw(8) << std::setprecision(5)
                  << tSum
                  << " s."
                  << resetiosflags( std::ios::fixed | std::ios::showpoint )
                  << std::endl;
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Iterated."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::minimizeTargetfct(
    std::string const& optimizer
)
{
    if (0 == optimizer.compare("lbfgsb")) {
        int optimVersion = 3; // Main version of optimizer L-BFGS-B.
        minimizeTargetfctUsingLbfgsb(optimVersion);
    }
    else {
        if (this->params().rankOutput() == this->myRank()
            && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                      << " * Optimizer " << optimizer << " not implemented."
                      << " Return." << std::endl;
        }
        return;
    }
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::isTracedTf() const
{
    return this->mIsTracedCompTf;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& 
Problem<OWNPRBLM, CT, DIM>::tapeIdComputeTfPartitionInner() const
{
    return this->mTapeIdComputeTfPartitionInner;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& 
Problem<OWNPRBLM, CT, DIM>::tapeIdComputeTfPartitionBorder() const
{
    return this->mTapeIdComputeTfPartitionBorder;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int 
const& Problem<OWNPRBLM, CT, DIM>::tapeIdComputeTfPartitionAll() const
{
    return this->mTapeIdComputeTfPartitionAll;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nParamsToOptimize() const
{
    return this->mNparamsToOptimize;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<bool> const& Problem<OWNPRBLM, CT, DIM>::optimizeDf() const
{
    return this->mOptimizeDf;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::minimizeTargetfctUsingLbfgsb(
    int const& optimVersion
)
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
         std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                this->myRank(), "SCADOPT")
                   << " * Minimize target functional using L-BFGS-B..."
                   << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    // Optimizer (MPI LOCAL): Declarations and initializations.
    boost::mpi::timer timerMinimize;
    double tMinimize = 0.0;
    boost::mpi::timer timerCommunicate;
    boost::mpi::timer timerLbfgsb;
    timerMinimize.restart();
    int kk = 0;
    int rankOptim = 0;

    /*------------------------------------------------------------------------*/
    // L-BFGS-B (MPI LOCAL): Declare quantities of L-BFGS-B
    // Declare variables for ALL processes such that LBFGS-B related variables
    // are known to all processes.
    // This must be done for ALL processes since there are several
    // "if (rank == rankOptim)" sections.
    // ==> Variables cannot be declared in the first local if section.
    // Instead, LBFGS-B related variables must be known to all processes.
    double scalarTfGlobal = 0.0;
    std::vector<double> memoryIterateGlobal;
    std::vector<double> memoryGradTfGlobal;
    int nCorrectionsInMatrix = 0;
    std::vector<double> lowerBound;
    std::vector<double> upperBound;
    std::vector<int> typeOfBounds;
    double factr = 0.0;
    double pgtol = 0.0;
    int info = 0;
    std::vector<double> wa;
    std::vector<int> iwa;
    //char* task = 0;
    std::vector<char> csave;
    std::vector<int> lsave;
    std::vector<int> isave;
    std::vector<double> dsave;
    const int N_CHARS_TASK = 60;
    //task = new char[N_CHARS_TASK];
    char task[N_CHARS_TASK];
    for (int ii = 0; ii < N_CHARS_TASK; ++ii) {
        task[ii] = static_cast<char>(0);
    }
    task[59] = '\0';
    
    task[0] = 'S';
    task[1] = 'T';
    task[2] = 'A';
    task[3] = 'R';
    task[4] = 'T';

    /*------------------------------------------------------------------------*/
    // nParamsToOptimizeGlobal = Number of geometrical parameters. e.g.
    //
    // It is assumed that all data fields have the same number of nodes.
    // As all data fields are scalar and all (global) data fields live on
    // the global grid, this assumption is true in the moment.
    // ==> nParamsToOptimizeGlobal = Global number of grid nodes multiplied
    // by the number of those data fields which should be opimized.
    int nParamsToOptimizeGlobal = 0;
    if (this->isParamsBased()) {
        nParamsToOptimizeGlobal = this->vectParamsCurr().size();
    }
    else {
        kk = 0;
        int tmpNvars = 0;
        for (std::size_t jj = 0; jj < this->isKnownDf().size(); ++jj) {
            if (!this->isKnownDf().at(jj)) {
                tmpNvars = this->vectUnknownDfsDomNew().at(kk).gridGlobal().nNodes().size();
                ++kk;
            }
            if (this->optimizeDf().at(jj)) {
                nParamsToOptimizeGlobal += tmpNvars;
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // L-BFGS-B (MPI GLOBAL):
    // Initialize quantities of L-BFGS-B depending on the number of unknowns
    // on process "rankOptim".
    if (rankOptim == this->myRank()) {
        nCorrectionsInMatrix = 5;
        memoryIterateGlobal.resize(nParamsToOptimizeGlobal, 0.0);
        memoryGradTfGlobal.resize(nParamsToOptimizeGlobal, 0.0);
        lowerBound.resize(nParamsToOptimizeGlobal, 0.0);
        upperBound.resize(nParamsToOptimizeGlobal, 0.0);
        // All variables are unbounded.
        typeOfBounds.resize(nParamsToOptimizeGlobal, 0);

        factr = 1.0e1;
        pgtol = 1.0e-12; // Criterium will be ignored.
        const int nmax = nParamsToOptimizeGlobal;
        const int mmax = nCorrectionsInMatrix;
        info = 0;

        int nWa = 0;
        if (2 < optimVersion) {
            // L-BFGS-B 3.0
            nWa = (2 * mmax + 5) * nmax + 11 * mmax * mmax + 8 * mmax;
        } 
        else {
            // L-BFGS-B 2.1
            nWa = (2 * mmax * nmax + 4 * nmax + 12 * mmax * mmax + 12 * mmax);
        }
        wa.resize(nWa, 0.0);
        iwa.resize(3 * nmax, 0);
        csave.resize(60, static_cast<char>(0));
        lsave.resize(4, 0);
        isave.resize(44, 0);
        dsave.resize(29, 0.0);
    }

    /*------------------------------------------------------------------------*/
    // (MPI LOCAL): Iterate.
    this->compNewTfAndGradTf();
    this->worldBarrier();

    /*------------------------------------------------------------------------*/
    // MPI: Prepare collecting and distributing.
    int nDfsToOptimize = this->getNdataFieldsForOptimization();
    int* nParamsToOptimizeLocalOfDf = 0;
    if (this->isParamsBased()) {
    }
    else {
        nParamsToOptimizeLocalOfDf = new int[nDfsToOptimize];
        for (int jj = 0; jj < nDfsToOptimize; ++jj) {
            nParamsToOptimizeLocalOfDf[jj] = 0;
        }
        int idDf = 0;
        int idOptimizeAndUnknownDf = 0;
        int tmpNvars = 0;
        for (std::size_t jj = 0; jj < this->isKnownDf().size(); ++jj) {
            if (!this->isKnownDf().at(jj)) {
                tmpNvars = this->vectUnknownDfsDomNew().at(idDf).memNormal().nNodesSub().size();
                if (this->optimizeDf().at(jj)) {
                    nParamsToOptimizeLocalOfDf[idOptimizeAndUnknownDf] = tmpNvars;
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // MPI: Prepare collecting and distributing.
    // Values: nDfsToOptimize approx 1 - 20
    //         nProcesses approx 100 - 10.000
    int** displs = 0;
    int** rcounts = 0;
    int nProcesses = this->params().myWorld().size();
    displs = new int*[nDfsToOptimize];
    rcounts = new int*[nDfsToOptimize];
    for (int jj = 0; jj < nDfsToOptimize; ++jj) {
        displs[jj] = new int[nProcesses];
        rcounts[jj] = new int[nProcesses];
    }

    for (int jj = 0; jj < nDfsToOptimize; ++jj) {
        // Collect portions of local arrays and send it to the
        // MPI process for optimization "rankOptim".
        // Exchange number of nodes of each process (=grid partition).
        // Case: Current rank != rankOptim
        // ---> Send local portion of array to rankOptim.
        //if (rankOptim != this->myRank()) {
            //this->params().myWorld().send(rankOptim, 0,
            //                               nParamsToOptimizeLocalOfDf[jj]);
        //}
        int nUnknownsLocalCurr;
        // Case: Current rank == rankOptim
        // ---> Receive local portion of array from all other processes.
        //if (rankOptim == this->myRank()) {
            for (int ii = 0; ii < nProcesses; ++ii) {
               // if (rankOptim != ii) {
                 nUnknownsLocalCurr = nParamsToOptimizeLocalOfDf[jj];
                 boost::mpi::broadcast(this->params().myWorld(),
                                       nUnknownsLocalCurr, ii);
                 //this->params().myWorld().recv(ii, 0, nUnknownsLocalCurr);
                    rcounts[jj][ii] = nUnknownsLocalCurr;
              //  }
              //  else {
              //      rcounts[jj][ii] = nParamsToOptimizeLocalOfDf[jj];
              //  }
            }
        //}
        this->worldBarrier();
    }

    int nCurrDf = 0;
    int nPreviousDfs = 0;
    for (int jj = 0; jj < nDfsToOptimize; ++jj) {
            for (int ii = 0; ii < nProcesses; ++ii) {
                nPreviousDfs = 0;
                for (int ll = 0; ll < jj; ++ll) {
                    for (int kk = 0; kk < nProcesses; ++kk) {
                        nPreviousDfs += rcounts[ll][kk];
                    }
                }
                nCurrDf = 0;
                for (int kk = 0; kk < ii; ++kk) {
                    nCurrDf += rcounts[jj][kk];
                }
                displs[jj][ii] = nPreviousDfs + nCurrDf;
            }
    }
    this->worldBarrier();

    /*------------------------------------------------------------------------*/
    if (this->params().rankOutput() == this->myRank()
            && (0 < this->params().indentDepth())) {
                std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                          << " * (Local) Value of target functional: "
                          << this->mTf
                          << std::endl;
    }

    /*------------------------------------------------------------------------*/
    // Collect local values of target functional to global value.
    timerCommunicate.restart();
    this->collectScalar(scalarTfGlobal,
                  static_cast<double>(this->mTf),
                  rankOptim);

    /*------------------------------------------------------------------------*/
    // Collect local arrays in corresponding global ones.
    if (this->isParamsBased()) {
            if (rankOptim == this->myRank()) {
                for (int ii = 0; ii < this->nColumns(); ++ii) {
                    memoryIterateGlobal[ii] = 0.0;
                }
            }
            // REMARK: Initial guess is already the same on all processes.
            // => No mpi_reduce() necessary.
            if (rankOptim == this->myRank()) {
                for (int ii = 0; ii < this->nColumns(); ++ii) {
                    memoryIterateGlobal[ii] = this->mMemoryVectIterate[ii];
                }
            }
            if (rankOptim == this->myRank()) {
                for (int ii = 0; ii < this->nColumns(); ++ii) {
                    memoryGradTfGlobal[ii] = 0.0;
                }
            }
            for (int ii = 0; ii < this->nColumns(); ++ii) {
                boost::mpi::reduce(
                    this->params().myWorld(),
                    this->mMemoryVectGradTf[ii],
                    memoryGradTfGlobal[ii],
                    std::plus<CT>(),
                    rankOptim
                );
            }
    }
    else {
        this->collectArray1D(memoryIterateGlobal,
                       this->mMemoryVectIterate,
                       rcounts,
                       displs,
                       nParamsToOptimizeLocalOfDf,
                       nParamsToOptimize(),
                       nDfsToOptimize,
                       rankOptim);
        this->collectArray1D(memoryGradTfGlobal,
                       this->mMemoryVectIterateGradTf,
                       rcounts,
                       displs,
                       nParamsToOptimizeLocalOfDf,
                       nParamsToOptimize(),
                       nDfsToOptimize,
                       rankOptim);
    }
    this->mWallClockTimesOptim["communic"] += timerCommunicate.elapsed();

    /*------------------------------------------------------------------------*/
    // Optimizer (MPI LOCAL): Loop.
    bool iterateOptimLoop = true;
    int currIteration = 0;
    do {

        if (this->params().rankOutput() == this->myRank()
            && (0 < this->params().indentDepth())) {
                std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                          << " * Process OPTIM_ITERATION="
                          << currIteration
                          << std::endl;
        }
        this->mParams.decreaseLevel();

        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank()
            && (0 < this->params().indentDepth())) {
                std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                          << " * (Global) parameter vector:" << std::endl;
            for (int ii = 0; ii < this->nColumns(); ++ii) {
                std::cout << " p[ " << ii << " ] = "
                           <<  memoryIterateGlobal.at(ii) << std::endl;
            }
        }
        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank()
            && (0 < this->params().indentDepth())) {
                std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                          << " * (Global) Value of target functional: "
                          << scalarTfGlobal
                          << std::endl;
        }

        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank()
            && (0 < this->params().indentDepth())) {
                std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                          << " * (Global) Value of gradient of tf: "
                          << std::endl;
            for (int ii = 0; ii < this->nColumns(); ++ii) {
                std::cout << " gradTf[ " << ii << " ] = "
                           <<  memoryGradTfGlobal.at(ii) << std::endl;
            }
        }

        std::string strtask("");
        strtask.resize(N_CHARS_TASK);
#ifdef SCADOPT_HAVE_LBFGSB
        /*--------------------------------------------------------------------*/
        // L-BFGS-B (MPI GLOBAL): Call FORTRAN routine.
        //void setulb_(int*, int*, double*, double*, double*, int*, double*,
        //             double*, double*, double*, double*, int*, char*, int*,
        //             char*, bool*, int*, double*)
        if (this->params().rankOutput() == this->myRank()
            && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                      << " * Call L-BFGS-B routine setulb()..." << std::endl;
        }
        this->mParams.decreaseLevel();
        timerLbfgsb.restart();

        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank()) {
            double* ptr2memoryIterateGlobal = memoryIterateGlobal.data();
            double* ptr2lowerBound = lowerBound.data();
            double* ptr2upperBound = upperBound.data();
            int* ptr2typeOfBounds = typeOfBounds.data();
            double* ptr2memoryGradTfGlobal = memoryGradTfGlobal.data();
            double* ptr2wa = wa.data();
            int* ptr2iwa = iwa.data();
            char* ptr2csave = csave.data();
            int* ptr2lsave = lsave.data();
            int* ptr2isave = isave.data();
            double* ptr2dsave = dsave.data();
        
            setulb_(
                &nParamsToOptimizeGlobal,
                &nCorrectionsInMatrix,
                ptr2memoryIterateGlobal,
                ptr2lowerBound,
                ptr2upperBound,
                ptr2typeOfBounds,
                &scalarTfGlobal,
                ptr2memoryGradTfGlobal,
                &factr,
                &pgtol,
                ptr2wa,
                ptr2iwa,
                task,
                &info,
                ptr2csave,
                ptr2lsave,
                ptr2isave,
                ptr2dsave
            );
            for (int ii = 0; ii < N_CHARS_TASK; ++ii) {
                strtask[ii] = task[ii];
            }
        }

        // Broadcast task to all other processes.
        this->params().myWorld().barrier();
        MPI_Bcast(task, N_CHARS_TASK+1, MPI_CHAR, rankOptim,
                   this->params().myWorld());
        // Alternative 1: 
        // MPI_Bcast(task, strlen(task) + 1, MPI_CHAR, rankOptim,
        //           this->params().myWorld());
        //
        // Alternative 2: 
        // boost::mpi::broadcast(this->params().myWorld(),
        //                      strtask,
        //                    rankOptim);

        // Set task variable to string variant strtask on all processes. 
        strtask = task;

        this->mWallClockTimesOptim["lbfgsb"] += timerLbfgsb.elapsed();
        this->mParams.increaseLevel();
        if (this->params().rankOutput() == this->myRank()
            && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "     * New task: " << strtask 
                      << std::endl
                      << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "   Called." << std::endl;
        }
#endif

        /*--------------------------------------------------------------------*/
        // L-BFGS-B (MPI LOCAL):
        // Case: Function and its gradient must be calculated
        // at current iterate.
        //if (0 == strncmp(task, "FG", 2)) {
        if (0 == strncmp(strtask.c_str(), "FG", 2)) {
            // Distribute global iterate to local arrays.
            timerCommunicate.restart();
            if (this->isParamsBased()) {
                if (rankOptim == this->myRank()) {
                    // Copy the parameter based vector got from the setulb().
                    for (int ll = 0; ll < this->nParamsToOptimize(); ++ll) {
                        this->mMemoryVectIterate[ll] = memoryIterateGlobal[ll];
                    }
                    for (int ll = 0; ll < this->nParamsToOptimize(); ++ll) {
                        this->mVectParamsCurr[ll] = this->mMemoryVectIterate[ll];
                    }
                    if (1 < this->params().myWorld().size()) {
                        // The vector of the geometrical parameters usually 
                        // consists only of a small number (1-50) of elements.
                        // --> Just broadcast the vector to all MPI processes.
                        // We use mVectParamsCurr instead of mMemoryVectIterate
                        // because boost::mpi can handle
                        // variables of type std::vector.
                        boost::mpi::broadcast(
                            this->params().myWorld(),
                            this->mVectParamsCurr,
                            rankOptim
                        );
                    }
                }
            }
            else {
                this->distributeArray1D(this->mMemoryVectIterate,
                                  memoryIterateGlobal,
                                  rcounts,
                                  displs,
                                  nParamsToOptimizeLocalOfDf,
                                  this->nParamsToOptimize(),
                                  nDfsToOptimize,
                                  rankOptim);
                 // mVectIterate works on nodes of grid memNormal!
                int idDf = 0;
                int idOptimizeAndUnknownDf = 0;
                for (std::size_t ii = 0; ii < this->optimizeDf().size(); ++ii) {
                    if (!this->isKnownDf().at(ii)) {
                        if (this->optimizeDf().at(ii)) {
                            this->mVectUnknownDfsDomOld[idDf].setValuesAtMemNormal(
                                this->mVectIterate[idOptimizeAndUnknownDf]);
                            ++idOptimizeAndUnknownDf;
                        }
                        ++idDf;
                    }
                }
            }
            this->mWallClockTimesOptim["communic"] += timerCommunicate.elapsed();

            /*----------------------------------------------------------------*/
            // mVectUnknownDfsDomOld has to contain the values of the 
            // current iterate.
            this->compNewTfAndGradTf();

            /*----------------------------------------------------------------*/
            // Collect local scalar and arrays in corresponding global ones.
            timerCommunicate.restart();
            this->collectScalar(
                scalarTfGlobal,
                this->mTf,
                rankOptim
            );
            if (this->isParamsBased()) {
                // REMARK: Initial guess is already the same on all processes.
                // => No mpi_reduce() necessary.
                if (rankOptim == this->myRank()) {
                    for (int ii = 0; ii < this->nColumns(); ++ii) {
                        memoryIterateGlobal[ii] = this->mMemoryVectIterate[ii];
                    }
                }
//                     for (int ii = 0; ii < nColumns(); ++ii) {
//                         boost::mpi::reduce(
//                             this->params().myWorld(),
//                             mMemoryVectIterate[ii],
//                             memoryIterateGlobal[ii],
//                             std::plus<CT>(),
//                             rankOptim
//                         );
//                     }
                    for (int ii = 0; ii < this->nColumns(); ++ii) {
                        boost::mpi::reduce(
                            this->params().myWorld(),
                            this->mMemoryVectGradTf[ii],
                            memoryGradTfGlobal[ii],
                            std::plus<CT>(),
                            rankOptim
                        );
                    }
            }
            else {
                this->collectArray1D(
                    memoryIterateGlobal,
                    this->mMemoryVectIterate,
                    rcounts,
                    displs,
                    nParamsToOptimizeLocalOfDf,
                    this->nParamsToOptimize(),
                    nDfsToOptimize,
                    rankOptim
                );
                this->collectArray1D(
                    memoryGradTfGlobal,
                    this->mMemoryVectIterateGradTf,
                    rcounts,
                    displs,
                    nParamsToOptimizeLocalOfDf,
                    this->nParamsToOptimize(),
                    nDfsToOptimize,
                    rankOptim
                );
            }
            this->mWallClockTimesOptim["communic"] 
                += timerCommunicate.elapsed();
        }
        /*--------------------------------------------------------------------*/
        // Case: New iterate was computed, the user can decide:
        // Stop or continue the iteration.
        else if (0 == strncmp(strtask.c_str(), "NEW_X", 5)) {
            // Currently, the iteration will be continued.
        }
        /*--------------------------------------------------------------------*/
        // Case: Iteration will be terminated.
        else {
            iterateOptimLoop = false;
        }

        this->mParams.increaseLevel();
        if (this->params().rankOutput() == this->myRank()
            && (0 < this->params().indentDepth())) {
                std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                          << "   Processed." << std::endl;
        }
        ++currIteration;
    } while (iterateOptimLoop &&  currIteration<4);

    /*------------------------------------------------------------------------*/
    // Print some statistics.
    double tSum = 0.0;
    for (auto it = this->mWallClockTimes.begin(); 
         it != this->mWallClockTimes.end(); ++it) {
        tSum += it->second;
    }
    for (auto it = this->mWallClockTimesOptim.begin(); 
         it != this->mWallClockTimesOptim.end(); ++it) {
        tSum+= it->second;
    }
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(), 
                                       this->myRank(), "SCADOPT")
                  << std::scientific << std::fixed
                  << " * Summary of Wall clock times(optimize):"
                  << std::endl;
        for (auto it = this->mWallClockTimes.begin(); 
             it != this->mWallClockTimes.end(); ++it) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(), 
                                           this->myRank(), "SCADOPT")
                      <<  "       t(" << std::setw(11) << it->first << ") = "
                      << std::setw(8) << std::setprecision(5)
                  << it->second
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << it->second / tSum * 100.0 << " %%)"
                  << std::endl;
        }
        for (auto it = this->mWallClockTimesOptim.begin(); 
             it != this->mWallClockTimesOptim.end(); ++it) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(), 
                                           this->myRank(), "SCADOPT")
                      <<  "       t(" << std::setw(11) << it->first << ") = "
                      << std::setw(8) << std::setprecision(5)
                  << it->second
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << it->second / tSum * 100.0 << " %%)"
                  << std::endl;
        }
        tMinimize = timerMinimize.elapsed();
        std::cout << ScaFES::getPrefix(this->params().indentDepth(), 
                                       this->myRank(), "SCADOPT")
                  << "       t(   sumItems) = "
                  << std::setw(8) << std::setprecision(5)
                  << tSum
                  << " s."
                  << resetiosflags( std::ios::fixed | std::ios::showpoint )
                  << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(), 
                                       this->myRank(), "SCADOPT")
                  << " ----> t(   optimize) = "
                  << std::setw(8) << std::setprecision(5)
                  << tMinimize
                  << " s. <-------"
                  << resetiosflags( std::ios::fixed | std::ios::showpoint )
                  << std::endl;

    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(), 
                                       this->myRank(), "SCADOPT")
                  << "   Minimized."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::collectScalar(
    CT& scalarGlobal,
    CT const& scalarLocal,
    int const& rankOptim
)
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                      << " * Collect scalar from all grid partitions..."
                      << std::endl;
    }
    this->mParams.decreaseLevel();

    const int N_COMPS = 1;
    double sendbuff[N_COMPS];
    double recvbuff[N_COMPS];

    if (1 < this->params().myWorld().size()) {
        // Alternative:
// // //         boost::mpi::reduce(
// // //             this->params().myWorld(),
// // //             scalarLocal,
// // //             scalarGlobal,
// // //             std::plus<CT>(),
// // //             rankOptim
// // //         );
        sendbuff[0] = scalarLocal;
        MPI_Reduce(sendbuff,
                   recvbuff,
                   N_COMPS,
                   MPI_DOUBLE,
                   MPI_SUM,
                   rankOptim,
                   this->params().myWorld());
        scalarGlobal = recvbuff[0];
        this->worldBarrier();
    }
    else {
        scalarGlobal = scalarLocal;
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                      << "   Collected."
                      << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::collectArray1D(
    std::vector<CT>& arrayGlobal,
    std::vector<CT>& arrayLocal,
    int** const& rcounts,
    int** const& displs,
    int* const& nVectLocal,
    int const& nTotalLocal,
    int const& nDfsToOptimize,
    int const& rankOptim
)
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                   this->myRank(), "SCADOPT")
                      << " * Collect 1D array from all grid partitions..."
                      << std::endl;
    }
    this->mParams.decreaseLevel();

    if (1 < this->params().myWorld().size()) {
        CT* sendbufferStart = arrayLocal.data();
        for (int jj = 0; jj < nDfsToOptimize; ++jj) {
            MPI_Gatherv(sendbufferStart,
                        nVectLocal[jj],
                        MPI_DOUBLE,
                        arrayGlobal.data(),
                        rcounts[jj],
                        displs[jj],
                        MPI_DOUBLE,
                        rankOptim,
                        this->params().myWorld());
            this->worldBarrier();
            sendbufferStart += nVectLocal[jj];
        }
    }
    else {
        for (int ii = 0; ii < nTotalLocal; ++ii) {
            arrayGlobal[ii] = arrayLocal[ii];
        }
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank(), "SCADOPT")
                  << "   Collected."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::distributeArray1D(
    std::vector<CT>& arrayLocal,
    std::vector<CT>& arrayGlobal,
    int** const& rcounts,
    int** const& displs,
    int* const& nVectLocal,
    int const& nTotalLocal,
    int const& nDfsToOptimize,
    int const& rankOptim
)
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Distribute 1D array to all grid partitions..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    if (1 < this->params().myWorld().size()) {
        // Distribute arrays from process for optimization to
        // all MPI processes.
        // Case: Current rank == rankOptim
        // ---> Send local array portion to all other processes.
        // Case: Current rank is NOT rankOptim.
        // ---> Receive local array from rankOptim.
        CT* ptrToArrayGlobal = arrayGlobal.data();
        CT* recvBufferStart = arrayLocal.data();
        for (int jj = 0; jj < nDfsToOptimize; ++jj) {
            MPI_Scatterv(ptrToArrayGlobal,
                         rcounts[jj],
                         displs[jj],
                         MPI_DOUBLE,
                         recvBufferStart,
                         nVectLocal[jj],
                         MPI_DOUBLE,
                         rankOptim,
                         this->params().myWorld());
            this->worldBarrier();
            recvBufferStart += nVectLocal[jj];
        }
    }
    else {
        for (int ii = 0; ii < nTotalLocal; ++ii) {
            arrayLocal[ii] = arrayGlobal[ii];
        }
    }
    this->worldBarrier();

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                   << "   Distributed."
                   << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::getNdataFieldsForOptimization() const
{
    int nDataFieldsForOptimization = 0;
    for (std::size_t ii = 0; ii < this->optimizeDf().size(); ++ii) {
        if (this->optimizeDf().at(ii)) {
            ++nDataFieldsForOptimization;
        }
    }

    return nDataFieldsForOptimization;
}


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::evalInner(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template evalInner<TT>(
        dfDep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::evalBorder(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template evalBorder<TT>(
        dfDep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initInner(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector<TT> const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template initInner<TT>(
        dfDep,
        dfIndep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initBorder(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector<TT> const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template initBorder<TT>(
        dfDep,
        dfIndep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initInner(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template initInner<TT>(
        dfDep,
        dfIndep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initBorder(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template initBorder<TT>(
        dfDep,
        dfIndep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateInner(
    std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
    std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template updateInner<TT>(
        dfDep,
        dfIndep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateBorder(
    std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
    std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template updateBorder<TT>(
        dfDep,
        dfIndep,
        idxNode,
        timestep
    );
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateInner2(
    std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
    std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template
        updateInner2<TT>(dfDep,dfIndep,idxNode,timestep);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateBorder2(
    std::vector< ScaFES::DataField<TT, DIM> >& dfDep,
    std::vector< ScaFES::DataField<TT, DIM> > const& dfIndep,
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& timestep
)
{
    static_cast<OWNPRBLM*>(this)->template
        updateBorder2<TT>(dfDep,dfIndep,idxNode,timestep);
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::compTargetfunc(int const& timestep)
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute target functional..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    CT tfTmp = static_cast<CT>(0);
    std::vector<CT> tfGradTmp(this->mMemoryVectGradTf.size(),
                              static_cast<CT>(0));
    
    if (this->useAsynchronMode()) {
        this->computeTfAsynchronously(
            &SCADOPT::Problem<OWNPRBLM,CT,DIM>::wrapperComputeTfPartitionInner,
            &SCADOPT::Problem<OWNPRBLM,CT,DIM>::wrapperComputeTfPartitionBorder,
            tfTmp,
            tfGradTmp,
            this->vectUnknownDfsDomNew(),
            this->mVectGradUnknownDfsDomNew,
            timestep
        );
    }
    else {
        this->computeTfSynchronously(
            &SCADOPT::Problem<OWNPRBLM, CT, DIM>::wrapperComputeTfPartitionAll,
            tfTmp,
            tfGradTmp,
            this->vectUnknownDfsDomNew(),
            this->mVectGradUnknownDfsDomNew,
            this->vectUnknownDfsBdryNew(),
            this->mVectGradUnknownDfsBdryNew,
            timestep
        );
    }

    if (this->params().isTfTimeDependent()) {
        this->mTf += tfTmp;
        for (std::size_t ii = 0; ii < tfGradTmp.size(); ++ii) {
            this->mMemoryVectGradTf[ii] += tfGradTmp[ii];
        }
    }
    else {
        this->mTf = tfTmp;
        for (std::size_t ii = 0; ii < tfGradTmp.size(); ++ii) {
            this->mMemoryVectGradTf[ii] = tfGradTmp[ii];
        }
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                   << "   Computed."
                   << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::computeTfAsynchronously(
    TypeFuncComputeTfToEval funcLocalInner,
    TypeFuncComputeTfToEval funcLocalBorder,
    CT& dfDep,
    std::vector<CT>& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    boost::mpi::timer timerSync;

    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Compute tf asynchronously at "
                  << dfIndep.at(0).name() << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    /*------------------------------------------------------------------------*/
    dfDep = static_cast<CT>(0);
    (this->*funcLocalBorder)(
        dfDep, 
        dfGradDep, 
        dfIndep, 
        dfGradIndep, 
        timestep
    );
    CT dfTmp = static_cast<CT>(0);
    std::vector<CT> dfGradTmp(dfGradDep.size());
    (this->*funcLocalInner)(
        dfTmp, 
        dfGradTmp, 
        dfIndep, 
        dfGradIndep, 
        timestep
    );
    dfDep += dfTmp;
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii] += dfGradTmp[ii];
    }

    /*------------------------------------------------------------------------*/
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   * (Local) tf value = " << dfDep << std::endl;
    }
    /*------------------------------------------------------------------------*/
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::computeTfSynchronously(
    TypeFuncComputeTfToEval funcGlobal,
    CT& dfDep,
    std::vector<CT>& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepDomain,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndepDomain,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepBdry,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndepBdry,
    const int& timestep
)
{
    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute tf synchronously at "
                  << dfIndepDomain.at(0).name() << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    /*------------------------------------------------------------------------*/
    // Compute tf for ALL (=DOMAIN+BOUNDARY) data fields
    // at all GLOBAL BORDER nodes.
    dfDep = static_cast<CT>(0);
    CT dfTmp = static_cast<CT>(0);
    std::vector<CT> dfGradTmp(dfGradDep.size(),
                              static_cast<CT>(0));
    std::vector< ScaFES::DataField<CT,DIM> > vectAllTmp(dfIndepDomain);
    for (std::size_t ii = 0; ii < dfIndepBdry.size(); ++ii) {
        vectAllTmp.push_back(dfIndepBdry[ii]);
    }
    std::vector< ScaFES::DataField<CT,DIM> > vectGradAllTmp(dfGradIndepDomain);
    for (std::size_t ii = 0; ii < dfGradIndepBdry.size(); ++ii) {
        vectGradAllTmp.push_back(dfGradIndepBdry[ii]);
    }
    (this->*funcGlobal)(
        dfTmp, 
        dfGradTmp, 
        vectAllTmp, 
        vectGradAllTmp, 
        timestep
    );
    dfDep += dfTmp;
    for (std::size_t ii = 0; ii < dfGradTmp.size(); ++ii) {
        dfGradDep[ii] += dfGradTmp[ii];
    }
    /*------------------------------------------------------------------------*/
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   * Local value(tf) = " << dfDep << std::endl;
    }
    /*------------------------------------------------------------------------*/
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperComputeTfPartitionAll(
    CT& dfDep,
    std::vector<CT>& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        this->traceComputeTf(
            dfDep,
            dfIndep,
            timestep
        );
        if (this->computeGradients()) {
            this->computeTf_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdComputeTfPartitionAll(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->computeTf_zosForward(
                dfDep,
                this->tapeIdComputeTfPartitionAll(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerTf;
        timerTf.restart();
        CT tmpDfDep = static_cast<CT>(0);
        this->computeTfGlobalInner<CT>(tmpDfDep, dfIndep, timestep);
        this->computeTfGlobalBorder<CT>(dfDep, dfIndep, timestep);
        dfDep += tmpDfDep;
        this->mWallClockTimesOptim["compTf"] += timerTf.elapsed();
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperComputeTfPartitionBorder(
    CT& dfDep,
    std::vector<CT>& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        this->traceComputeTf(
            dfDep,
            dfIndep,
            timestep
        );
        if (this->computeGradients()) {
            this->computeTf_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdComputeTfPartitionBorder(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->computeTf_zosForward(
                dfDep,
                this->tapeIdComputeTfPartitionBorder(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerTf;
        timerTf.restart();
        this->computeTfLocalBorder<CT>(dfDep, dfIndep, timestep);
        this->mWallClockTimesOptim["compTf"] += timerTf.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperComputeTfPartitionInner(
    CT& dfDep,
    std::vector<CT>& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        this->traceComputeTf(
            dfDep,
            dfIndep,
            timestep
        );
        if (this->computeGradients()) {
            this->computeTf_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdComputeTfPartitionInner(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->computeTf_zosForward(
                dfDep,
                this->tapeIdComputeTfPartitionInner(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerTf;
        timerTf.restart();
        this->computeTfLocalInner<CT>(dfDep, dfIndep, timestep);
        this->mWallClockTimesOptim["compTf"] += timerTf.elapsed();
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::computeTfLocalInner(
    TT& dfDep,
    std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute (normally) tf at all local inner nodes at "
                  << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    int comp = 0;
    std::size_t ii;
    ScaFES::Ntuple<int,DIM> idxNode;
    dfDep = static_cast<TT>(0);
    TT dfDepTmp = static_cast<TT>(0);
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfBorder(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfInner(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::computeTfLocalBorder(
    TT& dfDep,
    std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute (normally) tf at all local border nodes at "
                  << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    int comp = 0;
    std::size_t ii;
    ScaFES::Ntuple<int,DIM> idxNode;
    dfDep = static_cast<TT>(0);
    TT dfDepTmp = static_cast<TT>(0);
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfBorder(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfInner(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::computeTfGlobalBorder(
    TT& dfDep,
    std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute (normally) tf at all global border nodes at "
                  << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    int comp = 0;
    std::size_t ii;
    ScaFES::Ntuple<int,DIM> idxNode;
    dfDep = static_cast<TT>(0);
    TT dfDepTmp = static_cast<TT>(0);
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfBorder(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfBorder(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::computeTfGlobalInner(
    TT& dfDep,
    std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute (normally) tf at all global inner nodes at "
                  << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    int comp = 0;
    std::size_t ii;
    ScaFES::Ntuple<int,DIM> idxNode;
    dfDep = static_cast<TT>(0);
    TT dfDepTmp = static_cast<TT>(0);
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfInner(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        dfDepTmp = static_cast<TT>(0);
        static_cast<OWNPRBLM*>(this)->compTfInner(
            dfDepTmp,
            dfIndep,
            idxNode,
            timestep
        );
        dfDep += dfDepTmp;
    }

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::traceComputeTf(
    CT& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    int const& timestep
)
{
    if (this->isTracedTf()) {
        return;
    }

    if (!(this->params().enabledAdolc())) {
        throw std::runtime_error("Tracing not possible. ADOL-C is not enabled.");
    }

    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Create all tapes"
                  << " for tracing computeTf("
                  << dfIndep.at(0).name()
                  << ",...)."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    boost::mpi::timer timerTrace;
    timerTrace.restart();

#ifdef SCADOPT_HAVE_ADOLC
    // REMARK: Currently, tracing can only done using 1 OpenMP thread.
#ifdef _OPENMP
    const int NTHREADS_ENV = omp_get_num_threads();
    omp_set_num_threads(1);
#endif

    /*------------------------------------------------------------------------*/
    ScaFES::Ntuple<int, DIM> idxNode;
    // Copy constructor would not work because the types of the data fields
    // are different (adouble vs. double).
    unsigned long int nElemsDfIndepTotal = 1;
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        nElemsDfIndepTotal += dfIndep.at(ii).nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    std::vector<adouble> memoryAdfIndep;
    if (0 < nElemsDfIndepTotal) {
        memoryAdfIndep.resize(nElemsDfIndepTotal);
    }
    /*------------------------------------------------------------------------*/
    std::vector< ScaFES::DataField<adouble,DIM> > aDfIndep;
    aDfIndep.reserve(dfIndep.size());
    adouble* elemDataIndepCurr = memoryAdfIndep.data();
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        aDfIndep.push_back(
            ScaFES::DataField<adouble, DIM>(
                dfIndep.at(ii).name(),
                dfIndep.at(ii).params(),
                dfIndep.at(ii).gridGlobal(),
                dfIndep.at(ii).memAll(),
                dfIndep.at(ii).stencilWidth(),
                elemDataIndepCurr,
                1,
                dfIndep.at(ii).memHole(),
                dfIndep.at(ii).nLayers()
            )
        );
        elemDataIndepCurr += dfIndep.at(ii).nElemsAllocated();
    }
    /*------------------------------------------------------------------------*/
    dfDep = static_cast<CT>(0);
    adouble aDfDep = dfDep;

    /*------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------*/
    if (!(this->useAsynchronMode())) {
      this->createTapeForComputeTfPartitionAll(
        this->tapeIdComputeTfPartitionAll(),
        &SCADOPT::Problem<OWNPRBLM, CT, DIM>::template computeTfGlobalInner<adouble>,
        &SCADOPT::Problem<OWNPRBLM, CT, DIM>::template computeTfGlobalBorder<adouble>,
        dfDep,
        dfIndep,
        aDfDep,
        aDfIndep,
        timestep
       );
    }
    else {
      this->createTapesForComputeTf(
        this->tapeIdComputeTfPartitionBorder(),
        &SCADOPT::Problem<OWNPRBLM, CT, DIM>::template computeTfLocalBorder<adouble>,
        dfDep,
        dfIndep,
        aDfDep,
        aDfIndep,
        timestep
      );
      this->createTapesForComputeTf(
        this->tapeIdComputeTfPartitionInner(),
        &SCADOPT::Problem<OWNPRBLM, CT, DIM>::template computeTfLocalInner<adouble>,
        dfDep,
        dfIndep,
        aDfDep,
        aDfIndep,
        timestep
      );
    }

#ifdef _OPENMP
    omp_set_num_threads(NTHREADS_ENV);
#endif

#endif
    this->mWallClockTimes["trace"] += timerTrace.elapsed();
    this->mIsTracedCompTf = true;

    // REMARK: No synchronization is necessary as one just wants to trace
    // the mathematical operations.
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Created."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
#ifdef SCADOPT_HAVE_ADOLC
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::createTapeForComputeTfPartitionAll(
    short int const& tapeId,
    TypeFuncComputeTfToTrace funcInner,
    TypeFuncComputeTfToTrace funcBorder,
    CT& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    adouble& aDfDep,
    std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
    int const& timestep
)
{
    ScaFES::Ntuple<int, DIM> idxNode;
    /*------------------------------------------------------------------------*/
    trace_on(tapeId, 1);
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfIndep[ii].memAll().begin(),
              et = dfIndep[ii].memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfIndep[ii](idxNode) <<= dfIndep[ii](idxNode);
        }
        aDfIndep[ii].time() <<= dfIndep[ii].time();
    }
    // Remark:
    // Reset dependent variable to zero.
    // Just is due to the asynchronous MPI communication and the
    // introduction of a temporary vector.
    aDfDep = 0.0;
    adouble tmpAdfDep = 0.0;

    (this->*funcInner)(tmpAdfDep, aDfIndep, timestep);
    (this->*funcBorder)(aDfDep, aDfIndep, timestep);
    aDfDep += tmpAdfDep;
    
    /*------------------------------------------------------------------------*/
    aDfDep >>= dfDep;
    trace_off();
}
#endif
/*----------------------------------------------------------------------------*/
#ifdef SCADOPT_HAVE_ADOLC
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::createTapesForComputeTf(
    short int const& tapeId,
    TypeFuncComputeTfToTrace funcComputeTf,
    CT& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    adouble& aDfDep,
    std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
    int const& timestep
)
{
    ScaFES::Ntuple<int, DIM> idxNode;
    /*------------------------------------------------------------------------*/
    trace_on(tapeId, 1);
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfIndep[ii].memAll().begin(),
              et = dfIndep[ii].memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfIndep[ii](idxNode) <<= dfIndep[ii](idxNode);
        }
        aDfIndep[ii].time() <<= dfIndep[ii].time();
    }
    // Remark:
    // Reset dependent variable to zero.
    // Just is due to the asynchronous MPI communication and the
    // introduction of a temporary vector.
    aDfDep = 0.0;

    (this->*funcComputeTf)(aDfDep, aDfIndep, timestep);

    /*------------------------------------------------------------------------*/
    aDfDep >>= dfDep;
    trace_off();
}
#endif
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::computeTf_fovForward(
    CT& dfDep,
    std::vector<CT>& dfGradDep,
    short int const& tapeId,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    int const& /*timestep*/
)
{
    CT tmp;
    static_assert(std::is_same<decltype(tmp), double>::value,
                  "Only for CT=double!");

    boost::mpi::timer timerCompTf;
    timerCompTf.restart();

    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank(), "SCADOPT")
                  << " * Compute tf and its gradients using ADOL-C tape #"
                  << tapeId
                  << " at "
                  << dfIndep[0].name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

#ifdef SCADOPT_HAVE_ADOLC
    dfDep = static_cast<CT>(0);
    const int nDirections = this->nColumns();
    const int nDependents = 1;
    int nIndependents = 0;
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        nIndependents += dfIndep[ii].nElemsAllocated();
    }

    // Rearrange one-dimensional arrays to two-dimensional ones
    // such that these arrays can be passed to ADOL-C functions.
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank(), "SCADOPT")
                  << " * nIndependents = " << nIndependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * nDependents   = " << nDependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * nDirections   = " << nDirections << std::endl;
    }

    // Use "this->mMemoryVectGradUnknownDfsRowsTwo" 
    // because data fields will not be swapped between updateUnknownDfs() 
    // and compTargetfunc().
    CT* elemCurr = const_cast<CT*>(dfGradIndep[0].elemData());
    for (int ii = 0; ii < nIndependents; ++ii) {
        this->mMemoryVectGradUnknownDfsRowsTwo[ii] = elemCurr;
        elemCurr += nDirections;
    }

    // Points to first element of first vector of old iterate.
    double* vectIndep = dfIndep[0].elemData();
    // Points to first element of first vector of new iterate.
    double* vectDep = new double[1];
    double** vectIndepGrad = this->mMemoryVectGradUnknownDfsRowsTwo.data();
    double** vectDepGrad = new double*[1];
    vectDepGrad[0] = dfGradDep.data();

    ::fov_forward(
        tapeId,
        nDependents,
        nIndependents,
        nDirections,
        vectIndep,
        vectIndepGrad,
        vectDep,
        vectDepGrad
    );

    dfDep = vectDep[0];
    /*------------------------------------------------------------------------*/
    // Rearrange back two-dimensional arrays (gradients) 
    // to one-dimensional arrays.
    // NOT necessary: Values are already at correct position in memory.
    // Local variables vectDep, vectGradDep etc. can be thrown away.
    // These variables are just pointers to the memory of the data fields.
    // Not necessary here:
    // vectDepGrad[0] points already to this->mVectGradTf.

    delete[] vectDep;
    delete[] vectDepGrad;
#endif

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }

    this->mWallClockTimesOptim["compTf"] += timerCompTf.elapsed();

}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::computeTf_zosForward(
    CT& dfDep,
    short int const& tapeId,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    int const& /*timestep*/
)
{
    CT tmp;
    static_assert(std::is_same<decltype(tmp), double>::value,
                  "Only for CT=double!");

    boost::mpi::timer timerCompTf;
    timerCompTf.restart();

    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank(), "SCADOPT")
                  << " * Compute tf using ADOL-C tape #"
                  << tapeId
                  << " at "
                  << dfIndep[0].name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

#ifdef SCADOPT_HAVE_ADOLC
    dfDep = static_cast<CT>(0);
    const int nDirections = this->nColumns();
    const int nDependents = 1;
    int nIndependents = 0;
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        nIndependents += dfIndep[ii].nElemsAllocated();
    }

    // Rearrange one-dimensional arrays to two-dimensional ones
    // such that these arrays can be passed to ADOL-C functions.
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank(), "SCADOPT")
                  << " * nIndependents = " << nIndependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * nDependents   = " << nDependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * nDirections   = " << nDirections << std::endl;
    }

    // Points to first element of first vector of old iterate.
    double* vectIndep = static_cast<double*>(dfIndep[0].elemData());
    // Points to first element of a new arrary (= tf value).
    double* vectDep = new double[1];

    /*------------------------------------------------------------------------*/
    ::zos_forward(
        tapeId,
        nDependents,
        nIndependents,
        0, // Do not keep.
        vectIndep,
        vectDep
    );

    dfDep = vectDep[0];
    delete[] vectDep;
#endif

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }

    this->mWallClockTimesOptim["compTf"] += timerCompTf.elapsed();
}

} // End of namespace.
#endif
