#include <cstdlib>
#include <stdexcept>

#include <SCADOPT.hpp>

#ifdef _OPENMP
#include <omp.h>
#endif

template<typename CT, std::size_t DIM = 3>
class TestOpenMPFwd
    : public SCADOPT::Problem< TestOpenMPFwd<CT,DIM>, CT, DIM> {
private:
    typedef TestOpenMPFwd<CT,DIM> Self;
    typedef SCADOPT::Problem< Self, CT, DIM> Super;
    typedef std::vector<std::string>::size_type size_type;

    typedef ScaFES::Ntuple<int, DIM> Index;
    typedef typename Index::value_type index_type;

    static bool isOmpUsed()
    {
#ifdef _OPENMP
        const int omp_max = omp_get_max_threads();
        const int omp_curr = omp_get_num_threads();

        int omp_env = omp_max;
        const char* p = std::getenv("OMP_NUM_THREADS");
        if (p != nullptr) {
            //Currently, this is not used [KF, 05.04.2014].
            //const int nt = boost::lexical_cast<int>(p);
        }

        if ( (omp_max > omp_curr) && (omp_max == omp_env) ) {
            return false;
        }

        return true;
#else
        // hardcode non-failing test
        return true;
#endif
    }

public:

    TestOpenMPFwd(ScaFES::Parameters const& cl,
                  ScaFES::GridGlobal<DIM> const& gg,
            bool useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<bool> const& optimizeForDf = std::vector<bool>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile
                = std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>(),
            std::vector<CT> const& geomparamsSol = std::vector<CT>() )
            : SCADOPT::Problem<TestOpenMPFwd<CT, DIM>, CT, DIM>(
                     cl, gg, useLeapfrog, nameDatafield, stencilWidth,
                     isKnownDf, nLayers, optimizeForDf,
                     defaultValue, writeToFile,
                     computeError, geomparamsInit, geomparamsSol) { }

    template<typename TT>
    void initInner(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("initInner(): OpenMP not used, but enabled in compilation.");
    }
    template<typename TT>
    void initBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("initBorder(): OpenMP not used, but enabled in compilation.");
    }

    template<typename TT>
    void initInner(std::vector< ScaFES::DataField<TT, DIM> >& /*dfs*/,
            std::vector<TT> const& /*p*/,
            Index const& /*idx*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("initInner(): OpenMP not used, but enabled in compilation.");
    }
    template<typename TT>
    void initBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*dfs*/,
            std::vector<TT> const& /*p*/,
            Index const& /*idx*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("initBorder(): OpenMP not used, but enabled in compilation.");
    }

    template<typename TT>
    void updateInner(std::vector< ScaFES::DataField<TT, DIM> >& /*n*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*o*/,
            Index const& /*idx*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("updateInner(): OpenMP not used, but enabled in compilation.");
    }

    template<typename TT>
    void updateBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM>>const& /*vOld*/,
            Index const& /*idx*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("updateBorder(): OpenMP not used, but enabled in compilation.");
    }

    template<typename TT>
    void updateInner2(std::vector< ScaFES::DataField<TT, DIM> >& /*n*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*o*/,
            Index const& /*idx*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("updateInner2(): OpenMP not used, but enabled in compilation.");
    }

    template<typename TT>
    void updateBorder2(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM>>const& /*vOld*/,
            Index const& /*idx*/,
            int const& /*timestep*/)
    {
        if (!this->isOmpUsed())
            throw std::runtime_error("updateBorder2(): OpenMP not used, but enabled in compilation.");
    }

    template<typename TT>
    void compTfInner(TT& /*tf*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/) { }
    template<typename TT>
    void compTfBorder(TT& /*tf*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/) { }

    template<typename TT>
    void evalInner(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
        Index const& /*idxNode*/,
        int const& /*timestep*/) { }
    template<typename TT>
    void evalBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
        Index const& /*idxNode*/,
        int const& /*timestep*/) { }
};
