#include <SCADOPT.hpp>

template<typename CT, std::size_t DIM>
class OptHeatPara : public SCADOPT::Problem< OptHeatPara<CT,DIM>, CT, DIM > {
public:
    OptHeatPara(ScaFES::Parameters const& cl,
                ScaFES::GridGlobal<DIM> const& gg,
            bool useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<bool> const& optimizeForDf = std::vector<bool>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile
                = std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>(),
            std::vector<CT> const& geomparamsSol = std::vector<CT>() )
            : SCADOPT::Problem<OptHeatPara<CT, DIM>, CT, DIM>(
                     cl, gg, useLeapfrog, nameDatafield, stencilWidth,
                     isKnownDf, nLayers, optimizeForDf,
                     defaultValue, writeToFile,
                     computeError, geomparamsInit, geomparamsSol) { }

    template<typename TT>
    void evalInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                   ScaFES::Ntuple<int,DIM> const& idxNode,
                   int const& /*timestep*/) {
        const double X_HAT = 1.0;
        const double PI = 3.14159265358979323846;
        vNew[0](idxNode) = X_HAT * exp(-PI * PI / (2.0 * X_HAT * X_HAT) );
        ScaFES::Ntuple<double,DIM> x = this->coordinates(idxNode);
        for (std::size_t pp = 0; pp < DIM; ++pp) {
            vNew[0](idxNode) *= sin(2.0 * PI * x[pp] / X_HAT);
        }
        vNew[1](idxNode) = 0.0;
    }
    template<typename TT>
    void evalBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                    ScaFES::Ntuple<int,DIM> const& idxNode,
                    int const& timestep) {
        this->evalInner(vNew, idxNode, timestep);
    }
    template<typename TT>
    void initInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                  std::vector<TT> const& vOld,
                  ScaFES::Ntuple<int,DIM> const& idxNode,
                  int const& /*timestep*/) {
        ScaFES::Ntuple<double,DIM> x = this->coordinates(idxNode);
        vNew[0](idxNode) = vOld[0];
        for (std::size_t pp = 0; pp < DIM; ++pp) {
            vNew[0](idxNode) *= sin(x[pp] * vOld[pp+1]);
        }
    }
    template<typename TT>
    void initBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                    std::vector<TT> const& vOld,
                    ScaFES::Ntuple<int,DIM> const& idxNode,
                    int const& timestep) {
        this->template initInner<TT>(vNew, vOld, idxNode, timestep);
    }
    template<typename TT>
    void updateInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                     std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
                     ScaFES::Ntuple<int,DIM> const& idxNode,
                     int const& /*timestep*/) {
        vNew[0](idxNode) = vOld[0](idxNode);
        for (std::size_t pp = 0; pp < DIM; ++pp) {
            vNew[0](idxNode) -= this->tau() * (
                 2.0 * vOld[0](idxNode)
                     - vOld[0](this->connect(idxNode, 2*pp))
                     - vOld[0](this->connect(idxNode, 2*pp+1)) )
                     / (this->gridsize(pp) * this->gridsize(pp));
        }
    }
    template<typename TT>
    void updateBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT,DIM>> const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& /*timestep*/) {
        vNew[0](idxNode) = vOld[0](idxNode);
    }
    template<typename TT>
    void compTfInner(TT& vNew,
                     std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
                     ScaFES::Ntuple<int,DIM> const& idxNode,
                     int const& /*timestep*/) {
        const double LAMBDA = 1.0e-10;
        vNew = (vOld[0](idxNode) - this->knownDf(0,idxNode))
              * (vOld[0](idxNode) - this->knownDf(0,idxNode)) / 2.0
              +  LAMBDA * ( this->knownDf(1,idxNode) ) / 2.0;
    }
    template<typename TT>
    void compTfBorder(TT& vNew,
        std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
        ScaFES::Ntuple<int,DIM> const& idxNode, int const& timestep) {
        this->template compTfInner<TT>(vNew, vOld, idxNode, timestep);
    }
};
