/** ScaFES
 *  Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCAFES_PROBLEM_HPP_
#define SCAFES_PROBLEM_HPP_

#include "ScaFES_Config.hpp"

#include <iostream>
#include <iomanip>
#include <ios>
#include <fstream>
#include <string>
#include <cassert>
#include <vector>
#include <stdexcept>
#include <unordered_map>

extern "C" {
#ifdef _OPENMP
#include <omp.h>
#endif
}

#ifdef SCAFES_HAVE_ADOLC
#include <adolc/adolc.h>
#include <adolc/adouble.h>
#ifdef _OPENMP
#include <adolc/adolc_openmp.h>
#endif // _OPENMP 
#endif // SCAFES_HAVE_ADOLC


#ifdef SCAFES_HAVE_BOOST
#include <boost/type_traits.hpp>
#include <boost/assign/list_of.hpp>
#endif

#ifdef SCAFES_HAVE_BOOST_SERIALIZATION
namespace boost
{
namespace serialization
{
class access;
}
}
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/version.hpp>
#include <boost/serialization/pfto.hpp>
#include <boost/serialization/vector.hpp>

#ifdef SCAFES_HAVE_ADOLC
namespace boost
{
namespace serialization
{
template<class Archive>
void serialize(Archive& ar, adouble& g,
               const unsigned int /*version*/)
{
    double tmp = g.getValue();
    ar & (tmp);
#if defined(TAPELESS)
    adtl::ADVAL_TYPE tmp2 = g.getADValue();
    ar & (tmp2);
#endif
}
}
}
#endif // SCAFES_HAVE_ADOLC
#endif // SCAFES_HAVE_BOOST_SERIALIZATION

#ifdef SCAFES_HAVE_BOOST_MPI
#include <boost/mpi.hpp>
#endif

#ifdef VTRACE
#include "vt_user.h"
#endif

#include "ScaFES_Ntuple.hpp"
#include "ScaFES_Parameters.hpp"
#include "ScaFES_GridSub.hpp"
#include "ScaFES_GridGlobal.hpp"
#include "ScaFES_DataFile.hpp"
#include "ScaFES_DataField.hpp"

#ifdef SCAFES_HAVE_CUDA
#include "ScaFES_ProblemCuda.hpp"
#endif

namespace ScaFES
{

/*******************************************************************************
 ******************************************************************************/
/** The class \c Problem is designed to describe an initial boundary value 
 * problem (IBVP) on a given MPI partition. An IBVP usually consists of 
 * several data fields (some of them are unknown, some of them are given).
 *
 * \n In order to solve a particular IBVP, the user has to implement an own
 * class that is inherited by the class \c Problem and has to implement
 * different methods like \c updateInner, \c updateBorder.
 *
 * \n ScaFES provides the method \c iterateOverTime to handle the iteration
 * over a given number of time steps (forward computation).
 * The method will process the following steps:
 * Compute the new values of all data fields at all grid nodes, synchronise
 * values with neighboured partitions, redefine new situation (new data field
 * will be the old one in the next time step). It is sufficient to swap the
 * pointers to the memories of the old and the new data field.
 */
template<class OWNPRBLM, typename CT = double, std::size_t DIM = 3>
class Problem
{
    public:
        /*----------------------------------------------------------------------
        | TYPE DEFINITIONS.
        ----------------------------------------------------------------------*/
        /** Type definition for a free scalar valued function. */
        typedef void (*Scalarfunc)(
            CT&, 
            ScaFES::Ntuple<CT, DIM> const&, 
            CT const&);

        /** Type definition for a member scalar valued function. */
        typedef void (OWNPRBLM::*ClassScalarfunc)(
            CT&,
            ScaFES::Ntuple<CT, DIM> const&,
            CT const&);

#ifdef SCAFES_HAVE_ADOLC
        /** Type definition for the member init methods of this class. */
        typedef void (Problem<OWNPRBLM,CT,DIM>::*TypeFuncInitToTrace)(
                std::vector< ScaFES::DataField<adouble,DIM> >&,
                std::vector< adouble > const&,
                int const&);
#endif

        /** Type definition for the member init methods of this class. */
        typedef void (Problem<OWNPRBLM,CT,DIM>::*TypeFuncInitToEval)(
                std::vector< ScaFES::DataField<CT,DIM> >&,
                std::vector< ScaFES::DataField<CT,DIM> >&,
                std::vector<CT> const&,
                std::vector<CT> const&,
                int const&);

#ifdef SCAFES_HAVE_ADOLC
        /** Type definition for the member update methods of this class. */
        typedef void (Problem<OWNPRBLM,CT,DIM>::*TypeFuncUpdateToTrace)(
                std::vector< ScaFES::DataField<adouble,DIM> >&,
                std::vector< ScaFES::DataField<adouble,DIM> > const&,
                int const&);
#endif

        /** Type definition for the member update methods of this class. */
        typedef void (Problem<OWNPRBLM,CT,DIM>::*TypeFuncUpdateToEval)(
                std::vector< ScaFES::DataField<CT,DIM> >&,
                std::vector< ScaFES::DataField<CT,DIM> >&,
                std::vector< ScaFES::DataField<CT,DIM> > const&,
                std::vector< ScaFES::DataField<CT,DIM> > const&,
                int const&);

        /*----------------------------------------------------------------------
        | FRIEND CLASSES.
        ----------------------------------------------------------------------*/
        friend class boost::serialization::access;

        /*----------------------------------------------------------------------
        | LIFE CYCLE METHODS.
        ----------------------------------------------------------------------*/
        /** No default constructor. */
        Problem() = delete;

        /** Creates own constructor.
         * First, creates a partition file.
         * Second, runs the program based on the created partition file.
         * Third, generates the global grid from command line data.
         * Sets MPI-related stuff (communicator, MPI size,  MPI rank)
         * and adopts the global grid from the command line parameters.
         * Then, the master process broadcasts the parameters.
         * Creates one single memory lump for all unknown data fields.
         * Two data fields will be created: One for the last time step,
         * a second one for the current time step.
         */
        Problem(
            Parameters const& params,
            GridGlobal<DIM> const& gg,
            const bool& useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<WriteHowOften> const& writeToFile
                = std::vector<WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>()
        );

        /** Creates compiler provided copy constructor. */
        Problem(Problem<OWNPRBLM, CT, DIM> const& /*ppp*/) = default;

        /** Creates compiler provided assignment operator. */
        Problem& operator=(Problem<OWNPRBLM, CT, DIM> const&) = default;

        /** Creates own destructor. */
        virtual ~Problem() = default;

        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns the parameters. */
        Parameters const& params() const;

        /** Returns the grid size in a given direction. */
        double const& gridsize(int const& idx) const;

        /** Returns the time step size. */
        double const& tau() const;

        /** Returns the number of time steps. */
        int const& nTimesteps() const;

        /** Returns the pointer to global grid. */
        GridGlobal<DIM> const& globalGrid() const;

        /** Returns the node type of a given global node. */
        short int const& kind(ScaFES::Ntuple<int,DIM> const& idxNode) const;

        /** Returns the node type of a given global node. */
        ScaFES::DataField<short int,DIM> const& kind() const;

        /** Returns the number of nodes in each direction. */
        ScaFES::Ntuple<int,DIM> const& nNodes() const;

        /** Returns the number of nodes in a given direction. */
        int const& nNodes(int const& dim) const;

        /** Returns the coordinates of a given global node. */
        ScaFES::Ntuple<double,DIM> coordinates(
            ScaFES::Ntuple<int,DIM> const& idxNode
        ) const;

        /** Returns the global local node of a given global node which
         * resides in a given direction. */
        ScaFES::Ntuple<int,DIM> connect(
            ScaFES::Ntuple<int,DIM> const& idxNode, 
            int const& dir
        );

        /** Returns the numerical solution of parameters. */
        std::vector<CT> const& vectParamsCurr() const;

        /** Returns the analytical solution vector of all data fields. */
        std::vector< ScaFES::DataField<CT,DIM> > const& vectSol() const;

        /** Returns the value of analytical solution data field given by \c idx
         * at the grid node with node number \c idxNode.
         */
        CT const& vectSol(
            int const& idx, 
            ScaFES::Ntuple<int,DIM> const& idxNode
        ) const;

        /** Returns the value of the known data field given by \c idx
         * at the grid node with node number \c idxNode.
         */
        CT const& knownDf(
            int const& idx, 
            ScaFES::Ntuple<int,DIM> const& idxNode
        ) const;

        /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Iterates over time. */
        void iterateOverTime();

        /** Updates the data fields at all global inner grid nodes
         *  at a given time step. */
        template<typename TT>
        void updatePartitionAllInner(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Updates the data fields at all global border grid nodes
         *  at a given time step. */
        template<typename TT>
        void updatePartitionAllBorder(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Updates the data fields at all local inner grid nodes
         *  at a given time step. */
        template<typename TT>
        void updatePartitionInner(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Updates the data fields at all local border grid nodes
         *  at a given time step. */
        template<typename TT>
        void updatePartitionBorder(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

    protected:
        /*----------------------------------------------------------------------
        | SETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Increases output level by one. */
        void increaseLevel();

        /** Decreases output level by one. */
        void decreaseLevel();

        /*----------------------------------------------------------------------
        | GETTER METHODS.
        ----------------------------------------------------------------------*/
        /** Returns if the initialization phase was already traced. */
        bool const& isTracedInit() const;

        /** Returns the tape identifier corresponding to the tape
         * containing the initialization at all inner partition grid nodes. */
        short int tapeIdInitPartitionInner();

        /** Returns the tape identifier corresponding to the tape
         * containing the initialization at all local partition grid nodes. */
        short int tapeIdInitPartitionBorder();

        /** Returns the tape identifier corresponding to the tape of the
         * containing the initialization at all partition grid nodes. */
        short int tapeIdInitPartitionAll();

        /** Returns if the update phase was already traced. */
        bool const& isTracedUpdate() const;

        /** Returns the tape identifier corresponding to the tape
         * containing the update step at all inner partition grid nodes. */
        short int const& tapeIdUpdatePartitionInner() const;

        /** Returns the tape identifier corresponding to the tape
         * containing the update step at all border partition grid nodes. */
        short int const& tapeIdUpdatePartitionBorder() const;

        /** Returns the tape identifier corresponding to the tape
         * containing the update step at all all partition grid nodes. */
        short int const& tapeIdUpdatePartitionAll() const;

        /** Returns the tape identifier corresponding to the tape
         * containing the update2 step at all inner partition grid nodes. */
        short int const& tapeIdUpdatePartitionInner2() const;

        /** Returns the tape identifier corresponding to the tape
         * containing the update2 step at all border partition grid nodes. */
        short int const& tapeIdUpdatePartitionBorder2() const;

        /** Returns the tape identifier corresponding to the tape
         * containing the update2 step at all all partition grid nodes. */
        short int const& tapeIdUpdatePartitionAll2() const;

        /** Returns the rank of underlying MPI process. */
        int myRank() const;

        /** Returns the total number of nodes of the grid partition. */
        int const& nNodesTotalPartition() const;

        /** Returns the flags if a data field is a known one or not. */
        std::vector<bool> const& isKnownDf() const;

        /** Returns the flags where a data field is defined. */
        std::vector<int> const& nLayers() const;

        /** Returns the default values of all data fields. */
        std::vector<CT> const& defaultValue() const;

        /** Returns the number of elements of all unknown data fields. */
        int const& nElemsUnknownDfs() const;

        /** Returns the number of elements of all known data fields. */
        int const& nElemsKnownDfs() const;

        /** Returns the number of those data fields which are unknown and
         *  defined at the boundary. */
        int const& nUnknownBdryDfs() const;

        /** Returns the vector of all unknown data fields defined at domain. */
        std::vector< ScaFES::DataField<CT,DIM> > const& 
        vectUnknownDfsDomNew() const;

        /** Returns the old vector of all data fields defined at domain. */
        std::vector< ScaFES::DataField<CT,DIM> > const& 
        vectUnknownDfsDomOld() const;

        /** Returns the new vector of all data fields defined at boundary. */
        std::vector< ScaFES::DataField<CT,DIM> > const& 
        vectUnknownDfsBdryNew() const;

        /** Returns the old vector of all data fields defined at boundary. */
        std::vector< ScaFES::DataField<CT,DIM> > const& 
        vectUnknownDfsBdryOld() const;

        /** Returns the number of directions during tracing. */
        int const& nColumns() const;

        /** Returns if gradients should be computed during the
         * initialization and the update phase. */
        bool const& computeGradients() const;

        /** Returns if asynchron MPI communication will be used. */
        bool const& useAsynchronMode() const;

        /** Returns if leap frog scheme will be used. */
        bool const& useLeapfrogIntegration() const;

         /** Returns if the parameter based ansatz will be used. */
        bool const& isParamsBased() const;

       /*----------------------------------------------------------------------
        | WORK METHODS.
        ----------------------------------------------------------------------*/
        /** Creates new vector of unknown data fields.
         * Splits up allocated memory and distributes it to all data fields. */
        void createVectorOfUnknownDataFields(
            std::vector< ScaFES::DataField<CT,DIM> >& dfInitDomain,
            std::vector< ScaFES::DataField<CT,DIM> >& dfInitBdry,
            std::vector<CT> const& memoryOfDfs,
            std::vector<std::string> const& nameDatafield,
            std::string const& nameSuffix,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers,
            std::vector<int> const& stencilWidth,
            std::vector< ScaFES::Grid<DIM> > const& memAll,
            std::vector< ScaFES::GridSub<DIM> > const& memHolePart,
            std::vector<int> const& nRowsDf,
            std::vector<ScaFES::WriteHowOften> const& writeToFile,
            std::vector<CT> const& defaultValue,
            int const& nColumns
        );

        /** Initialise all unknown data fields. */
        void initUnknownDfs();

        /** Update values at all grid partition nodes of all data fields 
         *  (1st cycle). */
        void updateValsAtNodesOfUnknownDfs(int const& timestep);

        /** Update values at all grid partition nodes of all data fields 
         *  (2nd cycle) using the leapfrog scheme. */
        void update2ValsAtNodesOfUnknownDfs(int const& timestep);

        /** Update times of all unknown data fields. */
        void updateTimesOfUnknownDfs();

        /** Swap pointers of all unknown data fields. */
        void swapPointersOfUnknownDfs();

        /** Swap times of all unknown data fields. */
        void swapTimesOfUnknownDfs();

        /** Evaluate known data fields. */
        void evalKnownDfs(int const& timestep);

        /** Compute errors. */
        void compErrOfUnknownDfs();

        /** Write data fields to file. */
        void writeAllDfs(int const& timestep);

        /** Determines which MPI communication mode will be used. */
        void setCommunicationMode();

        /** Determines if gradients should be computed during
         * init and update phase. */
        void setCompGradients();

        /*--------------------------------------------------------------------*/
        /** Initialise data fields asynchronously. */
        void initAsynchronously(
            TypeFuncInitToEval funcLocalInner,
            TypeFuncInitToEval funcLocalBorder,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector<CT> const& dfIndep,
            std::vector<CT> const& dfGradIndep,
            int const& timestep
        );

        /** Initialise data fields synchronously. */
        void initSynchronously(
            TypeFuncInitToEval funcGlobal,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepDomain,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradDomain,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepBdry,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradBdry,
            std::vector<CT> const& dfIndep,
            std::vector<CT> const& dfGradIndep,
            int const& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperInitPartitionInnerPb(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector<CT> const& dfIndep,
            std::vector<CT> const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperInitPartitionBorderPb(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector<CT> const& dfIndep,
            std::vector<CT> const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperInitPartitionAllPb(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector<CT> const& dfIndep,
            std::vector<CT> const& dfGradIndep,
            const int& timestep
        );

        /** Initialises all data fields at all global inner nodes
         * depending on the function implemented in the sub class.
         */
        template<typename TT>
        void initPartitionAllInnerPb(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector<TT> const& dfIndep,
            int const& timestep
        );

        /** Initialises all data fields at all global border nodes
         * depending on the function implemented in the sub class.
         */
        template<typename TT>
        void initPartitionAllBorderPb(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector<TT> const& dfIndep,
            int const& timestep
        );

        /** Initialises all data fields at all local inner nodes
         * depending on the function implemented in the sub class.
         */
        template<typename TT>
        void initPartitionInnerPb(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector<TT> const& dfIndep,
            int const& timestep
        );

        /** Initialises all data fields at all local border nodes
         * depending on the function implemented in the sub class.
         */
        template<typename TT>
        void initPartitionBorderPb(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector<TT> const& dfIndep,
            int const& timestep
        );

        /** Initialises all data fields depending on the function
         * implemented in the sub class.
         */
        template<typename TT>
        void initGridBased(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep
        );

        /*--------------------------------------------------------------------*/
        /** Traces the initialization step.
         * Concrete example:
         * independent variables = all parameters which describe the ellipsoid.
         * dependent variables = all data fields, because they depend on the
         *                           material coefficients.
         */
        void traceInitParamsBased(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector<CT> const& dfIndep
        );

#ifdef SCAFES_HAVE_ADOLC
        /** Creates 2 tapes for init phase at all grid partition nodes:
         * 1 tape for the init step at all inner grid partition nodes.
         * 1 tape for the init step at all border grid partition nodes. */
        void createTapesForInit(
            short int const& tapeId,
            TypeFuncInitToTrace funcInit,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< CT > const& dfIndep,
            std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
            std::vector< adouble >& aDfIndep,
            int const& timestep
        );

        /** Creates 1 tape for init phase at all grid partition nodes. */
        void createTapeForInitPartitionAll(
            short int const& tapeId,
            TypeFuncInitToTrace funcInitInner,
            TypeFuncInitToTrace funcInitBorder,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< CT > const& dfIndep,
            std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
            std::vector< adouble >& aDfIndep,
            int const& timestep
        );
#endif

        /** Initialises data fields using ADOL-C tapes (params based). */
        void init_zosForward_paramsBased(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            short int const& tapeId,
            std::vector<CT> const& dfIndep,
            const int& timestep
        );

        /** Initialises the data fields and its gradients using ADOL-C tapes 
         *  (params based). */
        void init_fovForward_paramsBased(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            short int const& tapeId,
            std::vector<CT> const& dfIndep,
            std::vector<CT> const& dfGradIndep,
            const int& timestep
        );

        /*--------------------------------------------------------------------*/
        /** Updates unknown data fields asynchronously. */
        void updateAsynchronously(
            TypeFuncUpdateToEval funcLocalInner,
            TypeFuncUpdateToEval funcLocalBorder,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            int const& timestep
        );

        /** Updates unknown data fields synchronously. */
        void updateSynchronously(
            TypeFuncUpdateToEval funcGlobal,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepDomain,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradDomain,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepBdry,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradBdry,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepDomain,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndepDomain,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepBdry,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndepBdry,
            int const& timestep
        );


        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperUpdatePartitionInner(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperUpdatePartitionBorder(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperUpdatePartitionAll(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

       /*--------------------------------------------------------------------*/
        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperUpdatePartitionInner2(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperUpdatePartitionBorder2(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /** Wrapper: Same interface for normal and tape evaluation. */
        void wrapperUpdatePartitionAll2(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            const int& timestep
        );

        /*--------------------------------------------------------------------*/
        /** Traces the update phase using ADOL-C. */
        void traceUpdate(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            int const& timestep
        );

#ifdef SCAFES_HAVE_ADOLC
        /** Creates 1 tape for update phase at all grid partition nodes. */
        void createTapeForUpdatePartitionAll(
            short int const& tapeId,
            TypeFuncUpdateToTrace funcInner,
            TypeFuncUpdateToTrace funcBorder,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
            std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
            int const& timestep
        );

        /** Creates 2 tapes for update phase at all grid partition nodes:
         * 1 tape for the update step at all inner grid partition nodes.
         * 1 tape for the update step at all border grid partition nodes. */
        void createTapesForUpdate(
            short int const& tapeId,
            TypeFuncUpdateToTrace funcUpdate,
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
            std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
            int const& timestep
        );
#endif

        /** Computes new iterate using ADOL-C tape
         * and the AD forward mode. */
        void update_zosForward(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            short int const& tapeId,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            int const& currTimestep
        );

        /** Computes new iterate and its gradient using ADOL-C tape
         * and the AD forward mode. */
        void update_fovForward(
            std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
            short int const& tapeId,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
            std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
            int const& currTimestep
        );

        /*--------------------------------------------------------------------*/
        /** Calls barrier for MPI-communicator.
         * Returns, if all processes have entered it. */
        void worldBarrier();

        /** Sets up the kind field w.r.t. the global grid
         * (global inner node, global border node). */
        void setGridGlobalFlags();

        /** Sets up the kind field w.r.t. the underlying grid partition:
         * (local inner node, local border node). */
        void setGridPartitionFlags(int const& maxStencilWidth);

        /** Determines the number of grid partition nodes which are
         * inner local AND inner global. */
        int getNnodesLocInnerGlobInner() const;

        /** Determines the number of grid partition nodes which are
         * inner local AND border global. */
        int getNnodesLocInnerGlobBorder() const;

        /** Determines the number of grid partition nodes which are
         * border local AND inner global. */
        int getNnodesLocBorderGlobInner() const;

        /** Determines the number of grid partition nodes which are
         * border local AND border global. */
        int getNnodesLocBorderGlobBorder() const;

        /*--------------------------------------------------------------------*/
        /** Evaluates a given function at all global inner nodes of the
         * grid partition. */
        template<typename TT>
        void evalGlobInner(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            int const& timestep
        );

        /** Evaluates a given function at all global border nodes of the
         * grid partition. */
        template<typename TT>
        void evalGlobBorder(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            int const& timestep
        );

        /*--------------------------------------------------------------------*/
        /** Updates the data fields at all grid partition nodes
         *  at a given time step. */
        template<typename TT>
        void update(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Writes all data fields to one NetCDF file depending
         * on the parameter of the constructor. */
        void writeDfsToFile(int const& timestep);

        /** Updates the data fields at all local inner grid nodes
         *  at a given time step (LEAP FROG). */
        template<typename TT>
        void updatePartitionInner2(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Updates the data fields at all local border grid nodes
         *  at a given time step (LEAP FROG). */
        template<typename TT>
        void updatePartitionBorder2(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );
        /** Updates the data fields at all local inner grid nodes
         *  at a given time step (LEAP FROG). */
        template<typename TT>
        void updatePartitionAllInner2(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );

        /** Updates the data fields at all local border grid nodes
         *  at a given time step (LEAP FROG). */
        template<typename TT>
        void updatePartitionAllBorder2(
            std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
            std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep,
            int const& timestep
        );
        /*----------------------------------------------------------------------
        | MEMBER VARIABLES.
        ----------------------------------------------------------------------*/
        /** Parameters of the problem. */
        ScaFES::Parameters mParams;

        /** Use leapfrog integration scheme. */
        const bool mUseLeapfrogIntegration;

        /** List of interfaces to the neighbours. */
        /** Pointer to related global grid to access partitions and more. */
        ScaFES::GridGlobal<DIM> mGG;

        /** Names of all data fields. */
        std::vector<std::string> mNameDataField;

        /** Stencil widths of all data fields. */
        std::vector<int> mStencilWidth;

        /** Flags for all data fields if a data field is a known one or not. */
        std::vector<bool> mIsKnownDf;

        /** Flags for all dfs where a data field is defined. */
        std::vector<int> mNlayers;

        /** Default values for all dfs. */
        std::vector<CT> mDefaultValue;

        /** Flags for all dfs if a data field should be written or not. */
        std::vector<WriteHowOften> mWriteToFile;

        /** Flags for all data fields if error should be computed or not. */
        std::vector<bool> mComputeError;

        /** Number of given parameters. */
        int mNparams;

        /** Compute gradients? */
        bool mComputeGradients;

        /** Use asynchronous MPI communication? */
        bool mUseAsynchronMode;

        /** (Geometrical) parameters based ansatz? */
        bool mIsParamsBased;

        /** Tape identifier for initialization step. */
        short int mTapeIdInitPartitionInner;

        /** Tape identifier for initialization step. */
        short int mTapeIdInitPartitionBorder;

        /** Tape identifier for initialization step. */
        short int mTapeIdInitPartitionAll;

        /** Are methods \c initInner and \c initBorder traced? */
        bool mIsTracedInit;

        /** Tape identifier for update step at all local inner nodes. */
        short int mTapeIdUpdatePartitionInner;

        /** Tape identifier for update step at all local border nodes. */
        short int mTapeIdUpdatePartitionBorder;

        /** Tape identifier for update step at all global inner nodes. */
        short int mTapeIdUpdatePartitionAll;

        /** Tape identifier for update step 2 at all local inner nodes. */
        short int mTapeIdUpdatePartitionInner2;

        /** Tape identifier for update step 2 at all local border nodes. */
        short int mTapeIdUpdatePartitionBorder2;

        /** Tape identifier for update step 2 at all global inner nodes. */
        short int mTapeIdUpdatePartitionAll2;

        /** Are methods \c updateInner and \c updateBorder traced? */
        bool mIsTracedUpdate;

        /** Number of total grid nodes of the grid partition. */
        int mNnodesTotalPartition;

        /** Number of elements of all KNOWN data fields. */
        int mNelemsKnownDfs;

        /** Number of elements of all UNKNOWN data fields. */
        int mNelemsUnknownDfs;

        /** Number of UNKNOWN AND BOUNDARY data fields. */
        int mNunknownBdryDfs;

        /** Wall clock time for different sections. */
        std::unordered_map<std::string, double> mWallClockTimes;
        
        /** Global output file. */
        ScaFES::DataFile<CT, DIM> mCommonFile;

        /** Type of each grid node. */
        ScaFES::DataField<short int,DIM> mKind;

        /** Continuous memory lump for kind field. */
        std::vector<short int> mMemoryKind;

        /** List of all grid partition nodes which are inner grid partition 
         * nodes AND which are inner grid global nodes. */
        std::vector<int> mNodesLocInnerGlobInner;

         /** List of all grid partition nodes which are inner grid partition 
         * nodes AND which are border grid global nodes. */
        std::vector<int> mNodesLocInnerGlobBorder;

         /** List of all grid partition nodes which are border grid partition 
         * nodes AND which are inner grid global nodes. */
        std::vector<int> mNodesLocBorderGlobInner;

        /** List of all grid partition nodes which are border grid partition 
         * nodes AND which are border grid global nodes. */
        std::vector<int> mNodesLocBorderGlobBorder;

        /** Vector of all known data fields defined at the whole domain. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectKnownDfDom;

        /** Vector of all known data fields defined at the boundary, only. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectKnownDfBdry;

        /** Continuous memory lump for all known data fields. */
        std::vector<CT> mMemoryVectKnownDf;

        /** Numerical solution at last time step defined at the domain. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectUnknownDfsDomOld;

        /** Numerical solution at last time step defined at the boundary. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectUnknownDfsBdryOld;

        /** Numerical solution at current time step defined at the domain. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectUnknownDfsDomNew;

        /** Numerical solution at current time step defined at the boundary. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectUnknownDfsBdryNew;

        /** Continuous memory lump #1 for storing values of all unknown 
         *  data fields at all grid partition nodes (incl. halos) 
         *  (old / new scheme). */
        std::vector<CT> mMemoryVectUnknownDfsOne;

        /** Continuous memory lump #2 for storing values of all unknown 
         *  data fields at all grid partition nodes (incl. halos) 
         *  (old / new scheme). */
        std::vector<CT> mMemoryVectUnknownDfsTwo;

        /** Continuous memory lump #3 for storing values of all unknown 
         *  data fields at all grid partition nodes (incl. halos) 
         *  This memory lump will be allocated only if a temp. vector
         *  of data fields is needed (case: async. MPI comm. + ADOL-C). */
        std::vector<CT> mMemoryVectUnknownDfsThree;

        /** Gradient of numerical solution at last time step
         *  defined at the domain. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectGradUnknownDfsDomOld;

        /** Gradient of numerical solution at last time step. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectGradUnknownDfsBdryOld;

        /** Gradient of numerical solution at current time step. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectGradUnknownDfsDomNew;

        /** Gradient of numerical solution at current time step. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectGradUnknownDfsBdryNew;

        /** Continuous memory lump #1 for storing values of gradients 
         *  of all unknown data fields at all grid partition nodes (incl. halos) 
         *  (old / new scheme). */
        std::vector<CT> mMemoryVectGradUnknownDfsOne;

        /** Continuous memory lump #2 for storing values of gradients 
         *  of all unknown data fields at all grid partition nodes (incl. halos) 
         *  (old / new scheme). */
        std::vector<CT> mMemoryVectGradUnknownDfsTwo;

        /** Continuous memory lump #2 for storing values of gradients 
         *  of all unknown data fields at all grid partition nodes (incl. halos) 
         *  \remark This memory lump will be allocated only if a temp. vector
         *  of data fields is needed (case: async. MPI comm. + ADOL-C). */
        std::vector<CT> mMemoryVectGradUnknownDfsThree;

        /** Continuous memory lump #1 for storing pointers to rows of gradients 
         *  of all unknown data fields at all grid partition nodes (incl. halos) 
         *  \remark This memory lump will be allocated only if gradients should 
         *  be computed. ADOL-C needs variables of type double** for independent
         * and dependent gradients. */
        std::vector<CT*> mMemoryVectGradUnknownDfsRowsOne;

        /** Continuous memory lump #2 for storing pointers to rows of gradients 
         *  of all unknown data fields at all grid partition nodes (incl. halos) 
         *   \remark This memory lump will be allocated only if gradients should 
         *  be computed. ADOL-C needs variables of type double** for independent
         * and dependent gradients. */
        std::vector<CT*> mMemoryVectGradUnknownDfsRowsTwo;

        /** Temporary vector defined at the domain 
         * in order to realize asynchron. communication with ADOL-C. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectUnknownDfsDomTmp;

        /** Temporary vector defined at the boundary
         * in order to realize asynchron. communication with ADOL-C. */
        std::vector< ScaFES::DataField<CT,DIM> > mVectUnknownDfsBdryTmp;

        /** Temporary gradient in order to realize asynchron. communication
         * when using ADOL-C tapes. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectGradUnknownDfsDomTmp;

        /** Temporary gradient in order to realize asynchron. communication
         * when using ADOL-C tapes. */
        std::vector< ScaFES::DataField<CT, DIM> > mVectGradUnknownDfsBdryTmp;

        /** Vector containing current set of (geom.) parameters. */
        std::vector<CT> mVectParamsCurr;

        /** Vector containing gradient of current set of (geom.) parameters. */
        std::vector<CT> mVectGradParamsCurr;

        /** Pointer to memory for storing values of corresponding gradient. */
        std::vector<CT*> mMemoryVectGradParamsAsMatrix;

}; // End of class. //

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::Problem(
    Parameters const& params,
    GridGlobal<DIM> const& gg,
    bool const& useLeapfrog,
    std::vector<std::string> const& nameDatafield,
    std::vector<int> const& stencilWidth,
    std::vector<bool> const& isKnownDf,
    std::vector<int> const& nLayers,
    std::vector<CT> const& defaultValue,
    std::vector<ScaFES::WriteHowOften> const& writeToFile,
    std::vector<bool> const& computeError,
    std::vector<CT> const& geomparamsInit
)
    : mParams(params)
    , mUseLeapfrogIntegration(useLeapfrog)
    , mGG(gg)
    , mNameDataField(nameDatafield)
    , mStencilWidth(stencilWidth)
    , mIsKnownDf(isKnownDf)
    , mNlayers(nLayers)
    , mDefaultValue(defaultValue)
    , mWriteToFile(writeToFile)
    , mComputeError(computeError)
    , mNparams(1) // = 1 because gradients are of type R^{m,nParams}.
    , mComputeGradients(false)
    , mUseAsynchronMode(false)
    , mIsParamsBased(true) // Grid based ansatz currently not supported!
    , mTapeIdInitPartitionInner(0+params.myWorld().rank())
    , mTapeIdInitPartitionBorder(1592+params.myWorld().rank())
    , mTapeIdInitPartitionAll(3184+params.myWorld().rank())
    , mIsTracedInit(false)
    , mTapeIdUpdatePartitionInner(10922+params.myWorld().rank())
    , mTapeIdUpdatePartitionBorder(12922+params.myWorld().rank())
    , mTapeIdUpdatePartitionAll(14922+params.myWorld().rank())
    , mTapeIdUpdatePartitionInner2(16922+params.myWorld().rank())
    , mTapeIdUpdatePartitionBorder2(18922+params.myWorld().rank())
    , mTapeIdUpdatePartitionAll2(20922+params.myWorld().rank())
    , mIsTracedUpdate(false)
    , mNnodesTotalPartition(0)
    , mNelemsKnownDfs(0)
    , mNelemsUnknownDfs(0)
    , mNunknownBdryDfs(0)
    , mWallClockTimes()
    , mCommonFile()
    , mKind()
    , mMemoryKind()
    , mNodesLocInnerGlobInner()
    , mNodesLocInnerGlobBorder()
    , mNodesLocBorderGlobInner()
    , mNodesLocBorderGlobBorder()
    , mVectKnownDfDom()
    , mVectKnownDfBdry()
    , mMemoryVectKnownDf()
    , mVectUnknownDfsDomOld()
    , mVectUnknownDfsBdryOld()
    , mVectUnknownDfsDomNew()
    , mVectUnknownDfsBdryNew()
    , mMemoryVectUnknownDfsOne()
    , mMemoryVectUnknownDfsTwo()
    , mMemoryVectUnknownDfsThree()
    , mVectGradUnknownDfsDomOld()
    , mVectGradUnknownDfsBdryOld()
    , mVectGradUnknownDfsDomNew()
    , mVectGradUnknownDfsBdryNew()
    , mMemoryVectGradUnknownDfsOne()
    , mMemoryVectGradUnknownDfsTwo()
    , mMemoryVectGradUnknownDfsThree()
    , mMemoryVectGradUnknownDfsRowsOne()
    , mMemoryVectGradUnknownDfsRowsTwo()
    , mVectUnknownDfsDomTmp()
    , mVectUnknownDfsBdryTmp()
    , mVectGradUnknownDfsDomTmp()
    , mVectGradUnknownDfsBdryTmp()
    , mVectParamsCurr(geomparamsInit)
    , mVectGradParamsCurr()
    , mMemoryVectGradParamsAsMatrix()
{
#ifdef VTRACE
    VT_TRACER("ScaFES::Problem Ctor");
#endif
     this->mWallClockTimes["trace"] = 0.0;
     this->mWallClockTimes["write"] = 0.0; 
     this->mWallClockTimes["evalKnownDf"] = 0.0; 
     this->mWallClockTimes["sync"] = 0.0; 
     this->mWallClockTimes["update"] = 0.0; 
     this->mWallClockTimes["init"] = 0.0; 

    /*------------------------------------------------------------------------*/
    this->setCommunicationMode();
    this->setCompGradients();
    
    if (this->params().asynchronMode() 
         && this->params().enabledAdolc()  
         && this->computeGradients() ) { // AFTER calling setCompGradients().
        throw std::runtime_error("Async. MPI + ADOL-C + compGradients currently not supported!");
    }
    
    /*------------------------------------------------------------------------*/
    // Check sizes of parameters (basis: size of nameDatafield).
    // It suffices to check if one parameter has another size than the others.
    if (nameDatafield.size() != stencilWidth.size()) {
        throw std::runtime_error("Mismatch in sizes of parameters!");
    }
    if (nameDatafield.size() != isKnownDf.size()) {
        throw std::runtime_error("Mismatch in sizes of parameters!");
    }
    if (nameDatafield.size() != nLayers.size()) {
        throw std::runtime_error("Mismatch in sizes of parameters!");
    }
    if (nameDatafield.size() != defaultValue.size()) {
        throw std::runtime_error("Mismatch in sizes of parameters!");
    }
    if (nameDatafield.size() != writeToFile.size()) {
        throw std::runtime_error("Mismatch in sizes of parameters!");
    }
    if (nameDatafield.size() != computeError.size()) {
        throw std::runtime_error("Mismatch in sizes of parameters!");
    }

    /*------------------------------------------------------------------------*/
    // Determine number of unknown data fields.
    int nDataFields = stencilWidth.size();

    /*------------------------------------------------------------------------*/
    for (int ii = 0; ii < nDataFields; ++ii) {
        for (std::size_t jj = 0; jj < DIM; ++jj) {
            if (nLayers.at(ii) > gg.nNodes()[jj] / 2) {
                throw std::runtime_error("Number of layers does not fit to global grid!");
            }
        }
    }

    /*------------------------------------------------------------------------*/
    if (0 < geomparamsInit.size()) {
      this->mIsParamsBased = true;
      this->mVectGradParamsCurr.resize(
          geomparamsInit.size() * geomparamsInit.size(),
          static_cast<CT>(0)
      );
      this->mMemoryVectGradParamsAsMatrix.resize(
          geomparamsInit.size() * geomparamsInit.size(),
          static_cast<CT*>(0)
      );
    }
    
    /*------------------------------------------------------------------------*/
    if (this->isParamsBased()) {
        this->mNparams = 1; // Default = 1.
        if (0 < this->mVectParamsCurr.size()) {
            this->mNparams = this->mVectParamsCurr.size();
        }
    }

    /*------------------------------------------------------------------------*/
    // Determine number of total nodes of the grid partition.
    this->mNnodesTotalPartition
        = gg.partition(this->myRank()).nNodesSub().size();

    /*------------------------------------------------------------------------*/
    // Determine number of unknown data fields.
    int nUnknownDfs = 0;
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (!this->isKnownDf().at(ii)) {
            ++nUnknownDfs;
        }
    }

    /*------------------------------------------------------------------------*/
    // Determine number of unknown data fields which are defined at boundary.
    this->mNunknownBdryDfs = 0;
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (!(this->isKnownDf().at(ii))) {
            if (0 < nLayers.at(ii)) {
                ++this->mNunknownBdryDfs;
            }
        }
    }


    /*------------------------------------------------------------------------*/
    // Determine memAll and memNormal wrt. to the grid partition
    // for all data fields.
    std::vector< ScaFES::GridSub<DIM> > memNormal(nDataFields);
    std::vector< ScaFES::Grid<DIM> > memAll(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii) {
        ScaFES::Ntuple<int,DIM> idxNodeFirst
            = gg.partition(this->myRank()).idxNodeFirstSub()
              - ScaFES::Ntuple<int,DIM>(stencilWidth.at(ii));
        ScaFES::Ntuple<int,DIM> idxNodeLast
            = gg.partition(this->myRank()).idxNodeLastSub()
              + ScaFES::Ntuple<int,DIM>(stencilWidth.at(ii));
        ScaFES::Ntuple<int,DIM> idxNodeFirstNormal
            = gg.partition(this->myRank()).idxNodeFirstSub();
        ScaFES::Ntuple<int,DIM> idxNodeLastNormal
            = gg.partition(this->myRank()).idxNodeLastSub();
        memAll[ii] = ScaFES::Grid<DIM>(
            idxNodeFirst,
            idxNodeLast,
            gg.coordinates(idxNodeFirst),
            gg.coordinates(idxNodeLast)
        );
        memNormal[ii] = ScaFES::GridSub<DIM>(
            idxNodeFirstNormal,
            idxNodeLastNormal,
            memAll[ii]
        );
    }
    
    /*------------------------------------------------------------------------*/
    // Determine memHole wrt to global grid for all data fields.
    std::vector< ScaFES::GridSub<DIM> > memHoleGlobal(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii) {
        ScaFES::Ntuple<int,DIM> idxNodeFirst
            = gg.partition(this->myRank()).idxNodeFirstSub()
              - ScaFES::Ntuple<int,DIM>(stencilWidth.at(ii));
        ScaFES::Ntuple<int,DIM> idxNodeLast
            = gg.partition(this->myRank()).idxNodeLastSub()
              + ScaFES::Ntuple<int,DIM>(stencilWidth.at(ii));
        memHoleGlobal[ii] = ScaFES::GridSub<DIM>(
            gg.idxNodeFirst() + ScaFES::Ntuple<int,DIM>(nLayers.at(ii)),
            gg.idxNodeLast() - ScaFES::Ntuple<int,DIM>(nLayers.at(ii)),
            gg
        );
    }

    /*------------------------------------------------------------------------*/
    // Determine number of elements of all data fields.
    std::vector< ScaFES::GridSub<DIM> > memHolePart(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii) {
        memHolePart[ii] = ScaFES::GridSub<DIM>(
            ScaFES::max(memNormal.at(ii).idxNodeFirstSub(),
                        memHoleGlobal.at(ii).idxNodeFirstSub()),
            ScaFES::min(memNormal.at(ii).idxNodeLastSub(),
                        memHoleGlobal.at(ii).idxNodeLastSub()),
            memAll.at(ii)
        );
    }
    /*------------------------------------------------------------------------*/
    // Determine number of elements of all data fields (grid nodes incl. time).
    std::vector<int> nRowsDf(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (0 < nLayers.at(ii)) {
            nRowsDf[ii] = memAll[ii].nNodes().size() 
                           - memHolePart[ii].nNodesSub().size() + 1;
        }
        else {
            nRowsDf[ii] = memAll[ii].nNodes().size() + 1;
        }
    }

    /*------------------------------------------------------------------------*/
    // Determine number of elements of all unknown data fields.
    this->mNelemsUnknownDfs = 0;
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (!(this->isKnownDf().at(ii))) {
            this->mNelemsUnknownDfs += nRowsDf[ii];
        }
    }
    if (0 == this->nElemsUnknownDfs()) {
        throw std::runtime_error("#(unknown data fields) = 0.");
    }
    /*------------------------------------------------------------------------*/
    // Determine number of elements of all known data fields.
    this->mNelemsKnownDfs = 0;
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (this->isKnownDf().at(ii)) {
            this->mNelemsKnownDfs += nRowsDf[ii];
        }
    }
    
    /*------------------------------------------------------------------------*/
    // Check if reference vector for all UNKNOWN data fields should be created.
    bool createVectKnownDf = false;
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (isKnownDf.at(ii) || computeError.at(ii)) {
            createVectKnownDf = true;
            break;
        }
    }
    
    /*------------------------------------------------------------------------*/
    // Create known data fields only if solution is known and some error
    // should be computed.
    // Process all KNOWN data fields (including exact solutions of known dfs).
    // Create one large memory block for all KNOWN data fields.
    const int N_COLUMNS_KNOWN_DF = 1;
    int nElemsAllDfs = this->nElemsKnownDfs();
    if (createVectKnownDf) {
        nElemsAllDfs += this->nElemsUnknownDfs();
    }
    if (0 < nElemsAllDfs) {
        this->mMemoryVectKnownDf.resize(
            N_COLUMNS_KNOWN_DF * nElemsAllDfs ,
            static_cast<CT>(0)
        );
        this->mVectKnownDfDom.reserve(nDataFields);
        this->mVectKnownDfBdry.reserve(nDataFields);
    }
    CT* elemDataKnownDfCurr = this->mMemoryVectKnownDf.data();
    /*------------------------------------------------------------------------*/
    // First, all data fields which are defined at the whole domain.
    // Second, all data fields which are defined at the boundary, only.
    // This reordering of the data fields is important for the ADOL-C tracing
    // and evaluation as the fields must be available as a continuous memory
    // lump.
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (0 == nLayers.at(ii)) {
            if (this->isKnownDf().at(ii)) {
                this->mVectKnownDfDom.push_back(
                    ScaFES::DataField<CT, DIM>(
                        nameDatafield.at(ii),
                        this->params(),
                        this->globalGrid(),
                        memAll.at(ii),
                        stencilWidth.at(ii),
                        elemDataKnownDfCurr,
                        N_COLUMNS_KNOWN_DF,
                        memHolePart.at(ii),
                        nLayers.at(ii),
                        writeToFile.at(ii)
                    )
                );
                elemDataKnownDfCurr += nRowsDf.at(ii);
            }
            else {
                // Create reference data field regardless if error should be
                // computed or not. This important for the correct order of
                // the data fields.
                // BUT: Memory for reference data field will be created only in
                // case OF ERROR COMPUTATIONS.
                std::string nameDf = nameDatafield.at(ii) + "SolDom";
                this->mVectKnownDfDom.push_back(
                    ScaFES::DataField<CT, DIM>(
                            nameDf,
                            this->params(),
                            this->globalGrid(),
                            memAll.at(ii),
                            0,
                            elemDataKnownDfCurr,
                            N_COLUMNS_KNOWN_DF,
                            memHolePart.at(ii),
                            nLayers.at(ii),
                            writeToFile.at(ii)
                       )
                );
                if (computeError.at(ii)) {
                    elemDataKnownDfCurr += nRowsDf.at(ii);
                }
            }
        }
    }
    /*------------------------------------------------------------------------*/
    for (int ii = 0; ii < nDataFields; ++ii) {
        if (0 < nLayers.at(ii)) {
            if (this->isKnownDf().at(ii)) {
                this->mVectKnownDfBdry.push_back(
                    ScaFES::DataField<CT, DIM>(
                        nameDatafield.at(ii),
                        this->params(),
                        this->globalGrid(),
                        memAll.at(ii),
                        stencilWidth.at(ii),
                        elemDataKnownDfCurr,
                        N_COLUMNS_KNOWN_DF,
                        memHolePart.at(ii),
                        nLayers.at(ii),
                        writeToFile.at(ii)
                    )
                );
                elemDataKnownDfCurr += nRowsDf.at(ii);
            }
            else {
                // Create reference data field regardless if error should be
                // computed or not. This important for the correct order of
                // the data fields.
                // BUT: Memory for reference data field will be created only in
                // case OF COMPUTING THE ERROR.
                std::string nameDf = nameDatafield.at(ii) + "SolBdry";
                this->mVectKnownDfBdry.push_back(
                    ScaFES::DataField<CT, DIM>(
                        nameDf,
                        this->params(),
                        this->globalGrid(),
                        memAll.at(ii),
                        0,
                        elemDataKnownDfCurr,
                        N_COLUMNS_KNOWN_DF,
                        memHolePart.at(ii),
                        nLayers.at(ii),
                        writeToFile.at(ii)
                   )
                );
                if (computeError.at(ii)) {
                    elemDataKnownDfCurr += nRowsDf.at(ii);
                }
            }
        }
    }
    /*------------------------------------------------------------------------*/
    // Set default value for each known data field.
// //     int idDf = 0;
// //     for (int ii = 0; ii < nDataFields; ++ii) {
// //         if (0 == nLayers.at(ii)) {
// //             if (this->isKnownDf().at(ii)) {
// //                 this->mVectKnownDfDom.at(idDf) = defaultValue.at(ii);
// //                 ++idDf;
// //             }
// //         }
// //     }
// //     idDf = 0;
// //     for (int ii = 0; ii < nDataFields; ++ii) {
// //         if (0 < nLayers.at(ii)) {
// //             if (this->isKnownDf().at(ii)) {
// //                 this->mVectKnownDfBdry.at(idDf) = defaultValue.at(ii);
// //                 ++idDf;
// //             }
// //         }
// //     }

    /*------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------*/
    // Process all UNKNOWN data fields.
    // Create one large memory block for all UNKNOWN data fields.
    // Use old / new scheme.
    // First, allocate memory for all old data fields (independent ones),
    // then, allocate memory for all new data fields (dependent ones).
    const int N_COLUMNS_VECTUNKNOWN = 1;
    if (0 < this->nElemsUnknownDfs()) {
        this->mMemoryVectUnknownDfsOne.resize(
            N_COLUMNS_VECTUNKNOWN * this->nElemsUnknownDfs(),
            static_cast<CT>(0)
        );
    }
    if (0 < this->nElemsUnknownDfs()) {
        this->mMemoryVectUnknownDfsTwo.resize(
            N_COLUMNS_VECTUNKNOWN * this->nElemsUnknownDfs(),
            static_cast<CT>(0)
        );
    }
    /*------------------------------------------------------------------------*/
    std::string nameSuffixOld = "Old";
    this->createVectorOfUnknownDataFields(
        this->mVectUnknownDfsDomOld,
        this->mVectUnknownDfsBdryOld,
        this->mMemoryVectUnknownDfsOne,
        nameDatafield,
        nameSuffixOld,
        isKnownDf,
        nLayers,
        stencilWidth,
        memAll,
        memHolePart,
        nRowsDf,
        writeToFile,
        defaultValue,
        N_COLUMNS_VECTUNKNOWN
    );
    /*------------------------------------------------------------------------*/
    std::string nameSuffixNew = "New";
    this->createVectorOfUnknownDataFields(
        this->mVectUnknownDfsDomNew,
        this->mVectUnknownDfsBdryNew,
        this->mMemoryVectUnknownDfsTwo,
        nameDatafield,
        nameSuffixNew,
        isKnownDf,
        nLayers,
        stencilWidth,
        memAll,
        memHolePart,
        nRowsDf,
        writeToFile,
        defaultValue,
        N_COLUMNS_VECTUNKNOWN
    );
    /*------------------------------------------------------------------------*/
    // Create temporary vector only in case of
    // ADOL-C + asynchronously MPI communication.
    if ( params.enabledAdolc() && this->useAsynchronMode() ) {
        if (0 < this->nElemsUnknownDfs()) {
            this->mMemoryVectUnknownDfsThree.resize(
                N_COLUMNS_VECTUNKNOWN * this->nElemsUnknownDfs(),
                static_cast<CT>(0)
            );
        }
        std::string nameSuffixTmp = "Tmp";
        this->createVectorOfUnknownDataFields(
            this->mVectUnknownDfsDomTmp,
            this->mVectUnknownDfsBdryTmp,
            this->mMemoryVectUnknownDfsThree,
            nameDatafield,
            nameSuffixTmp,
            isKnownDf,
            nLayers,
            stencilWidth,
            memAll,
            memHolePart,
            nRowsDf,
            writeToFile,
            defaultValue,
            N_COLUMNS_VECTUNKNOWN
        );
    }

    /*------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------*/
    if (this->computeGradients()) {
        // Process GRADIENTS of all UNKNOWN data fields.
        // Use old / new scheme.
        // Gradients are \in R^{nxm}:
        // Nevertheless, create an one-dimensional array:
        // +++ one continuous memory lump
        // +++ Methods of class DataField can be used, 
        //     if pointer to one-dimensional
        //     array will be passed to the constructor of the class DataField.
        const int N_COLUMNS_VECTGRADUNKNOWN = this->nColumns();
        if (0 < this->nElemsUnknownDfs()) {
            this->mMemoryVectGradUnknownDfsOne.resize(
                N_COLUMNS_VECTGRADUNKNOWN * this->nElemsUnknownDfs(),
                static_cast<CT>(0)
            );
        }
        if (0 < this->nElemsUnknownDfs()) {
            this->mMemoryVectGradUnknownDfsTwo.resize(
                (N_COLUMNS_VECTGRADUNKNOWN * (this->nElemsUnknownDfs())),
                static_cast<CT>(0)
            );
        }
        /*--------------------------------------------------------------------*/
        std::string nameSuffixGradOld = "GradOld";
        this->createVectorOfUnknownDataFields(
            this->mVectGradUnknownDfsDomOld,
            this->mVectGradUnknownDfsBdryOld,
            this->mMemoryVectGradUnknownDfsOne,
            nameDatafield,
            nameSuffixGradOld,
            isKnownDf,
            nLayers,
            stencilWidth,
            memAll,
            memHolePart,
            nRowsDf,
            writeToFile,
            defaultValue,
            N_COLUMNS_VECTGRADUNKNOWN
        );
        /*------------------------------------------------------------------------*/
        std::string nameSuffixGradNew = "GradNew";
        this->createVectorOfUnknownDataFields(
            this->mVectGradUnknownDfsDomNew,
            this->mVectGradUnknownDfsBdryNew,
            this->mMemoryVectGradUnknownDfsTwo,
            nameDatafield,
            nameSuffixGradNew,
            isKnownDf,
            nLayers,
            stencilWidth,
            memAll,
            memHolePart,
            nRowsDf,
            writeToFile,
            defaultValue,
            N_COLUMNS_VECTGRADUNKNOWN
        );
        /*------------------------------------------------------------------------*/
        // Create temporary vector of data fields only in case of
        // ADOL-C + asynchronously MPI communication.
        if (params.enabledAdolc() && this->useAsynchronMode()) {
            if (0 < this->nElemsUnknownDfs()) {
                this->mMemoryVectGradUnknownDfsThree.resize(
                    ( N_COLUMNS_VECTGRADUNKNOWN * (this->nElemsUnknownDfs()) ),
                    static_cast<CT>(0)
                );
            }
            std::string nameSuffixGradTmp = "GradTmp";
            this->createVectorOfUnknownDataFields(
                this->mVectGradUnknownDfsDomTmp,
                this->mVectGradUnknownDfsBdryTmp,
                this->mMemoryVectGradUnknownDfsThree,
                nameDatafield,
                nameSuffixGradTmp,
                isKnownDf,
                nLayers,
                stencilWidth,
                memAll,
                memHolePart,
                nRowsDf,
                writeToFile,
                defaultValue,
                N_COLUMNS_VECTGRADUNKNOWN
            );
        }
    }
 
    /*------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------*/
    // Create memory for pointers only if gradients should be computed 
    // using ADOL-C.
    if (params.enabledAdolc() && this->computeGradients()) {
        if (0 < this->nElemsUnknownDfs()) {
            this->mMemoryVectGradUnknownDfsRowsOne.resize(
                1 * this->nElemsUnknownDfs(),
                static_cast<CT*>(0)
            );
            this->mMemoryVectGradUnknownDfsRowsTwo.resize(
                1 * this->nElemsUnknownDfs(),
                static_cast<CT*>(0)
            );
        }
    }
    
    /*------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------*/
    int nDataFieldsToWrite = 0;
    for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
        if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
            ++nDataFieldsToWrite;
        }
    }

    /*------------------------------------------------------------------------*/
    std::vector<std::string> tmpNames;
    tmpNames.reserve(nDataFieldsToWrite);
    int idxKnownDf = 0;
    int idxUnknownDf = 0;
    for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
        if (0 == nLayers.at(ii)) {
            if (this->isKnownDf().at(ii)) {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpNames.push_back(this->mVectKnownDfDom[idxKnownDf].name());
                }
                ++idxKnownDf;
            }
            else {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpNames.push_back(this->vectUnknownDfsDomNew()[idxUnknownDf].name());
                }
                ++idxUnknownDf;
            }
        }
    }
    idxKnownDf = 0;
    idxUnknownDf = 0;
    for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
        if (0 < nLayers.at(ii)) {
            if (this->isKnownDf().at(ii)) {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpNames.push_back(this->mVectKnownDfBdry[idxKnownDf].name());
                }
                ++idxKnownDf;
            }
            else {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpNames.push_back(this->vectUnknownDfsBdryNew()[idxUnknownDf].name());
                }
                ++idxUnknownDf;
            }
        }
    }

    /*------------------------------------------------------------------------*/
    std::vector< ScaFES::GridSub<DIM> > tmpMemNormal;
    tmpMemNormal.reserve(nDataFieldsToWrite);
    idxKnownDf = 0;
    idxUnknownDf = 0;
    for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
        if (0 == nLayers.at(ii)) {
            if (this->isKnownDf().at(ii)) {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpMemNormal.push_back(this->mVectKnownDfDom[idxKnownDf].memNormal());
                }
                ++idxKnownDf;
            }
            else {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpMemNormal.push_back(this->vectUnknownDfsDomNew()[idxUnknownDf].memNormal());
                }
                ++idxUnknownDf;
            }
        }
    }
    idxKnownDf = 0;
    idxUnknownDf = 0;
    for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
        if (0 < nLayers.at(ii)) {
            if (this->isKnownDf().at(ii)) {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpMemNormal.push_back(this->mVectKnownDfBdry[idxKnownDf].memNormal());
                }
                ++idxKnownDf;
            }
            else {
                if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                    tmpMemNormal.push_back(this->vectUnknownDfsBdryNew()[idxUnknownDf].memNormal());
                }
                ++idxUnknownDf;
            }
        }
    }

    std::ostringstream tmpStringstream;
    tmpStringstream << this->params().nameDataFile();

    this->mCommonFile = ScaFES::DataFile<CT, DIM>(
        tmpStringstream.str(),
        tmpNames,
        this->params().myWorld(),
        this->globalGrid().nNodes(),
        tmpMemNormal
    );

    /*------------------------------------------------------------------------*/
    // Read in partition file.
    if (this->params().readPartitionFile()) {
        try {
            std::ifstream ifs(this->params().nameReadPartitionFile().c_str());
            boost::archive::binary_iarchive ia(ifs);
            ia >> (this->mGG);

            // The partition file is okay.
            // Check if number of partitions == number of MPI procs.
            if (this->params().myWorld().size() 
                != this->globalGrid().nPartitions()) {
                throw std::runtime_error("Number of partitions != mpiSize().");
            }
        } catch (std::exception& e) {
            throw std::runtime_error("Could not open partition file.");
        }
    }

    /*------------------------------------------------------------------------*/
    // Serialize the global grid into an archive and write file.
    if (this->params().writePartitionFile()) {
        std::ostringstream tmpStringstream;
        tmpStringstream << this->params().nameWritePartitionFile()
                        << ".partfile";
        std::string tmpStr = tmpStringstream.str();
        std::ofstream ofs(tmpStr.c_str());
        boost::archive::binary_oarchive oa(ofs);

        try {
            oa << (this->mGG);
        } catch (...) {
            throw std::runtime_error("Global grid could be not be serialized.");
        }

        ofs.close();
    }

    /*------------------------------------------------------------------------*/
    ScaFES::Ntuple<int,DIM> idxNodeFirstKindAll
            = gg.partition(this->myRank()).idxNodeFirstSub();
    ScaFES::Ntuple<int,DIM> idxNodeLastKindAll
            = gg.partition(this->myRank()).idxNodeLastSub();
    ScaFES::Ntuple<int,DIM> idxNodeFirstKindNormal
            = gg.partition(this->myRank()).idxNodeFirstSub();
    ScaFES::Ntuple<int,DIM> idxNodeLastKindNormal
            = gg.partition(this->myRank()).idxNodeLastSub();
    ScaFES::Grid<DIM> memAllKind(
        idxNodeFirstKindAll,
        idxNodeLastKindAll,
        gg.coordinates(idxNodeFirstKindAll),
        gg.coordinates(idxNodeLastKindAll)
    );
    ScaFES::GridSub<DIM> memNormalKind(
            idxNodeFirstKindNormal,
            idxNodeLastKindNormal,
            memAllKind
    );
    ScaFES::GridSub<DIM> memHolePartKind(
        memNormalKind.idxNodeFirstSub(),
        memNormalKind.idxNodeLastSub(),
        memAllKind
    );

    if (0 < this->nNodesTotalPartition()) {
        this->mMemoryKind.resize(
            this->nNodesTotalPartition() + 1, // incl. time
            static_cast<short int>(0)
        );
    }
    short int* elemDataKind = this->mMemoryKind.data();
    ScaFES::WriteHowOften writeKind = ScaFES::WriteHowOften::NEVER;
    if (this->params().writeKindFile()) {
        writeKind = ScaFES::WriteHowOften::LIKE_GIVEN_AT_CL;
    }
    this->mKind = ScaFES::DataField<short int,DIM>(
        "kind",
        this->params(),
        this->globalGrid(),
        memNormalKind,                                                  
        0,
        elemDataKind,
        1,
        memHolePartKind,
        0,
        writeKind
    );

     /*------------------------------------------------------------------------*/
    // Read in a given kind file?
    if (this->params().readKindFile()) {
        this->mKind.read(0);
    }
    else {
        this->setGridGlobalFlags();
    }
    int maxStencilWidth = 0;
    for (std::size_t ii = 0; ii < this->vectUnknownDfsDomNew().size(); ++ii) {
        if (maxStencilWidth < this->vectUnknownDfsDomNew()[ii].stencilWidth()) {
            maxStencilWidth = this->vectUnknownDfsDomNew()[ii].stencilWidth();
        }
    }
    for (std::size_t ii = 0; ii < this->vectUnknownDfsBdryNew().size(); ++ii) {
        if (maxStencilWidth < this->vectUnknownDfsBdryNew()[ii].stencilWidth()) {
            maxStencilWidth = this->vectUnknownDfsBdryNew()[ii].stencilWidth();
        }
    }
    this->setGridPartitionFlags(maxStencilWidth);

    /*------------------------------------------------------------------------*/
    // Determine node list of all global inner and all global border nodes.
    this->mNodesLocInnerGlobInner.resize(
        this->getNnodesLocInnerGlobInner()
    );
    int nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_INNER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_INNER_NODE) {
                this->mNodesLocInnerGlobInner[nNodes] = ii;
                ++nNodes;
            }
        }
    }
    this->mNodesLocInnerGlobBorder.resize(
        this->getNnodesLocInnerGlobBorder()
    );
    nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_INNER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_BORDER_NODE) {
                this->mNodesLocInnerGlobBorder[nNodes] = ii;
                ++nNodes;
            }
        }
    }
    this->mNodesLocBorderGlobInner.resize(
        this->getNnodesLocBorderGlobInner()
    );
    nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_BORDER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_INNER_NODE) {
                this->mNodesLocBorderGlobInner[nNodes] = ii;
                ++nNodes;
            }
        }
    }
    this->mNodesLocBorderGlobBorder.resize(
        this->getNnodesLocBorderGlobBorder()
    );
    nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_BORDER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_BORDER_NODE) {
                this->mNodesLocBorderGlobBorder[nNodes] = ii;
                ++nNodes;
            }
        }
    }
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::createVectorOfUnknownDataFields(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDomain,
    std::vector< ScaFES::DataField<CT,DIM> >& dfBdry,
    std::vector<CT> const& memoryOfDfs,
    std::vector<std::string> const& nameDatafield,
    std::string const& nameSuffix,
    std::vector<bool> const& isKnownDf,
    std::vector<int> const& nLayers,
    std::vector<int> const& stencilWidth,
    std::vector< ScaFES::Grid<DIM> > const& memAll,
    std::vector< ScaFES::GridSub<DIM> > const& memHolePart,
    std::vector<int> const& nRowsDf,
    std::vector<ScaFES::WriteHowOften> const& writeToFile,
    std::vector<CT> const& defaultValue,
    int const& nColumns
)
{
    CT* elemDataCurr = const_cast<CT*>(memoryOfDfs.data());
    int idDf = 0;

    for (std::size_t ii = 0; ii < nLayers.size(); ++ii) {
        if (!(isKnownDf.at(ii))) {
            if (0 == nLayers.at(ii)) {
                std::string nameDf = nameDatafield.at(ii) + nameSuffix + "Dom";
                dfDomain.push_back(
                    ScaFES::DataField<CT, DIM>(
                        nameDf,
                        this->params(),
                        this->globalGrid(),
                        memAll.at(ii),
                        stencilWidth.at(ii),
                        elemDataCurr,
                        nColumns,
                        memHolePart.at(ii),
                        nLayers.at(ii),
                        writeToFile.at(ii)
                    )
                );
                elemDataCurr += (nRowsDf[ii]  * nColumns);
                ++idDf;
            }
        }
    }

    idDf = 0;
    for (std::size_t ii = 0; ii < nLayers.size(); ++ii) {
        if (!isKnownDf.at(ii)) {
            if (0 < nLayers.at(ii)) {
                std::string nameDf = nameDatafield.at(ii) + nameSuffix + "Bdry";
                dfBdry.push_back(
                    ScaFES::DataField<CT, DIM>(
                        nameDf,
                        this->params(),
                        this->globalGrid(),
                        memAll.at(ii),
                        stencilWidth.at(ii),
                        elemDataCurr,
                        nColumns,
                        memHolePart.at(ii),
                        nLayers.at(ii),
                        writeToFile.at(ii)
                    )
                );
                elemDataCurr += (nRowsDf[ii] * nColumns);
                ++idDf;
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Set default value for each unknown data field.
    idDf = 0;
    for (std::size_t ii = 0; ii < nLayers.size(); ++ii) {
        if (0 == nLayers.at(ii)) {
            if (!(isKnownDf.at(ii))) {
                dfDomain.at(idDf) = defaultValue.at(ii);
                ++idDf;
            }
        }
    }
    idDf = 0;
    for (std::size_t ii = 0; ii < nLayers.size(); ++ii) {
        if (0 < nLayers.at(ii)) {
            if (!(isKnownDf.at(ii))) {
                dfBdry.at(idDf) = defaultValue.at(ii);
                ++idDf;
            }
        }
    }
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline Parameters const& Problem<OWNPRBLM, CT, DIM>::params() const
{
    return this->mParams;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nTimesteps() const
{
    return this->params().nTimesteps();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline double const& Problem<OWNPRBLM, CT, DIM>::gridsize(int const& idx) const
{
    return (this->mGG.gridsize(idx));
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline double const& Problem<OWNPRBLM, CT, DIM>::tau() const
{
    return this->params().tau();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline GridGlobal<DIM> const& Problem<OWNPRBLM, CT, DIM>::globalGrid() const
{
    return this->mGG;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& Problem<OWNPRBLM, CT, DIM>::kind(ScaFES::Ntuple<int,DIM> const& idxNode) const
{
    return this->mKind[idxNode];
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline ScaFES::DataField<short int, DIM> const& Problem<OWNPRBLM, CT, DIM>::kind() const
{
    return this->mKind;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline ScaFES::Ntuple<int,DIM> const& Problem<OWNPRBLM, CT, DIM>::nNodes() const
{
    return this->globalGrid().nNodes();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nNodes(int const& dim) const
{
    return this->nNodes().elem(dim);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline ScaFES::Ntuple<double,DIM> Problem<OWNPRBLM, CT, DIM>::coordinates(
    ScaFES::Ntuple<int,DIM> const& idxNode
) const
{
    return this->globalGrid().coordinates(idxNode);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline ScaFES::Ntuple<int,DIM> Problem<OWNPRBLM, CT, DIM>::connect(
    ScaFES::Ntuple<int,DIM> const& idxNode,
    int const& dir
)
{
    return this->mGG.neigh(idxNode, dir);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<CT> const& Problem<OWNPRBLM, CT, DIM>::vectParamsCurr() const
{
    return this->mVectParamsCurr;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline CT const& Problem<OWNPRBLM, CT, DIM>::vectSol(
    int const& idx,
    ScaFES::Ntuple<int,DIM> const& idxNode
) const
{
    return this->mVectKnownDfDom[idx](idxNode);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline CT const& Problem<OWNPRBLM, CT, DIM>::knownDf(
    int const& idx,
    ScaFES::Ntuple<int,DIM> const& idxNode
) const
{
    return this->mVectKnownDfDom[idx](idxNode);
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::initUnknownDfs()
{
   if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise all data fields..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    int currTimestep = 0;

    // Remark: Initialise "New" data fields because memory of "New" and "Old"
    // data fields will be swapped in each time step.
    if (this->useAsynchronMode()) {
        this->initAsynchronously(
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperInitPartitionInnerPb,
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperInitPartitionBorderPb,
            this->mVectUnknownDfsDomNew,
            this->mVectGradUnknownDfsDomNew,
            this->vectParamsCurr(),
            this->mVectGradParamsCurr,
            currTimestep
        );
    }
    else {
        this->initSynchronously(
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperInitPartitionAllPb,
            this->mVectUnknownDfsDomNew,
            this->mVectGradUnknownDfsDomNew,
            this->mVectUnknownDfsBdryNew,
            this->mVectGradUnknownDfsBdryNew,
            this->vectParamsCurr(),
            this->mVectGradParamsCurr,
            currTimestep
        );
    }

    for (std::size_t ii = 0; ii < this->mVectUnknownDfsDomNew.size(); ++ii) {
        this->mVectUnknownDfsDomOld[ii].setValuesAtMemAll(this->mVectUnknownDfsDomNew[ii]);
    }
    for (std::size_t ii = 0; ii < this->mVectUnknownDfsBdryNew.size(); ++ii) {
        this->mVectUnknownDfsBdryOld[ii].setValuesAtMemAll(this->mVectUnknownDfsBdryNew[ii]);
    }

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::setCommunicationMode()
{
    this->mUseAsynchronMode = this->params().asynchronMode();
    if (0 < this->nUnknownBdryDfs() && this->mUseAsynchronMode) {
        this->mUseAsynchronMode = false; // Force synchronous MPI communication.
        if (this->params().rankOutput() == this->myRank()) {
            std::cout << "\n *** WARNING: There exists at least one border df!"
                      << "===> Synchronous MPI communication must be used! ***\n"
                      << std::endl;
        }
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::setCompGradients()
{
    this->mComputeGradients = this->params().computeGradients();
    if (!(this->params().enabledAdolc()) && this->mComputeGradients) {
        this->mComputeGradients = false;
        if (this->params().rankOutput() == this->myRank()) {
            std::cout << "\n *** WARNING: ADOL-C is not enabled!"
                      << "===> Gradients can not be computed! ***\n"
                      << std::endl;
        }
    }
}

/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::initAsynchronously(
    TypeFuncInitToEval funcLocalInner,
    TypeFuncInitToEval funcLocalBorder,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector<CT> const& dfIndep,
    std::vector<CT> const& dfGradIndep,
    const int& timestep
)
{
    boost::mpi::timer timerSync;

    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(pb) asynchronously data fields and its gradients "
                  << dfDep.at(0).name() << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    /*------------------------------------------------------------------------*/
    (this->*funcLocalBorder)(
        dfDep,
        dfGradDep,
        dfIndep,
        dfGradIndep,
        timestep
    );
    /*------------------------------------------------------------------------*/
    timerSync.restart();
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].copyValuesFromMemCommToSendBuffer(timestep);
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].copyValuesFromMemCommToSendBuffer(timestep);
    }
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].exchangeValuesInBuffers(timestep);
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].exchangeValuesInBuffers(timestep);
    }
    this->mWallClockTimes["sync"] += timerSync.elapsed();
    /*------------------------------------------------------------------------*/
    if (this->params().enabledAdolc()) {
        (this->*funcLocalInner)(
            this->mVectUnknownDfsDomTmp,
            this->mVectGradUnknownDfsDomTmp,
            dfIndep,
            dfGradIndep,
            timestep
        );
        for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
            dfDep[ii] += this->mVectUnknownDfsDomTmp[ii];
            dfDep[ii] -= this->defaultValue()[ii];
        }
        for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
            dfGradDep[ii] += this->mVectGradUnknownDfsDomTmp[ii]; // - 0;
        }
    }
    else {
        (this->*funcLocalInner)(
            dfDep,
            dfGradDep,
            dfIndep,
            dfGradIndep,
            timestep
        );
    }
    /*------------------------------------------------------------------------*/
    timerSync.restart();
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].waitAll();
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].waitAll();
    }
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].copyValuesFromReceiveBufferToMemGhost(timestep);
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].copyValuesFromReceiveBufferToMemGhost(timestep);
    }
    this->mWallClockTimes["sync"] += timerSync.elapsed();

    /*------------------------------------------------------------------------*/
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::initSynchronously(
    TypeFuncInitToEval funcGlobal,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepDomain,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradDomain,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepBdry,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradBdry,
    std::vector<CT> const& dfIndep,
    std::vector<CT> const& dfGradIndep,
    const int& timestep
)
{
    boost::mpi::timer timerSync;

    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(pb) synchronously data fields "
                  << dfDepDomain.at(0).name() << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    std::vector< ScaFES::DataField<CT,DIM> > vectAllTmp(dfDepDomain);
    for (std::size_t ii = 0; ii < dfDepBdry.size(); ++ii) {
        vectAllTmp.push_back(dfDepBdry[ii]);
    }
    std::vector< ScaFES::DataField<CT,DIM> > vectGradAllTmp(dfDepGradDomain);
    for (std::size_t ii = 0; ii < dfDepGradBdry.size(); ++ii) {
        vectGradAllTmp.push_back(dfDepGradBdry[ii]);
    }
    (this->*funcGlobal)(
        vectAllTmp,
        vectGradAllTmp,
        dfIndep,
        dfGradIndep,
        timestep
    );
    /*------------------------------------------------------------------------*/
    timerSync.restart();
    for (std::size_t ii = 0; ii < dfDepDomain.size(); ++ii) {
        dfDepDomain[ii].sync(timestep);
    }
    for (std::size_t ii = 0; ii < dfDepBdry.size(); ++ii) {
        dfDepBdry[ii].sync(timestep);
    }
    this->mWallClockTimes["sync"] += timerSync.elapsed();
    /*------------------------------------------------------------------------*/
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperInitPartitionInnerPb(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector<CT> const& dfIndep,
    std::vector<CT> const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        if (this->computeGradients()) {
            this->init_fovForward_paramsBased(
                dfDep,
                dfGradDep,
                this->tapeIdInitPartitionInner(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->init_zosForward_paramsBased(
                dfDep,
                this->tapeIdInitPartitionInner(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerInit;
        timerInit.restart();
        this->template initPartitionInnerPb<CT>(dfDep, dfIndep, timestep);
        this->mWallClockTimes["init"] += timerInit.elapsed();
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperInitPartitionBorderPb(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector<CT> const& dfIndep,
    std::vector<CT> const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        if (this->computeGradients()) {
            this->init_fovForward_paramsBased(
                dfDep,
                dfGradDep,
                this->tapeIdInitPartitionBorder(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->init_zosForward_paramsBased(
                dfDep,
                this->tapeIdInitPartitionBorder(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerInit;
        timerInit.restart();
        this->template initPartitionBorderPb<CT>(dfDep, dfIndep, timestep);
        this->mWallClockTimes["init"] += timerInit.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperInitPartitionAllPb(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector<CT> const& dfIndep,
    std::vector<CT> const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        if (this->computeGradients()) {
            this->init_fovForward_paramsBased(
                dfDep,
                dfGradDep,
                this->tapeIdInitPartitionAll(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->init_zosForward_paramsBased(
                dfDep,
                this->tapeIdInitPartitionAll(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerInit;
        timerInit.restart();
        this->template initPartitionAllBorderPb<CT>(dfDep, dfIndep, timestep);
        this->template initPartitionAllInnerPb<CT>(dfDep, dfIndep, timestep);
        this->mWallClockTimes["init"] += timerInit.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initPartitionInnerPb(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector<TT> const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(pb) dfs at all local inner nodes..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    std::size_t ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initBorder(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initPartitionBorderPb(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector<TT> const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(pb) dfs at all local border nodes..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    std::size_t ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initBorder(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initPartitionAllInnerPb(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector<TT> const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(pb) dfs at all global inner nodes..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    std::size_t ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initPartitionAllBorderPb(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector<TT> const& dfIndep,
    const int& timestep
)
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(pb) dfs at all border nodes..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    std::size_t ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initBorder(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfIndep, dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->initInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initGridBased(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<TT,DIM> > const& dfIndep
)
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(gb) dfs at all nodes..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    const int nNodesTotalPartition = this->kind().memNormal().nNodesSub().size();
    int ii;
    const int timestep = 0;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
// #ifdef _OPENMP
// #pragma omp parallel for
//     schedule(static)
//     default(none)
//     private(ii, idxNode, comp)
//     shared(dfIndep, dfDep, timestep, nNodesTotalPartition)
// #endif
    for (ii = 0; ii < nNodesTotalPartition; ++ii) {
        this->kind().memoryPos2IdxNode(idxNode, comp, ii);
        static_cast<OWNPRBLM*>(this)->template initialise<TT>(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }
}

/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::traceInitParamsBased(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector<CT> const& dfIndep
)
{
    if (this->isTracedInit()) {
        return;
    }

    if (!(this->params().enabledAdolc())) {
        throw std::runtime_error("Tracing not possible. ADOL-C is not enabled.");
    }


    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Create tapes"
                  << " for tracing initialise(" << dfDep.at(0).name()
                  << ",...)."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    // Measuring time can be done here because initialization phase will
    // be traced only once.
    boost::mpi::timer timerTrace;
    timerTrace.restart();
    int const timestep = 0;

#ifdef SCAFES_HAVE_ADOLC
    // REMARK: Currently, tracing can only done using 1 OpenMP thread.
#ifdef _OPENMP
    const int NTHREADS_ENV = omp_get_max_threads();
    omp_set_num_threads(1);
#endif

    ScaFES::Ntuple<int, DIM> idxNode;
    unsigned long int nElemsDfDepTotal = 0;
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        nElemsDfDepTotal += dfDep.at(ii).nElemsAllocated();
    }

    /*-------------------------------------------------&-----------------------*/
    std::vector<adouble> aDfIndep;
    aDfIndep.reserve(dfIndep.size());
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        aDfIndep.push_back( dfIndep.at(ii) );
    }

    /*------------------------------------------------------------------------*/
    std::vector<adouble> memoryAdfs;
    if (0 < nElemsDfDepTotal) {
        memoryAdfs.resize(nElemsDfDepTotal);
    }

    /*------------------------------------------------------------------------*/
    adouble* elemDataCurr = memoryAdfs.data();
    std::vector< ScaFES::DataField<adouble,DIM> > aDfDep;
    aDfDep.reserve(dfDep.size());
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        aDfDep.push_back(
            ScaFES::DataField<adouble, DIM>(
                dfDep.at(ii).name(),
                dfDep.at(ii).params(),
                dfDep.at(ii).gridGlobal(),
                dfDep.at(ii).memAll(),
                dfDep.at(ii).stencilWidth(),
                elemDataCurr,
                1,
                dfDep.at(ii).memHole(),
                dfDep.at(ii).nLayers()
            )
        );
        elemDataCurr += dfDep.at(ii).nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------*/
    // * asynchron: 2 tapes: 1 tape (localBorder) and 1 tape (localInner)
    // * synchron:  1 tape:  1 tape (globalInner + globalBorder)
    if (this->useAsynchronMode()) {
        this->createTapesForInit(
            this->tapeIdInitPartitionBorder(),
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template initPartitionBorderPb<adouble>,
            dfDep,
            dfIndep,
            aDfDep,
            aDfIndep,
            timestep
        );
        this->createTapesForInit(
            this->tapeIdInitPartitionInner(),
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template initPartitionInnerPb<adouble>,
            dfDep,
            dfIndep,
            aDfDep,
            aDfIndep,
            timestep
        );
    }
    else {
        this->createTapeForInitPartitionAll(
            this->tapeIdInitPartitionAll(),
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template initPartitionAllInnerPb<adouble>,
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template initPartitionAllBorderPb<adouble>,
            dfDep,
            dfIndep,
            aDfDep,
            aDfIndep,
            timestep
        );
    }

#ifdef _OPENMP
    omp_set_num_threads(NTHREADS_ENV);
#endif

#endif // HAVE_ADOLC
    
    this->mWallClockTimes["trace"] += timerTrace.elapsed();
    this->mIsTracedInit = true;

    /*------------------------------------------------------------------------*/
    // REMARK: No synchronization is necessary as one just wants to trace
    // the mathematical operations.
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Created."
                  << std::endl;
    }
}

/*----------------------------------------------------------------------------*/
#ifdef SCAFES_HAVE_ADOLC
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::createTapesForInit(
    short int const& tapeId,
    TypeFuncInitToTrace funcInit,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< CT > const& dfIndep,
    std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
    std::vector< adouble >& aDfIndep,
    int const& timestep
)
{
    ScaFES::Ntuple<int, DIM> idxNode;
    /*------------------------------------------------------------------------*/
    trace_on(tapeId, 1);
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        aDfIndep[ii] <<= dfIndep.at(ii);
    }
    
    // Remark:
    // Reset vector of dependent variables to default values.
    // Just is due to the asynchronous MPI communication and the
    // introduction of a temporary vector.
    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = aDfDep.at(ii).memAll().begin(),
             et = aDfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) = this->defaultValue().at(ii);
        }
    }

    // Set t = t_S.
    // The rhs need not to be an independent variable the time at the
    // initialization will always be t=t_S.
    // ==> Variable of type double for rhs is sufficient.
    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        aDfDep[ii].time() = this->params().timeIntervalStart();
    }

    (this->*funcInit)(aDfDep, aDfIndep, timestep);

    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfDep.at(ii).memAll().begin(),
            et = dfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) >>= dfDep.at(ii)(idxNode);
        }
        aDfDep[ii].time() >>= dfDep.at(ii).time();
    }
    trace_off();
}
#endif
/*----------------------------------------------------------------------------*/
#ifdef SCAFES_HAVE_ADOLC
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::createTapeForInitPartitionAll(
    short int const& tapeId,
    TypeFuncInitToTrace funcInitInner,
    TypeFuncInitToTrace funcInitBorder,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< CT > const& dfIndep,
    std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
    std::vector< adouble >& aDfIndep,
    int const& timestep
)
{
    ScaFES::Ntuple<int, DIM> idxNode;
    /*------------------------------------------------------------------------*/
    trace_on(tapeId, 1);
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        aDfIndep[ii] <<= dfIndep.at(ii);
    }
    // Remark:
    // Reset vector of dependent variables to zero.
    // Just is due to the asynchronous MPI communication and the
    // introduction of a temporary vector.
    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = aDfDep.at(ii).memAll().begin(),
             et = aDfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) = this->defaultValue().at(ii);
        }
    }

    // Set t = t_S.
    // The rhs need not to be an independent variable the time at the
    // initialization will always be t=t_S.
    // ==> Variable of type double for rhs is sufficient.
    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        aDfDep[ii].time() = this->params().timeIntervalStart();
    }

    (this->*funcInitInner)(aDfDep, aDfIndep, timestep);
    (this->*funcInitBorder)(aDfDep, aDfIndep, timestep);

    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfDep.at(ii).memAll().begin(),
            et = dfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) >>= dfDep.at(ii)(idxNode);
        }
        aDfDep[ii].time() >>= dfDep.at(ii).time();
    }
    trace_off();
}
#endif
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::init_zosForward_paramsBased(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    short int const& tapeId,
    std::vector<CT> const& dfIndep,
    const int& /*timestep*/
)
{
    CT tmp;
    static_assert(std::is_same<decltype(tmp), double>::value,
                  "Only for CT=double!");

    this->traceInitParamsBased(dfDep, dfIndep);

    boost::mpi::timer timerInit;
    timerInit.restart();

    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Initialise(pb) iterate"
                  << " using ADOL-C tape #"
                  << tapeId << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

#ifdef SCAFES_HAVE_ADOLC
#ifdef _OPENMP
    const int NTHREADS_ENV = omp_get_max_threads();
    omp_set_num_threads(1);
#endif
// // #ifdef _OPENMP
// // #pragma omp parallel // firstprivate(ADOLC_OpenMP_Handler)
// //     {
// // #endif
    const int nIndependents = dfIndep.size();
    int nDependents = 0;
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        nDependents += dfDep[ii].nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nIndependents = " << nIndependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nDependents   = " << nDependents << std::endl;
    }

    /*------------------------------------------------------------------------*/
    // Points to first element of first vector of old iterate.
    //    double* vectIndep = dfIndep[0].elemData();
    const double* vectIndep = static_cast<const double*>(dfIndep.data());
    // Points to first element of first vector of new iterate.
    double* vectDep = static_cast<double*>(dfDep[0].elemData());

    /*------------------------------------------------------------------------*/
    ::zos_forward(tapeId,
                  nDependents,
                  nIndependents,
                  0, // Do not keep.
                  vectIndep,
                  vectDep);
// // #ifdef _OPENMP
// //     }
// // #endif
#ifdef _OPENMP
    omp_set_num_threads(NTHREADS_ENV);
#endif

#endif // HAVE_ADOLC

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Initialised." << std::endl;
    }

    this->mWallClockTimes["init"] += timerInit.elapsed();

}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::init_fovForward_paramsBased(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    short int const& tapeId,
    std::vector<CT> const& dfIndep,
    std::vector<CT> const& dfGradIndep,
    const int& /*timestep*/
)
{
    CT tmp;
    static_assert(std::is_same<decltype(tmp), double>::value,
                  "Only for CT=double!");

    this->traceInitParamsBased(dfDep, dfIndep);

    boost::mpi::timer timerInit;
    timerInit.restart();

    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank())
                  << " * Initialise(pb) iterate and its gradients"
                  << " using ADOL-C tape #" << tapeId
                  << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
#ifdef SCAFES_HAVE_ADOLC
#ifdef _OPENMP
    const int NTHREADS_ENV = omp_get_max_threads();
    omp_set_num_threads(1);
#endif
    const int nDirections = this->nColumns();
    const int nIndependents = dfIndep.size();
    int nDependents = 0;
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        nDependents += dfDep[ii].nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    // Rearrange one-dimensional arrays to two-dimensional ones
    // such that these arrays can be passed to ADOL-C functions.
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank())
                  << " * nIndependents = " << nIndependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nDependents   = " << nDependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nDirections   = " << nDirections << std::endl;
    }

    /*------------------------------------------------------------------------*/
    // Points to first element of gradient of parameter vector.
    CT* elemCurrIndep = const_cast<CT*>(dfGradIndep.data());
    for (int ii = 0; ii < nIndependents; ++ii) {
        this->mMemoryVectGradParamsAsMatrix[ii] = elemCurrIndep;
        elemCurrIndep += nDirections;
    }
    CT* elemCurrDep = dfGradDep[0].elemData();
    // Just take memory lump #1 OR memory lump #2.
    for (int ii = 0; ii < nDependents; ++ii) {
        this->mMemoryVectGradUnknownDfsRowsTwo[ii] = elemCurrDep;
        elemCurrDep += nDirections;
    }

    /*------------------------------------------------------------------------*/
    // Set gradient of independent variables to I.
    for (int ii = 0; ii < nIndependents; ++ii) {
        for (int jj = 0; jj < nDirections; ++jj) {
            this->mMemoryVectGradParamsAsMatrix[ii][jj] = static_cast<CT>(0);
            if (jj == ii) {
                this->mMemoryVectGradParamsAsMatrix[ii][jj] = static_cast<CT>(1);
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Points to first element of first vector.
    const double* vectIndep = dfIndep.data();
    double* vectDep = dfDep[0].elemData();
    double** vectIndepGrad = this->mMemoryVectGradParamsAsMatrix.data();
    double** vectDepGrad = this->mMemoryVectGradUnknownDfsRowsTwo.data();

    /*------------------------------------------------------------------------*/
    ::fov_forward(tapeId,
                  nDependents,
                  nIndependents,
                  nDirections,
                  vectIndep,
                  vectIndepGrad,
                  vectDep,
                  vectDepGrad);

// // #ifdef _OPENMP
// //     }
// // #endif
#ifdef _OPENMP
    omp_set_num_threads(NTHREADS_ENV);
#endif

#endif // HAVE_ADOLC

    /*------------------------------------------------------------------------*/
    // Rearrange back two-dimensional arrays (gradients) 
    // to one-dimensional arrays.
    // NOT necessary: Values are already at correct position in memory.
    // Local variables vectDep, vectGradDep etc. can be thrown away.
    // These variables are just pointers to the memory of the data fields.

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank())
                  << "   Initialised." << std::endl;
    }

    this->mWallClockTimes["init"] += timerInit.elapsed();
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::traceUpdate(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    int const& timestep
)
{
    if (this->isTracedUpdate()) {
        return;
    }

    if (!(this->params().enabledAdolc())) {
        throw std::runtime_error("Tracing not possible. ADOL-C is not enabled.");
    }

    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Create all tapes"
                  << " for tracing update(" << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name()
                  << ",...)."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    boost::mpi::timer timerTrace;
    timerTrace.restart();

#ifdef SCAFES_HAVE_ADOLC

#ifdef _OPENMP
    const int NTHREADS_ENV = omp_get_max_threads();
    omp_set_num_threads(1);
#endif

    ScaFES::Ntuple<int, DIM> idxNode;
    // Copy constructor would not work because the types of the data fields
    // are different (adouble vs. double).
    unsigned long int nElemsDfIndepTotal = 0;
    unsigned long int nElemsDfDepTotal = 0;
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        nElemsDfIndepTotal += dfIndep.at(ii).nElemsAllocated();
    }
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        nElemsDfDepTotal += dfDep.at(ii).nElemsAllocated();
    }
    /*------------------------------------------------------------------------*/
    std::vector<adouble> memoryAdfIndep;
    if (0 < nElemsDfIndepTotal) {
        memoryAdfIndep.resize(nElemsDfIndepTotal);
    }
    /*------------------------------------------------------------------------*/
    std::vector<adouble> memoryAdfDep;
    if (0 < nElemsDfDepTotal) {
        memoryAdfDep.resize(nElemsDfDepTotal);
    }
    /*------------------------------------------------------------------------*/
    std::vector< ScaFES::DataField<adouble,DIM> > aDfIndep;
    aDfIndep.reserve(dfIndep.size());
    adouble* elemDataIndepCurr = memoryAdfIndep.data();
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        aDfIndep.push_back(
            ScaFES::DataField<adouble, DIM>(
                dfIndep.at(ii).name(),
                dfIndep.at(ii).params(),
                dfIndep.at(ii).gridGlobal(),
                dfIndep.at(ii).memAll(),
                dfIndep.at(ii).stencilWidth(),
                elemDataIndepCurr,
                1,
                dfIndep.at(ii).memHole(),
                dfIndep.at(ii).nLayers()
            )
        );
        elemDataIndepCurr += dfIndep.at(ii).nElemsAllocated();
    }
    std::vector< ScaFES::DataField<adouble,DIM> > aDfDep;
    aDfDep.reserve(dfDep.size());
    adouble* elemDataDepCurr = memoryAdfDep.data();
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        aDfDep.push_back(
            ScaFES::DataField<adouble, DIM>(
                dfDep.at(ii).name(),
                dfDep.at(ii).params(),
                dfDep.at(ii).gridGlobal(),
                dfDep.at(ii).memAll(),
                dfDep.at(ii).stencilWidth(),
                elemDataDepCurr,
                1,
                dfDep.at(ii).memHole(),
                dfDep.at(ii).nLayers()
            )
        );
        elemDataDepCurr += dfDep.at(ii).nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    // * asynchron: 2 tapes: 1 tape (localBorder) and 1 tape (localInner)
    // * synchron:  1 tape:  1 tape (globalInner + globalBorder)
// #ifdef _OPENMP
// //#pragma omp parallel 
// //    default(none)
// //    shared(dfIndep, dfDep, timestep, aDfDep, aDfIndep)
// //    firstprivate(ADOLC_OpenMP_Handler)
// //    {
// #endif
    if (this->useAsynchronMode()) {
        this->createTapesForUpdate(
            this->tapeIdUpdatePartitionBorder(),
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionBorder<adouble>,
            dfDep,
            dfIndep,
            aDfDep,
            aDfIndep,
            timestep
         );
// //#ifdef _OPENMP
// //#pragma omp barrier
// //#endif
        this->createTapesForUpdate(
            this->tapeIdUpdatePartitionInner(),
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionInner<adouble>,
            dfDep,
            dfIndep,
            aDfDep,
            aDfIndep,
            timestep
         );
    }
    else {
        this->createTapeForUpdatePartitionAll(
            this->tapeIdUpdatePartitionAll(),
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionAllInner<adouble>,
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionAllBorder<adouble>,
            dfDep,
            dfIndep,
            aDfDep,
            aDfIndep,
            timestep
         );
// //#ifdef _OPENMP
// //      }
// //#endif
    }

    /*------------------------------------------------------------------------*/
    if (this->useLeapfrogIntegration()) {
        if (this->useAsynchronMode()) {
            this->createTapesForUpdate(
                this->tapeIdUpdatePartitionBorder2(),
                &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionBorder2<adouble>,
                dfDep,
                dfIndep,
                aDfDep,
                aDfIndep,
                timestep
            );
            this->createTapesForUpdate(
                this->tapeIdUpdatePartitionInner2(),
                &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionInner2<adouble>,
                dfDep,
                dfIndep,
                aDfDep,
                aDfIndep,
                timestep
            );
        }
        else {
            this->createTapeForUpdatePartitionAll(
                this->tapeIdUpdatePartitionAll2(),
                &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionAllInner2<adouble>,
                &ScaFES::Problem<OWNPRBLM, CT, DIM>::template updatePartitionAllBorder2<adouble>,
                dfDep,
                dfIndep,
                aDfDep,
                aDfIndep,
                timestep
            );
        }
    }

#ifdef _OPENMP
    omp_set_num_threads(NTHREADS_ENV);
#endif

#endif // HAVE_ADOLC

    this->mWallClockTimes["trace"] += timerTrace.elapsed();
    this->mIsTracedUpdate = true;

    // REMARK: No synchronization is necessary as one just wants to trace
    // the mathematical operations.
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Created."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
#ifdef SCAFES_HAVE_ADOLC
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::createTapeForUpdatePartitionAll(
    short int const& tapeId,
    TypeFuncUpdateToTrace funcUpdateInner,
    TypeFuncUpdateToTrace funcUpdateBorder,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
    std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    int idxThread = omp_get_thread_num();
#else
    int idxThread = 0;    
#endif
    ScaFES::Ntuple<int, DIM> idxNode;
    /*------------------------------------------------------------------------*/
    short int tmpTapeId = tapeId + idxThread;
    trace_on(tmpTapeId, 1);
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfIndep.at(ii).memAll().begin(),
             et = dfIndep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfIndep[ii](idxNode) <<= dfIndep.at(ii)(idxNode);
        }
        aDfIndep[ii].time() <<= dfIndep[ii].time();
    }
    // Remark:
    // Reset vector of dependent variables to zero.
    // Especially, values at all halos will be set to zero.
    // Values at all memNormal nodes will be overwritten via calling
    // funcUpdateInner, funcUpdateBorder.
    // This is due to the asynchronous MPI communication and the
    // introduction of a temporary vector.
    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = aDfDep.at(ii).memAll().begin(),
            et = aDfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) = aDfIndep.at(ii)(idxNode);
        }
    }

    // Set t = t_S + timestep * \tau = t_l + \tau.
    // Attention: Update time of dep. var. BEFORE calls of update methods.
    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        // \tau is a constant. ==> \tau need not to be of type adouble. ==> OK.
        aDfDep[ii].time() = aDfIndep[ii].time() + this->params().tau();
    }

    (this->*funcUpdateInner)(aDfDep, aDfIndep, timestep);
    (this->*funcUpdateBorder)(aDfDep, aDfIndep, timestep);

    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfDep.at(ii).memAll().begin(),
            et = dfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) >>= dfDep.at(ii)(idxNode);
        }
        aDfDep[ii].time() >>= dfDep.at(ii).time();
    }
    trace_off();
}
#endif
/*----------------------------------------------------------------------------*/
#ifdef SCAFES_HAVE_ADOLC
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::createTapesForUpdate(
    short int const& tapeId,
    TypeFuncUpdateToTrace funcUpdate,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<adouble,DIM> >& aDfDep,
    std::vector< ScaFES::DataField<adouble,DIM> > const& aDfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    int idxThread = omp_get_thread_num();
#else
    int idxThread = 0;
#endif
    ScaFES::Ntuple<int, DIM> idxNode;
    /*------------------------------------------------------------------------*/
    short int tmpTapeId = tapeId + idxThread;
    trace_on(tmpTapeId, 1);
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfIndep.at(ii).memAll().begin(),
             et = dfIndep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfIndep[ii](idxNode) <<= dfIndep.at(ii)(idxNode);
        }
        aDfIndep[ii].time() <<= dfIndep.at(ii).time();
    }

    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = aDfDep.at(ii).memAll().begin(),
            et = aDfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) = aDfIndep.at(ii)(idxNode);
        }
    }
    
    // Set t = t_S + timestep * \tau = t_l + \tau.
    // Attention: Update time of dep. var. BEFORE calls of update methods.
    for (std::size_t ii = 0; ii < aDfDep.size(); ++ii) {
        // \tau is a constant. ==> \tau need not to be of type adouble. ==> OK.
        aDfDep[ii].time() = aDfIndep[ii].time() + this->params().tau();
    }

    (this->*funcUpdate)(aDfDep, aDfIndep, timestep);

    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        for (typename ScaFES::Grid<DIM>::iterator it = dfDep.at(ii).memAll().begin(),
            et = dfDep.at(ii).memAll().end(); it < et; ++it) {
            idxNode = it.idxNode();
            aDfDep[ii](idxNode) >>= dfDep.at(ii)(idxNode);
        }
        aDfDep[ii].time() >>= dfDep[ii].time();
    }
    trace_off();
}
#endif
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::update_zosForward(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    short int const& tapeId,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    int const& /*timestep*/
)
{
    CT tmp;
    static_assert(std::is_same<decltype(tmp), double>::value,
                  "Only for CT=double!");

    boost::mpi::timer timerUpdate;
    timerUpdate.restart();

    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Compute new iterates ("
                  << dfDep.at(0).name() << ",..."
                  << ") using ADOL-C zos_forward() and tape #"
                  << tapeId << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

#ifdef SCAFES_HAVE_ADOLC
#ifdef _OPENMP
    const int NTHREADS_ENV = omp_get_max_threads();
    omp_set_num_threads(1);
#endif
// // #ifdef _OPENMP
// // #pragma omp parallel // firstprivate(ADOLC_OpenMP_Handler)
// //     {
// // #endif
    int nDependents = 0;
    int nIndependents = 0;
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        nIndependents += dfIndep.at(ii).nElemsAllocated();
    }
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        nDependents += dfDep.at(ii).nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    // Rearrange one-dimensional arrays to two-dimensional ones
    // such that these arrays can be passed to ADOL-C functions.
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank())
                  << " * nIndependents = " << nIndependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nDependents   = " << nDependents << std::endl;
    }

    /*------------------------------------------------------------------------*/
    // Points to first element of first vector of old iterate.
    const double* vectIndep = static_cast<const double*>(dfIndep[0].elemData());
    // Points to first element of first vector of new iterate.
    double* vectDep = static_cast<double*>(dfDep[0].elemData());

    /*------------------------------------------------------------------------*/
    ::zos_forward(tapeId,
                  nDependents,
                  nIndependents,
                  0, // Do not keep.
                  vectIndep,
                  vectDep);

// // #ifdef _OPENMP
// //     }
// // #endif
#ifdef _OPENMP
    omp_set_num_threads(NTHREADS_ENV);
#endif

#endif // HAVE_ADOLC

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }

    this->mWallClockTimes["update"] += timerUpdate.elapsed();

}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::update_fovForward(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    short int const& tapeId,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    int const& /*timestep*/
)
{
    CT tmp;
    static_assert(std::is_same<decltype(tmp), double>::value,
                  "Only for CT=double!");

    boost::mpi::timer timerUpdate;
    timerUpdate.restart();

    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Compute new iterates ("
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << ") and its gradients"
                  << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   using ADOL-C fov_forward() and tape #"
                  << tapeId << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

#ifdef SCAFES_HAVE_ADOLC
#ifdef _OPENMP
    const int NTHREADS_ENV = omp_get_max_threads();
    omp_set_num_threads(1);
#endif
// // #ifdef _OPENMP
// // #pragma omp parallel // firstprivate(ADOLC_OpenMP_Handler)
// //     {
// // #endif
//    int kk = 0;
    const int nDirections = this->nColumns();
    int nDependents = 0;
    int nIndependents = 0;
    for (std::size_t ii = 0; ii < dfIndep.size(); ++ii) {
        nIndependents += dfIndep[ii].nElemsAllocated();
    }
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        nDependents += dfDep[ii].nElemsAllocated();
    }

    // Rearrange one-dimensional arrays to two-dimensional ones
    // such that these arrays can be passed to ADOL-C functions.
    if ((this->params().rankOutput() == this->myRank())
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                               this->myRank())
                  << " * nIndependents = " << nIndependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nDependents   = " << nDependents << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nDirections   = " << nDirections << std::endl;
    }

    CT* elemCurr = const_cast<CT*>(dfGradIndep[0].elemData());
    for (int ii = 0; ii < nIndependents; ++ii) {
        this->mMemoryVectGradUnknownDfsRowsOne[ii] = elemCurr;
        elemCurr += nDirections;
    }
    elemCurr = dfGradDep[0].elemData();
    for (int ii = 0; ii < nDependents; ++ii) {
        this->mMemoryVectGradUnknownDfsRowsTwo[ii] = elemCurr;
        elemCurr += nDirections;
    }

    // Points to first element of first vector of old iterate.
    double* vectIndep = dfIndep[0].elemData();
    // Points to first element of first vector of new iterate.
    double* vectDep = dfDep[0].elemData();
    double** vectIndepGrad = this->mMemoryVectGradUnknownDfsRowsOne.data();
    double** vectDepGrad = this->mMemoryVectGradUnknownDfsRowsTwo.data();

    /*------------------------------------------------------------------------*/
    ::fov_forward(tapeId,
                  nDependents,
                  nIndependents,
                  nDirections,
                  vectIndep,
                  vectIndepGrad,
                  vectDep,
                  vectDepGrad);

    /*------------------------------------------------------------------------*/
    // Rearrange back two-dimensional arrays (gradients) 
    // to one-dimensional arrays.
    // NOT necessary: Values are already at correct position in memory.
    // Local variables vectDep, vectGradDep etc. can be thrown away.
    // These variables are just pointers to the memory of the data fields.
// // #ifdef _OPENMP
// //     }
// // #endif
#ifdef _OPENMP
    omp_set_num_threads(NTHREADS_ENV);
#endif

#endif // HAVE_ADOLC

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }

    this->mWallClockTimes["update"] += timerUpdate.elapsed();

}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateValsAtNodesOfUnknownDfs(
    int const& timestep
)
{
    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update all unknown data fields..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    if (this->useAsynchronMode()) {
        this->updateAsynchronously(
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionInner,
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionBorder,
            this->mVectUnknownDfsDomNew,
            this->mVectGradUnknownDfsDomNew,
            this->vectUnknownDfsDomOld(),
            this->mVectGradUnknownDfsDomOld,
            timestep
        );
    }
    else {
        this->updateSynchronously(
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionAll,
            this->mVectUnknownDfsDomNew,
            this->mVectGradUnknownDfsDomNew,
            this->mVectUnknownDfsBdryNew,
            this->mVectGradUnknownDfsBdryNew,
            this->vectUnknownDfsDomOld(),
            this->mVectGradUnknownDfsDomOld,
            this->vectUnknownDfsBdryOld(),
            this->mVectGradUnknownDfsBdryOld,
            timestep
        );
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::update2ValsAtNodesOfUnknownDfs(
    int const& timestep
)
{
    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update (2nd cycle) all unknown data fields"
                  << " using the leapfrog scheme..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    if (this->useAsynchronMode()) {
        this->updateAsynchronously(
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionInner2,
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionBorder2,
            this->mVectUnknownDfsDomNew,
            this->mVectGradUnknownDfsDomNew,
            this->vectUnknownDfsDomOld(),
            this->mVectGradUnknownDfsDomOld,
            timestep
        );
    }
    else {
        this->updateSynchronously(
            &ScaFES::Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionAll2,
            this->mVectUnknownDfsDomNew,
            this->mVectGradUnknownDfsDomNew,
            this->mVectUnknownDfsBdryNew,
            this->mVectGradUnknownDfsBdryNew,
            this->vectUnknownDfsDomOld(),
            this->mVectGradUnknownDfsDomOld,
            this->vectUnknownDfsBdryOld(),
            this->mVectGradUnknownDfsBdryOld,
            timestep
        );
    }

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateAsynchronously(
    TypeFuncUpdateToEval funcLocalInner,
    TypeFuncUpdateToEval funcLocalBorder,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    int const& timestep
)
{
    boost::mpi::timer timerSync;

    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update asynchronously data fields "
                  << dfDep.at(0).name() << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
    /*------------------------------------------------------------------------*/
    (this->*funcLocalBorder)(
        dfDep,
        dfGradDep,
        dfIndep,
        dfGradIndep,
        timestep
    );
    /*------------------------------------------------------------------------*/
    timerSync.restart();
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].copyValuesFromMemCommToSendBuffer(timestep);
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].copyValuesFromMemCommToSendBuffer(timestep);
    }
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].exchangeValuesInBuffers(timestep);
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].exchangeValuesInBuffers(timestep);
    }
    this->mWallClockTimes["sync"] += timerSync.elapsed();
    /*------------------------------------------------------------------------*/
    if (this->params().enabledAdolc()) {
        (this->*funcLocalInner)(
            this->mVectUnknownDfsDomTmp,
            this->mVectGradUnknownDfsDomTmp,
            dfIndep,
            dfGradIndep,
            timestep
        );
        for (std::size_t ii = 0; ii < this->mVectUnknownDfsDomTmp.size(); ++ii) {
            dfDep[ii] += (this->mVectUnknownDfsDomTmp[ii] - dfIndep[ii]);
        }
        for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsDomTmp.size(); ++ii) {
            dfGradDep[ii] += this->mVectGradUnknownDfsDomTmp[ii];
            // Subtract identity matrix from dfGradDep [KF, 15.05.2014].
            // dfGradDep[ii] -= IdMat;
        }
    }
    else {
        (this->*funcLocalInner)(
            dfDep,
            dfGradDep,
            dfIndep,
            dfGradIndep,
            timestep
        );
    }
    timerSync.restart();
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].waitAll();
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].waitAll();
    }
    for (std::size_t ii = 0; ii < dfDep.size(); ++ii) {
        dfDep[ii].copyValuesFromReceiveBufferToMemGhost(timestep);
    }
    for (std::size_t ii = 0; ii < dfGradDep.size(); ++ii) {
        dfGradDep[ii].copyValuesFromReceiveBufferToMemGhost(timestep);
    }
    this->mWallClockTimes["sync"] += timerSync.elapsed();
    /*------------------------------------------------------------------------*/
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateSynchronously(
    TypeFuncUpdateToEval funcGlobal,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepDomain,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradDomain,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepBdry,
    std::vector< ScaFES::DataField<CT,DIM> >& dfDepGradBdry,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepDomain,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepGradDomain,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepBdry,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndepGradBdry,
    int const& timestep
)
{
    boost::mpi::timer timerSync;

    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update synchronously data fields "
                  << dfDepDomain.at(0).name() << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    // 1) Update DOMAIN data fields at all GLOBAL INNER nodes.
    // Remark: Only those data fields are involved which are defined at
    // the global inner nodes.
    // In detail: the dependent data fields are only dependent on data fields
    // defined on the global inner nodes, i.e. "domain data fields".
    // 2) Update ALL (=DOMAIN+BOUNDARY) data fields at all GLOBAL BORDER nodes.
    std::vector< ScaFES::DataField<CT,DIM> > vectIndepAllTmp(dfIndepDomain);
    for (std::size_t ii = 0; ii < dfIndepBdry.size(); ++ii) {
        vectIndepAllTmp.push_back(dfIndepBdry[ii]);
    }
    std::vector< ScaFES::DataField<CT,DIM> > vectIndepGradAllTmp(dfIndepGradDomain);
    for (std::size_t ii = 0; ii < dfIndepGradBdry.size(); ++ii) {
        vectIndepGradAllTmp.push_back(dfIndepGradBdry[ii]);
    }
    std::vector< ScaFES::DataField<CT,DIM> > vectDepAllTmp(dfDepDomain);
    for (std::size_t ii = 0; ii < dfDepBdry.size(); ++ii) {
        vectDepAllTmp.push_back(dfDepBdry[ii]);
    }
    std::vector< ScaFES::DataField<CT,DIM> > vectDepGradAllTmp(dfDepGradDomain);
    for (std::size_t ii = 0; ii < dfDepGradBdry.size(); ++ii) {
        vectDepGradAllTmp.push_back(dfDepGradBdry[ii]);
    }
    (this->*funcGlobal)(
        vectDepAllTmp, 
        vectDepGradAllTmp,
        vectIndepAllTmp, 
        vectIndepGradAllTmp,
        timestep
    );
    /*------------------------------------------------------------------------*/
    timerSync.restart();
    for (std::size_t ii = 0; ii < dfDepDomain.size(); ++ii) {
        dfDepDomain[ii].sync(timestep);
    }
    for (std::size_t ii = 0; ii < dfDepGradDomain.size(); ++ii) {
        dfDepGradDomain[ii].sync(timestep);
    }
    for (std::size_t ii = 0; ii < dfDepBdry.size(); ++ii) {
        dfDepBdry[ii].sync(timestep);
    }
    for (std::size_t ii = 0; ii < dfDepGradBdry.size(); ++ii) {
        dfDepGradBdry[ii].sync(timestep);
    }
    this->mWallClockTimes["sync"] += timerSync.elapsed();
    /*------------------------------------------------------------------------*/    
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionInner(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        this->traceUpdate(
            dfDep,
            dfIndep,
            timestep
        );
        if (this->computeGradients()) {
            this->update_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdUpdatePartitionInner(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->update_zosForward(
                dfDep,
                this->tapeIdUpdatePartitionInner(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerUpdate;
        timerUpdate.restart();
#ifdef _OPENMP
#pragma omp parallel \
    default(none) \
    shared(dfDep, dfIndep, timestep)
      {
#endif
        this->template updatePartitionInner<CT>(dfDep, dfIndep, timestep);
#ifdef _OPENMP
      }
#endif
        this->mWallClockTimes["update"] += timerUpdate.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionBorder(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        this->traceUpdate(
            dfDep,
            dfIndep,
            timestep
        );
        if (this->computeGradients()) {
            this->update_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdUpdatePartitionBorder(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->update_zosForward(
                dfDep,
                this->tapeIdUpdatePartitionBorder(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerUpdate;
        timerUpdate.restart();
#ifdef _OPENMP
#pragma omp parallel \
    default(none) \
    shared(dfDep, dfIndep, timestep)
      {
#endif
        this->template updatePartitionBorder<CT>(dfDep, dfIndep, timestep);
#ifdef _OPENMP
      }
#endif
        this->mWallClockTimes["update"] += timerUpdate.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionAll(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        this->traceUpdate(
            dfDep,
            dfIndep,
            timestep
        );
        if (this->computeGradients()) {
            this->update_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdUpdatePartitionAll(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->update_zosForward(
                dfDep,
                this->tapeIdUpdatePartitionAll(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerUpdate;
        timerUpdate.restart();
#ifdef _OPENMP
#pragma omp parallel \
    default(none) \
    shared(dfDep, dfIndep, timestep)
      {
#endif
        this->template updatePartitionAllBorder<CT>(dfDep, dfIndep, timestep);
        this->template updatePartitionAllInner<CT>(dfDep, dfIndep, timestep);
#ifdef _OPENMP
      }
#endif
        this->mWallClockTimes["update"] += timerUpdate.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionInner2(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        if (this->computeGradients()) {
            this->update_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdUpdatePartitionInner2(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->update_zosForward(
                dfDep,
                this->tapeIdUpdatePartitionInner2(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerUpdate;
        timerUpdate.restart();
#ifdef _OPENMP
#pragma omp parallel \
    default(none) \
    shared(dfDep, dfIndep, timestep)
      {
#endif
        this->template updatePartitionInner2<CT>(dfDep, dfIndep, timestep);
#ifdef _OPENMP
      }
#endif
        this->mWallClockTimes["update"] += timerUpdate.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionBorder2(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        if (this->computeGradients()) {
            this->update_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdUpdatePartitionBorder2(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->update_zosForward(
                dfDep,
                this->tapeIdUpdatePartitionBorder2(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerUpdate;
        timerUpdate.restart();
#ifdef _OPENMP
#pragma omp parallel \
    default(none) \
    shared(dfDep, dfIndep, timestep)
      {
#endif
        this->template updatePartitionBorder2<CT>(dfDep, dfIndep, timestep);
#ifdef _OPENMP
      }
#endif
        this->mWallClockTimes["update"] += timerUpdate.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::wrapperUpdatePartitionAll2(
    std::vector< ScaFES::DataField<CT,DIM> >& dfDep,
    std::vector< ScaFES::DataField<CT,DIM> >& dfGradDep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfIndep,
    std::vector< ScaFES::DataField<CT,DIM> > const& dfGradIndep,
    const int& timestep
)
{
    if (this->params().enabledAdolc()) {
        if (this->computeGradients()) {
            this->update_fovForward(
                dfDep,
                dfGradDep,
                this->tapeIdUpdatePartitionAll2(),
                dfIndep,
                dfGradIndep,
                timestep
            );
        }
        else {
            this->update_zosForward(
                dfDep,
                this->tapeIdUpdatePartitionAll2(),
                dfIndep,
                timestep
            );
        }
    }
    else {
        boost::mpi::timer timerUpdate;
        timerUpdate.restart();

#ifdef _OPENMP
#pragma omp parallel \
    default(none) \
    shared(dfDep, dfIndep, timestep)
      {
#endif
        this->template updatePartitionAllBorder2<CT>(dfDep, dfIndep, timestep);
        this->template updatePartitionAllInner2<CT>(dfDep, dfIndep, timestep);
#ifdef _OPENMP
      }
#endif
        this->mWallClockTimes["update"] += timerUpdate.elapsed();
   }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::swapPointersOfUnknownDfs()
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Swap pointers of all unknown data fields..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    // Swap old and new iterate.
    for (std::size_t ii = 0; ii < this->vectUnknownDfsDomNew().size(); ++ii) {
        this->mVectUnknownDfsDomNew[ii].swapPointerToMemoryWith(
            this->mVectUnknownDfsDomOld[ii]
        );
    }
    for (std::size_t ii = 0; ii < this->vectUnknownDfsBdryNew().size(); ++ii) {
        this->mVectUnknownDfsBdryNew[ii].swapPointerToMemoryWith(
            this->mVectUnknownDfsBdryOld[ii]
        );
    }
    /*------------------------------------------------------------------------*/
    // Swap gradients of old and new iterate.
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsDomNew.size(); ++ii) {
        this->mVectGradUnknownDfsDomNew[ii].swapPointerToMemoryWith(
            this->mVectGradUnknownDfsDomOld[ii]
        );
    }
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsBdryNew.size(); ++ii) {
        this->mVectGradUnknownDfsBdryNew[ii].swapPointerToMemoryWith(
            this->mVectGradUnknownDfsBdryOld[ii]
        );
    }

    /*------------------------------------------------------------------------*/
    // Update current time of "new" data fields AFTER swapping.
    for (std::size_t ii = 0; ii < this->mVectUnknownDfsDomNew.size(); ++ii) {
        this->mVectUnknownDfsDomNew[ii].time()
            = this->mVectUnknownDfsDomOld[ii].time() + this->params().tau();
    }
    for (std::size_t ii = 0; ii < this->mVectUnknownDfsBdryNew.size(); ++ii) {
        this->mVectUnknownDfsBdryNew[ii].time()
            = this->mVectUnknownDfsBdryOld[ii].time() + this->params().tau();
    }
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsDomNew.size(); ++ii) {
        this->mVectGradUnknownDfsDomNew[ii]
            = this->mVectGradUnknownDfsDomOld[ii].time() + this->params().tau();
    }
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsBdryNew.size(); ++ii) {
        this->mVectGradUnknownDfsBdryNew[ii]
            = this->mVectGradUnknownDfsBdryOld[ii].time() + this->params().tau();
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                   << "   Swapped."
                   << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::updateTimesOfUnknownDfs()
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update times of all unknown data fields..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    // Update current time of "new" data fields AFTER swapping.
    for (std::size_t ii = 0; ii < this->mVectUnknownDfsDomNew.size(); ++ii) {
        this->mVectUnknownDfsDomNew[ii].time()
            = this->mVectUnknownDfsDomOld[ii].time() + this->params().tau();
    }
    for (std::size_t ii = 0; ii < this->mVectUnknownDfsBdryNew.size(); ++ii) {
        this->mVectUnknownDfsBdryNew[ii].time()
            = this->mVectUnknownDfsBdryOld[ii].time() + this->params().tau();
    }
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsDomNew.size(); ++ii) {
        this->mVectGradUnknownDfsDomNew[ii]
            = this->mVectGradUnknownDfsDomOld[ii].time() + this->params().tau();
    }
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsBdryNew.size(); ++ii) {
        this->mVectGradUnknownDfsBdryNew[ii]
            = this->mVectGradUnknownDfsBdryOld[ii].time() + this->params().tau();
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                   << "   Swapped."
                   << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::swapTimesOfUnknownDfs()
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Swap times of all unknown data fields..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    double tmpTime = 0.0;
    for (std::size_t ii = 0; ii < this->mVectUnknownDfsDomNew.size(); ++ii) {
        tmpTime = this->mVectUnknownDfsDomOld[ii].time();
        this->mVectUnknownDfsDomOld[ii].time() 
            = this->mVectUnknownDfsDomNew[ii].time();
        this->mVectUnknownDfsDomNew[ii].time() = tmpTime;
    }
    for (std::size_t ii = 0; ii < this->mVectUnknownDfsBdryNew.size(); ++ii) {
        tmpTime = this->mVectUnknownDfsBdryOld[ii].time();
        this->mVectUnknownDfsBdryOld[ii].time() 
            = this->mVectUnknownDfsBdryNew[ii].time();
        this->mVectUnknownDfsBdryNew[ii].time() = tmpTime;
    }
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsDomNew.size(); ++ii) {
        tmpTime = this->mVectGradUnknownDfsDomOld[ii].time();
        this->mVectGradUnknownDfsDomOld[ii].time() 
            = this->mVectGradUnknownDfsDomNew[ii].time();
        this->mVectGradUnknownDfsDomNew[ii].time() = tmpTime;
    }
    for (std::size_t ii = 0; ii < this->mVectGradUnknownDfsBdryNew.size(); ++ii) {
        tmpTime = this->mVectGradUnknownDfsBdryOld[ii].time();
        this->mVectGradUnknownDfsBdryOld[ii].time() 
            = this->mVectGradUnknownDfsBdryNew[ii].time();
        this->mVectGradUnknownDfsBdryNew[ii].time() = tmpTime;
    }

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Swapped." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::writeAllDfs(int const& timestep)
{
    boost::mpi::timer timerWrite;
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Write all data fields to file..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    timerWrite.restart();
    this->writeDfsToFile(timestep);
    this->mKind.write(timestep);
    this->mWallClockTimes["write"] += timerWrite.elapsed();
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Written." << std::endl;
    }
}

/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::evalKnownDfs(int const& timestep)
{
    bool evalDf = false;
    for (std::size_t ii = 0; ii < this->mComputeError.size(); ++ii) {
       if (this->mComputeError.at(ii) || this->isKnownDf().at(ii)) {
           evalDf = true;
           break;
       }
    }
    if (!evalDf) {
        return;
    }

    boost::mpi::timer timerEval;
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Evaluate all known data fields..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    timerEval.restart();

    this->evalGlobInner(this->mVectKnownDfDom, timestep);
    this->evalGlobBorder(this->mVectKnownDfDom, timestep);
    for (std::size_t ii = 0; ii < this->mVectKnownDfDom.size(); ++ii) {
        this->mVectKnownDfDom[ii].sync(timestep);
    }
    if (0 < this->nUnknownBdryDfs()) {
        this->evalGlobBorder(this->mVectKnownDfBdry, timestep);
        for (std::size_t ii = 0; ii < this->mVectKnownDfBdry.size(); ++ii) {
            this->mVectKnownDfBdry[ii].sync(timestep);
        }
    }
    this->mWallClockTimes["evalKnownDf"] += timerEval.elapsed();

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Evaluated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::compErrOfUnknownDfs()
{
    boost::mpi::timer timerCompErr;
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Compute error for all unknown data fields..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    timerCompErr.restart();
        // Compute nodal error.
        int idDf = 0;
        for (std::size_t ii = 0; ii < this->mComputeError.size(); ++ii) {
            if (0 == this->nLayers().at(ii)) {
                if (!(this->isKnownDf().at(ii))) {
                    if (this->mComputeError.at(ii)) {
                        this->mVectUnknownDfsDomNew[idDf].compErrLinf(
                            this->mVectKnownDfDom[idDf]);
                    }
                    ++idDf;
                }
            }
        }
        if (0 < this->nUnknownBdryDfs()) {
            idDf = 0;
            for (std::size_t ii = 0; ii < this->mComputeError.size(); ++ii) {
                if (0 < this->nLayers().at(ii)) {
                    if (!(this->isKnownDf().at(ii))) {
                        if (this->mComputeError.at(ii)) {
                            this->mVectUnknownDfsBdryNew[idDf].compErrLinf(
                                this->mVectKnownDfBdry[idDf]);
                        }
                        ++idDf;
                    }
                }
            }
        }
    this->mWallClockTimes["compErr"] += timerCompErr.elapsed();

    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Computed." << std::endl;
    }
}


/*******************************************************************************
 * WOKR METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::iterateOverTime()
{
    int currTimestep = 0;

    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Iterate over time..." << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
#ifdef VTRACE
    unsigned int mid;
    mid = VT_MARKER_DEF("Prblm_iterate", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Prblm_iterate_start");
#endif

    for (currTimestep = 0; currTimestep <= this->nTimesteps(); ++currTimestep) {

        /*--------------------------------------------------------------------*/
        if ((this->params().rankOutput() == this->myRank()) 
            && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank())
                      << " * Process TIME STEP=" << currTimestep
                      << "..." << std::endl;
        }
        this->mParams.decreaseLevel();

        /*--------------------------------------------------------------------*/
        this->evalKnownDfs(currTimestep);
        if (0 < currTimestep) {
            this->updateValsAtNodesOfUnknownDfs(currTimestep);
            if (this->useLeapfrogIntegration()) {
                // Swap old / new data fields, BUT DO NOT SWAP times.
                // ==> Swap times 2x.
                this->swapPointersOfUnknownDfs(); // 1st swap of times.
                this->swapTimesOfUnknownDfs();    // 2nd swap of times.
                this->update2ValsAtNodesOfUnknownDfs(currTimestep);
            }
        }
        else {
            this->initUnknownDfs();
        }
        this->writeAllDfs(currTimestep);
        this->compErrOfUnknownDfs();
        this->swapPointersOfUnknownDfs();
        // Update times of old / new data fields AFTER swapping.
        this->updateTimesOfUnknownDfs();

        /*--------------------------------------------------------------------*/
        this->mParams.increaseLevel();
        if ((this->params().rankOutput() == this->myRank()) 
            && (0 < this->params().indentDepth())) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank())
                      << "   Processed." << std::endl;
        }
    }

    /*------------------------------------------------------------------------*/
    // Print some statistics.
    double tSum = 0.0;
    for (auto it = this->mWallClockTimes.begin(); 
         it != this->mWallClockTimes.end(); ++it) {
        tSum += it->second;
    }
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << std::scientific << std::fixed
                  << " * Summary of Wall clock times(iterateOverTime):"
                  << std::endl;
        for (auto it = this->mWallClockTimes.begin(); it != this->mWallClockTimes.end(); ++it) {
            std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                           this->myRank())
                      <<  "       t(" << std::setw(11) << it->first << ") = "
                      << std::setw(8) << std::setprecision(5)
                  << it->second
                  << " s ("
                  << std::setw(5) << std::setprecision(2)
                  << it->second / tSum * 100.0 << " %%)"
                  << std::endl;
        }
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "       --------------------------------------"
                  << std::endl
                  << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "       t(    iterate) = "
                  << std::setw(8) << std::setprecision(5)
                  << tSum
                  << " s."
                  << resetiosflags( std::ios::fixed | std::ios::showpoint )
                  << std::endl;
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Iterated."
                  << std::endl;
    }
#ifdef VTRACE
    mid = VT_MARKER_DEF("EndLoop", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Prblm_iterate_end");
#endif
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::increaseLevel()
{
    this->mParams.increaseLevel();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::decreaseLevel()
{
    this->mParams.decreaseLevel();
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::isTracedInit() const
{
    return this->mIsTracedInit;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int Problem<OWNPRBLM, CT, DIM>::tapeIdInitPartitionInner()
{
    return this->mTapeIdInitPartitionInner;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int Problem<OWNPRBLM, CT, DIM>::tapeIdInitPartitionBorder()
{
    return this->mTapeIdInitPartitionBorder;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int Problem<OWNPRBLM, CT, DIM>::tapeIdInitPartitionAll()
{
    return this->mTapeIdInitPartitionAll;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::isTracedUpdate() const
{
    return this->mIsTracedUpdate;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& Problem<OWNPRBLM, CT, DIM>::tapeIdUpdatePartitionInner() const
{
    return this->mTapeIdUpdatePartitionInner;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& Problem<OWNPRBLM, CT, DIM>::tapeIdUpdatePartitionBorder() const
{
    return this->mTapeIdUpdatePartitionBorder;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& Problem<OWNPRBLM, CT, DIM>::tapeIdUpdatePartitionAll() const
{
    return this->mTapeIdUpdatePartitionAll;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& Problem<OWNPRBLM, CT, DIM>::tapeIdUpdatePartitionInner2() const
{
    return this->mTapeIdUpdatePartitionInner2;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& Problem<OWNPRBLM, CT, DIM>::tapeIdUpdatePartitionBorder2() const
{
    return this->mTapeIdUpdatePartitionBorder2;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const& Problem<OWNPRBLM, CT, DIM>::tapeIdUpdatePartitionAll2() const
{
    return this->mTapeIdUpdatePartitionAll2;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::myRank() const
{
    return this->params().myWorld().rank();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<bool> const& Problem<OWNPRBLM, CT, DIM>::isKnownDf() const
{
    return this->mIsKnownDf;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<int> const& Problem<OWNPRBLM, CT, DIM>::nLayers() const
{
    return this->mNlayers;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<CT> const& Problem<OWNPRBLM, CT, DIM>::defaultValue() const
{
    return this->mDefaultValue;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nElemsUnknownDfs() const
{
    return this->mNelemsUnknownDfs;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nElemsKnownDfs() const
{
    return this->mNelemsKnownDfs;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nUnknownBdryDfs() const
{
    return this->mNunknownBdryDfs;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nNodesTotalPartition() const
{
    return this->mNnodesTotalPartition;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector< DataField<CT,DIM> > const& 
Problem<OWNPRBLM, CT, DIM>::vectUnknownDfsDomNew() const
{
    return this->mVectUnknownDfsDomNew;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector< DataField<CT,DIM> > const& 
Problem<OWNPRBLM, CT, DIM>::vectUnknownDfsDomOld() const
{
    return this->mVectUnknownDfsDomOld;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector< DataField<CT,DIM> > const& 
Problem<OWNPRBLM, CT, DIM>::vectUnknownDfsBdryNew() const
{
    return this->mVectUnknownDfsBdryNew;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector< DataField<CT,DIM> > const& 
Problem<OWNPRBLM, CT, DIM>::vectUnknownDfsBdryOld() const
{
    return this->mVectUnknownDfsBdryOld;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nColumns() const
{
    return this->mNparams;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::computeGradients() const
{
    return this->mComputeGradients;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::useAsynchronMode() const
{
    return this->mUseAsynchronMode;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::isParamsBased() const
{
    return this->mIsParamsBased;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::useLeapfrogIntegration() const
{
    return this->mUseLeapfrogIntegration;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::worldBarrier()
{
    mParams.myWorld().barrier();
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::setGridGlobalFlags()
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Set grid global related flags..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    int nLayersAtBorder = this->params().nLayersAtBorder();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * nLayersAtGlobalBorder = " << nLayersAtBorder
                  << std::endl;
    }

    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    int ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(comp)
#endif
    // Default: Set all nodes of the grid partition as MPI global inner nodes.
    for (ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        this->mKind.memoryPos2IdxNode(idxNode, comp, ii);
        this->mKind(idxNode) = this->globalGrid().SCAFES_GLOBAL_INNER_NODE;
    }
    
    ScaFES::Ntuple<int,DIM> idxNodeFirstGlobal 
        = this->globalGrid().idxNodeFirst();
    ScaFES::Ntuple<int,DIM> idxNodeLastGlobal 
        = this->globalGrid().idxNodeLast();
    ScaFES::Ntuple<int,DIM> idxNodeFirstPart
        = this->globalGrid().partition(this->myRank()).idxNodeFirstSub();
    ScaFES::Ntuple<int,DIM> idxNodeLastPart
        = this->globalGrid().partition(this->myRank()).idxNodeLastSub();

    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * nLayersAtBorder READ access --> shared
    // * idxNodeFirst... READ access --> shared
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(comp, nLayersAtBorder, idxNodeFirstGlobal, idxNodeLastGlobal)
#endif
    for (ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        this->mKind.memoryPos2IdxNode(idxNode, comp, ii);

        for (std::size_t jj = 0; jj < DIM; ++jj) {
            if ( ((idxNodeFirstGlobal.elem(jj) + nLayersAtBorder) > idxNode.elem(jj))
               || ((idxNodeLastGlobal.elem(jj) - nLayersAtBorder) < idxNode.elem(jj)) ) {
                this->mKind(idxNode)
                    += this->globalGrid().SCAFES_GLOBAL_BORDER_NODE;
                this->mKind(idxNode)
                    -= this->globalGrid().SCAFES_GLOBAL_INNER_NODE;
                break;
            }
        }
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Set." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::setGridPartitionFlags(
    int const& maxStencilWidth
)
{
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Set grid partition related flags..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * maxStencilWidth = " << maxStencilWidth
                  << std::endl;
    }

    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    int ii;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(comp)
#endif
    // Default: Set all nodes of the grid partition as MPI local inner nodes.
    for (ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        this->mKind.memoryPos2IdxNode(idxNode, comp, ii);
        this->mKind(idxNode) += this->globalGrid().SCAFES_PARTITION_INNER_NODE;
    }
    ScaFES::Ntuple<int,DIM> idxNodeFirstGlobal 
        = this->globalGrid().idxNodeFirst();
    ScaFES::Ntuple<int,DIM> idxNodeLastGlobal 
        = this->globalGrid().idxNodeLast();
    ScaFES::Ntuple<int,DIM> idxNodeFirstPart
          = this->globalGrid().partition(this->myRank()).idxNodeFirstSub();
    ScaFES::Ntuple<int,DIM> idxNodeLastPart
          = this->globalGrid().partition(this->myRank()).idxNodeLastSub();

    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * maxStencilWidth READ access --> shared
    // * idxNodeFirst... READ access --> shared
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode) \
    shared(comp, maxStencilWidth, idxNodeFirstPart, idxNodeLastPart)
#endif
   for (ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
       this->mKind.memoryPos2IdxNode(idxNode, comp, ii);

        for (std::size_t jj = 0; jj < DIM; ++jj) {
            if ( ((idxNodeFirstPart.elem(jj) + maxStencilWidth) > idxNode.elem(jj))
               || ((idxNodeLastPart.elem(jj) - maxStencilWidth) < idxNode.elem(jj)) ) {
                this->mKind(idxNode)
                    += this->globalGrid().SCAFES_PARTITION_BORDER_NODE;
                this->mKind(idxNode)
                    -= this->globalGrid().SCAFES_PARTITION_INNER_NODE;
                break;
            }
        }
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Set." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::evalGlobInner(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    int const& timestep
)
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Evaluate data fields " << dfDep.at(0).name()
                  << "... at global inner at timestep " << timestep << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    std::size_t ii;
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->evalInner(
            dfDep,
            idxNode,
            timestep
        );
    }
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->evalInner(
            dfDep,
            idxNode,
            timestep
        );
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Evaluated."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::evalGlobBorder(
    std::vector< ScaFES::DataField<TT,DIM> >& dfDep,
    int const& timestep
)
{
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Evaluate data fields " << dfDep.at(0).name()
                  << "... at global border at timestep " << timestep << "..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    std::size_t ii;
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->evalBorder(
            dfDep,
            idxNode,
            timestep
        );
    }
#ifdef _OPENMP
#pragma omp parallel for \
    schedule(static) \
    default(none) \
    private(ii, idxNode, comp) \
    shared(dfDep, timestep)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        static_cast<OWNPRBLM*>(this)->evalBorder(
            dfDep,
            idxNode,
            timestep
        );
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank()
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Evaluated."
                  << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::getNnodesLocInnerGlobInner() const
{
    int nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_INNER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_INNER_NODE) {
                ++nNodes;
            }
        }
    }
    return nNodes;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::getNnodesLocInnerGlobBorder() const
{
    int nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_INNER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_BORDER_NODE) {
                ++nNodes;
            }
        }
    }
    return nNodes;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::getNnodesLocBorderGlobInner() const
{
    int nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_BORDER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_INNER_NODE) {
                ++nNodes;
            }
        }
    }
    return nNodes;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::getNnodesLocBorderGlobBorder() const
{
    int nNodes = 0;
    for (int ii = 0; ii < this->nNodesTotalPartition(); ++ii) {
        if (this->mKind(ii) & this->globalGrid().SCAFES_PARTITION_BORDER_NODE) {
            if (this->mKind(ii) & this->globalGrid().SCAFES_GLOBAL_BORDER_NODE) {
                ++nNodes;
            }
        }
    }
    return nNodes;
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::update(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
    this->template updatePartitionInner<TT>(dfDep, dfIndep, timestep);
    this->template updatePartitionBorder<TT>(dfDep, dfIndep, timestep);
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionInner(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update at local inner nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }  
    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;

#ifdef SCAFES_HAVE_CUDA
    std::vector<TT*> vectDfDepElemData;
    std::vector<TT*> vectDfIndepElemData;
    for (ii = 0; ii < dfDep.size(); ++ii) {
        vectDfDepElemData.push_back( dfDep[ii].elemData() );
    }
    for (ii = 0; ii < dfIndep.size(); ++ii) {
        vectDfIndepElemData.push_back( dfIndep[ii].elemData() );
    }
    updatePartitionAllInnerUsingCudaVect(
        this->mKind.elemData(),
        vectDfDepElemData,
        vectDfIndepElemData,
        timestep,
        this->globalGrid().SCAFES_GLOBAL_INNER_NODE,
        this->nNodesTotalPartition()
    );
#else
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
    // * timestep READ access --> shared
#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
#pragma omp barrier
#endif

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }
#endif

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionBorder(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update at local border nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }
    
    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;
    // * ii Iteration variable ---> private
    // * idxNode WRITE access Corresponding global node number ---> private
    // * dfIndep READ access --> shared
    // * dfDep WRITE access, but only dfDep(idxNode) will be changed
    //   within the functions given by the corresponding function pointers!
    //   --> The memory locaton is different for all threads --> shared.
    // * timestep READ access --> shared
#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
#pragma omp barrier
#endif

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionAllInner(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update at global inner nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }

    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp barrier
#endif

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionAllBorder(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update at global border nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }

    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp barrier
#endif

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionInner2(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update 2 at local inner nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }

    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
#pragma omp barrier
#endif

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionBorder2(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update 2 at local border nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }

    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
#pragma omp barrier
#endif

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionAllInner2(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update 2 at global inner nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }

    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
#pragma omp barrier
#endif

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobInner.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobInner[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateInner2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
template<typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updatePartitionAllBorder2(
    std::vector< DataField<TT,DIM> >& dfDep,
    std::vector< DataField<TT,DIM> > const& dfIndep,
    int const& timestep
)
{
#ifdef _OPENMP
    #pragma omp single
#endif
  {
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Update 2 at global border nodes: "
                  << dfDep.at(0).name()
                  << ", " << dfIndep.at(0).name() << ",...."
                  << std::endl;
    }
    this->mParams.decreaseLevel();
  }

    std::size_t ii;
    std::size_t jj;
    ScaFES::Ntuple<int, DIM> idxNode;
    int comp = 0;

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocBorderGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocBorderGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
#pragma omp for \
    schedule(static)
#endif
    for (ii = 0; ii < this->mNodesLocInnerGlobBorder.size(); ++ii) {
        int idxNodeScalar = this->mNodesLocInnerGlobBorder[ii];
        this->mKind.memoryPos2IdxNode(idxNode, comp, idxNodeScalar);
        for (jj = 0; jj < dfDep.size(); ++jj) {
            dfDep[jj](idxNode) = dfIndep[jj](idxNode);   
        }
        static_cast<OWNPRBLM*>(this)->updateBorder2(
            dfDep,
            dfIndep,
            idxNode,
            timestep
        );
    }

#ifdef _OPENMP
    #pragma omp single
#endif
  {
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Updated." << std::endl;
    }
  }
}
/*----------------------------------------------------------------------------*/
template<class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::writeDfsToFile(int const& timestep)
{
    // Check if file should be written at current time step.
    int intervalOfWrites = this->params().nTimesteps();

    if (0 < this->params().nSnapshots()) {
        if (this->params().nSnapshots() <= this->params().nTimesteps()) {
            intervalOfWrites = this->params().nTimesteps()
                               / this->params().nSnapshots();
        }
    }

    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << " * Write data fields to file "
                  << mCommonFile.nameDataFile() << " ..." << std::endl;
    }
    this->mParams.decreaseLevel();

    // Do not consider those data fields which should be never written.
    int nDataFieldsToWrite = 0;
     for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
        if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
            ++nDataFieldsToWrite;
        }
    }

    if (0 < nDataFieldsToWrite) {
        int kk = 0;
        std::vector<bool> writeData(nDataFieldsToWrite, false);
        for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
            if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                switch (this->mWriteToFile.at(ii)) {
                case ScaFES::WriteHowOften::AT_START:
                    if (0 == timestep) {
                        writeData.at(kk) = true;
                    }
                    break;
                case ScaFES::WriteHowOften::AT_START_AND_END:
                    if (0 == timestep
                        || this->params().nTimesteps() == timestep) {
                        writeData.at(kk) = true;
                    }
                    break;
                case ScaFES::WriteHowOften::LIKE_GIVEN_AT_CL:
                    if (0 == timestep % intervalOfWrites) {
                        writeData.at(kk) = true;
                    }
                    break;
                case ScaFES::WriteHowOften::ALWAYS:
                default:
                    writeData.at(kk) = true;
                    break;
                }
                if (this->params().rankOutput() == this->myRank() 
                    && (0 < this->params().indentDepth())) {
                    std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                                   this->myRank())
                      << " * Status of data field #"
                      << ii << ": "
                      <<  ((writeData.at(kk)) ? "write" : "do not write")
                      << std::endl;
                }
                ++kk;
            }
        }

        std::vector<CT*> tmpElemData;
        tmpElemData.reserve(nDataFieldsToWrite);
        int idxKnownDf = 0;
        int idxUnknownDf = 0;
        for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
            if (0 == this->nLayers().at(ii)) {
                if (this->isKnownDf().at(ii)) {
                    if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                        tmpElemData.push_back(this->mVectKnownDfDom[idxKnownDf].elemData());
                        ++idxKnownDf;
                    }
                }
                else {
                    if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                        tmpElemData.push_back(this->vectUnknownDfsDomNew()[idxUnknownDf].elemData());
                    }
                    ++idxUnknownDf;
                }
            }
        }
        idxKnownDf = 0;
        idxUnknownDf = 0;
        for (std::size_t ii = 0; ii < this->mWriteToFile.size(); ++ii) {
            if (0 < this->nLayers().at(ii)) {
                if (this->isKnownDf().at(ii)) {
                    if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                        tmpElemData.push_back(this->mVectKnownDfBdry[idxKnownDf].elemData());
                        ++idxKnownDf;
                    }
                }
                else {
                    if (ScaFES::WriteHowOften::NEVER != this->mWriteToFile[ii]) {
                        tmpElemData.push_back(this->vectUnknownDfsBdryNew()[idxUnknownDf].elemData());
                    }
                    ++idxUnknownDf;
                }
            }
        }

        this->mCommonFile.write(
            tmpElemData,
            writeData,
            timestep
        );
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank() 
        && (0 < this->params().indentDepth())) {
        std::cout << ScaFES::getPrefix(this->params().indentDepth(),
                                       this->myRank())
                  << "   Written." << std::endl;
    }
}

} // End of namespace.
#endif
